
//			CHARENCODING command:


#ifndef CHARENCODING_COMMAND_VER
#define CHARENCODING_COMMAND_VER		1.0.0.0


#include "HtmlEntities.cpp"

#define EXPAND_ESCAPES_REAL_BACKSPACE
#include "ExpandEscapes.cpp"

#ifndef MACRO_STR
#define MACRO_EXP( arg )				#arg
#define MACRO_STR( arg )				MACRO_EXP( arg )
#endif	//	MACRO_STR


#define CE_FLAG_SHOW_DECIMAL			0x0010
#define CE_FLAG_SHOW_CLASS				0x0020
#define CE_FLAG_SHOW_UTF16				0x0040
#define CE_FLAG_SHOW_UTF8				0x0080
#define CE_FLAG_SHOW_CHARACTER			0x0100



const wchar_t CharEncodingHelpText[]	= L"Show UTF-16 and UTF-8 encodings for characters.\n\
\n\
CHARENCODING /16 /8 /C /D /K /X value \"string\" ...\r\n\
\n\
\t/16\tUTF-16 encoding\n\
\t/8\tUTF-8 encoding\n\
\t/C\tcharacters\n\
\t/D\tdecimal values\n\
\t/K\tshow character class\n\
\t/X\texpand C-style escapes\n\
\n\
You may enter characters as quoted string literals, character\n\
values, HTML character entities, or any combination.  You may  \n\
prefix hex values with 0x or U+ but neither is required.  With  \n\
or without either prefix, hexadecimal is assumed.  Separate\n\
values with spaces.  If you specify neither /16 nor /8, the\n\
default is to show both.\n";



BOOL Noncharacter( unsigned int ch )
{
	if ( ( ch >= 0xfdd0 ) && ( ch <= 0xfdef ) )
		return TRUE;

	if ( ( ch & 0xffff ) >= 0xfffe )
		return TRUE;

	return FALSE;
}



DLLExports int WINAPI charencoding( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t temp[MAX_ARG_LEN]		= L"";
	wchar_t tmp2[MAX_ARG_LEN]		= L"";

	wchar_t *sep1			= NULL;
	unsigned int val		= 0;
	size_t elen				= 0;
	size_t j				= 0;
	unsigned int k			= 0;

	BOOL Expand				= FALSE;
	unsigned int Flags		= 0;
	
	wchar_t Class			= '\0';
	BYTE b1					= 0;
	BYTE b2					= 0;
	BYTE b3					= 0;
	BYTE b4					= 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_RETAIN_QUOTES );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( CharEncodingHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( CHARENCODING_COMMAND_VER ) );
			else if ( !wcscmp( arg[i], L"/16" ) )
				Flags |= CE_FLAG_SHOW_UTF16;
			else if ( !wcscmp( arg[i], L"/8" ) )
				Flags |= CE_FLAG_SHOW_UTF8;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags |= CE_FLAG_SHOW_CHARACTER;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				Flags |= CE_FLAG_SHOW_DECIMAL;
			else if ( !_wcsicmp( arg[i], L"/K" ) )
				Flags |= CE_FLAG_SHOW_CLASS;
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;

				if ( !_wcsnicmp( arg[i] + j, L"TF", 2 ) )
					j += 2;

				if ( arg[i][j] == '-' )
					j++;

				if ( !wcscmp( arg[i] + j, L"16" ) )
					Flags |= CE_FLAG_SHOW_UTF16;
				else if ( !wcscmp( arg[i] + j, L"8" ) )
					Flags |= CE_FLAG_SHOW_UTF8;
				else {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
					return 0;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				Expand = TRUE;
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( arg[i][0] == QUOTE ) {
				wcscpy_s( tmp2, MAX_ARG_LEN, arg[i] + 1 );

				j = wcslen( tmp2 );
				if ( j )
					if ( tmp2[j - 1] == QUOTE )
						tmp2[--j] = '\0';

				if ( tmp2[0] ) {
					if ( Expand )
						ExpandEscapes( tmp2, MAX_ARG_LEN );

					wcscat_s( temp, MAX_ARG_LEN, tmp2 );
				}
			}
			else {
				val		= 0;
				sep1	= NULL;

				if ( !_wcsicmp( arg[i], L"BOM" ) )
					val = BOM;
				else if ( ( arg[i][0] == '0' ) && ( towupper( arg[i][1] ) == 'X' ) && iswxdigit( arg[i][2] ) )
					val = wcstoul( arg[i] + 2, &sep1, 16 );
				else if ( ( towupper( arg[i][0] ) == 'U' ) && ( arg[i][1] == '+' ) && iswxdigit( arg[i][2] ) )
					val = wcstoul( arg[i] + 2, &sep1, 16 );
				else if ( ( arg[i][0] == '#' ) && iswdigit( arg[i][1] ) )
					val = wcstoul( arg[i] + 1, &sep1, 10 );
				else if ( iswxdigit( arg[i][0] ) )
					val = wcstoul( arg[i], &sep1, 16 );

#ifdef NUM_HTML_ENTITIES

				else if ( arg[i][0] == '&' ) {
					val = ParseEntity( arg[i], &elen );

					if ( !val ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unknown entity: \"%s\"\r\n", arg[i] );
						return 1;
					}
				}

#endif	//	NUM_HTML_ENTITIES

				else if ( ( arg[i][0] > 0 ) && ( ( arg[i][0] & SURROGATE_SIG_MASK ) != SURROGATE_SIG_HIGH ) && ( ( arg[i][0] & SURROGATE_SIG_MASK ) != SURROGATE_SIG_LOW ) && ( arg[i][1] == '\0' ) )
					val = arg[i][0];
				else if ( ( ( arg[i][0] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) && ( ( arg[i][1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) && ( arg[i][2] == '\0' ) )
					val = ( ( ( arg[i][0] & SURROGATE_MASK ) << 10 ) | ( arg[i][1] & SURROGATE_MASK ) ) + SURROGATE_OFFSET;
				else {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown character: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( sep1 ) {
					if ( *sep1 ) {
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad character value: \"%s\"\r\n", arg[i] );
						return 1;
					}
				}

				if ( ( val == 0 ) || ( val > 0x10ffff ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad character value; should be 0x0001 - 0x10FFFF: 0x%04X\r\n", val );
					return 1;
				}

				k = ( val > 0xffff ) ? 2 : 1;

				if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
					return 1;
				}

				if ( k == 2 ) {
					temp[j++]	= ( ( val - SURROGATE_OFFSET ) >> 10 ) | SURROGATE_SIG_HIGH;
					temp[j++]	= ( ( val - SURROGATE_OFFSET ) & SURROGATE_MASK ) | SURROGATE_SIG_LOW;
					temp[j]		= '\0';
					//	wprintf( L"   * Surrogates:  %04X  %04X\n", temp[j - 2], temp[j - 1] );
				}
				else {
					temp[j++]	= val;
					temp[j]		= '\0';
				}
			}
		}
	}


	if ( ( Flags & ( CE_FLAG_SHOW_UTF16 | CE_FLAG_SHOW_UTF8 ) ) == 0 )
		Flags |= ( CE_FLAG_SHOW_UTF16 | CE_FLAG_SHOW_UTF8 );


	//		Show header :

	CrLf();

	QPuts( L"Unicode   " );

	if ( Flags & CE_FLAG_SHOW_DECIMAL )
		QPuts( L"Decimal  " );

	if ( Flags & CE_FLAG_SHOW_CLASS )
		QPuts( L"K  " );

	if ( Flags & CE_FLAG_SHOW_UTF16 )
		QPuts( L"UTF-16     " );

	if ( Flags & CE_FLAG_SHOW_UTF8 )
		QPuts( L"UTF-8        " );

	if ( Flags & CE_FLAG_SHOW_CHARACTER )
		QPuts( L"Ch  " );

	QPuts( L"\r\n" );

	QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_DECIMAL )
		QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_CLASS )
		QPuts( L"\u2015  " );

	if ( Flags & CE_FLAG_SHOW_UTF16 )
		QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_UTF8 )
		QPuts( L"\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015  " );

	if ( Flags & CE_FLAG_SHOW_CHARACTER )
		QPuts( L"\u2015\u2015  " );

	CrLf();


	//		Iterate through string :

	for ( size_t i = 0; temp[i]; i++ ) {

		val = temp[i];
		Class = '-';

		if ( ( ( val & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) && ( ( temp[i + 1] & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) ) {
			val = ( ( ( val & SURROGATE_MASK ) << 10 ) | ( temp[i + 1] & SURROGATE_MASK ) ) + SURROGATE_OFFSET;
			i++;
		}

		if ( val == 0xfeff )
			Class = 'B';
		else if ( Noncharacter( val ) )
			Class = 'N';
		else if ( ( val < 0x10000 ) && ( ( val & SURROGATE_SIG_MASK ) == SURROGATE_SIG_HIGH ) )
			Class = 'H';
		else if ( ( val < 0x10000 ) && ( ( val & SURROGATE_SIG_MASK ) == SURROGATE_SIG_LOW ) )
			Class = 'L';
		else if ( iswcntrl( val ) )
			Class = 'C';
		else if ( iswspace( val ) )
			Class = 'W';
		else if ( iswalpha( val ) )
			Class = 'A';
		else if ( iswdigit( val ) )
			Class = 'D';
		else if ( iswpunct( val ) )
			Class = 'P';


		Printf( L"U+%06X  ", val );

		if ( Flags & CE_FLAG_SHOW_DECIMAL )
			Printf( L"%7u  ", val );

		if ( Flags & CE_FLAG_SHOW_CLASS )
			Printf( L"%c  ", Class );

		if ( Flags & CE_FLAG_SHOW_UTF16 ) {
			if ( val > 0xffff )
				Printf( L"%04X %04X  ", ( ( val - SURROGATE_OFFSET ) >> 10 ) | SURROGATE_SIG_HIGH, ( ( val - SURROGATE_OFFSET ) & SURROGATE_MASK ) | SURROGATE_SIG_LOW );
			else
				Printf( L"%04X       ", val );
		}

		if ( Flags & CE_FLAG_SHOW_UTF8 ) {
			if ( val < 0x80 )
				Printf( L"%02X           ", val );
			else if ( val < 0x800 ) {
				b1 = ( val >> 6 ) | 0xc0;
				b2 = ( val & 0x3f ) | 0x80;
				Printf( L"%02X %02X        ", b1, b2 );
			}
			else if ( val < 0x10000 ) {
				b1 = ( val >> 12 ) | 0xe0;
				b2 = ( ( val >> 6 ) & 0x3f ) | 0x80;
				b3 = ( val & 0x3f ) | 0x80;
				Printf( L"%02X %02X %02X     ", b1, b2, b3 );
			}
			else {
				b1 = ( val >> 18 ) | 0xf0;
				b2 = ( ( val >> 12 ) & 0x3f ) | 0x80;
				b3 = ( ( val >> 6 ) & 0x3f ) | 0x80;
				b4 = ( val & 0x3f ) | 0x80;
				Printf( L"%02X %02X %02X %02X  ", b1, b2, b3, b4 );
			}
		}

		if ( Flags & CE_FLAG_SHOW_CHARACTER ) {
			if ( wcschr( L"-ADPW", Class ) ) {
				if ( val <= 0xffff )
					Printf( L"%c   ", val );
				else
					Printf( L"%c%c  ", ( ( val - SURROGATE_OFFSET ) >> 10 ) | SURROGATE_SIG_HIGH, ( ( val - SURROGATE_OFFSET ) & SURROGATE_MASK ) | SURROGATE_SIG_LOW );
			}
			else
				QPuts( L"    " );
		}

		CrLf();
	}

	return 0;
}


#endif	//	CHARENCODING_COMMAND_VER


/*

	1.0.0.0		2023-07-27

*/
