
//			COUNTCHARS command:


#ifndef COUNTCHARS_COMMAND_VER
#define COUNTCHARS_COMMAND_VER				1.0.1.1


#define COUNTCHARS_FLAG_VERBOSE_HEADER		0x0001
#define COUNTCHARS_FLAG_UPPERCASE			0x0002
#define COUNTCHARS_FLAG_ZEROS				0x0004
#define COUNTCHARS_FLAG_SORT				0x0008
#define COUNTCHARS_FLAG_CONSOLE				0x0010
#define COUNTCHARS_FLAG_RANGE_TOTALS		0x0020
#define COUNTCHARS_FLAG_NO_CHAR_COUNTS		0x0040
#define COUNTCHARS_FLAG_NO_OTHERS_COUNT		0x0080
#define COUNTCHARS_FLAG_NO_TOTAL_COUNT		0x0100
#define COUNTCHARS_FLAG_STRIP_ACCENTS		0x0200

#define COUNTCHARS_MAX_RANGES				32




const wchar_t CountCharsHelpText[]	= L"Count characters in text files.\n\
\n\
COUNTCHARS /C:x-y /O /P /R /RO /S /U /V /W /X filespec....\r\n\
\n\
\t/C:   range of characters to count\n\
\t/O    sort by frequency\n\
\t/P    page output\n\
\t/R    report counts for ranges as well as characters\n\
\t/RO   report range counts only\n\
\t/S    search in subdirectories for matching files\n\
\t/U    force characters to uppercase\n\
\t/V    do not merge overlapping ranges\n\
\t/W    do not report count of other characters\n\
\t/X    do not report total characters count\n\
\n\
COUNTCHARS will count characters in the specified ranges,\n\
in all matching files.  If no ranges are given with /C:,\n\
the default is /C:0-127 (ASCII characters).  You may list\n\
up to 32 ranges.\n";



const wchar_t CountChars_nSpacesText[]	= L"                    ";




struct _CountCharsDuo {
	unsigned __int64 InFile;
	unsigned __int64 Total;
};


struct _CountCharsSortEntry {
	unsigned __int64 Count;
	int ch;
};


struct _CountCharsRange {
	int FirstChar;
	int LastChar;
	int n;
	_CountCharsDuo RangeTotal;
	_CountCharsDuo *Counts;
};


struct _CountChars_Ernie {
	int FirstChar;
	int LastChar;
	unsigned __int64 Count;
};


struct _CountCharsArgs {
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
	_CountCharsDuo Other;
	_CountCharsDuo TotalChars;
	unsigned int FilesProcessed;
	unsigned int NumRanges;
	_CountCharsRange *Ranges;
};




int CountChars_RangeCmp$( const void *A, const void *B )
{
	_CountCharsRange *a		= (_CountCharsRange*) A;
	_CountCharsRange *b		= (_CountCharsRange*) B;


	if ( a->FirstChar < b->FirstChar )
		return -1;
	else if ( a->FirstChar > b->FirstChar )
		return 1;
	else if ( a->LastChar < b->LastChar )
		return -1;
	else if ( b->LastChar > b->LastChar )
		return 1;
	else
		return 0;
}



int CountChars_ErnieCmp$( const void *A, const void *B )
{
	_CountChars_Ernie *a	= (_CountChars_Ernie*) A;
	_CountChars_Ernie *b	= (_CountChars_Ernie*) B;


	if ( a->Count < b->Count )
		return 1;
	else if ( a->Count > b->Count )
		return -1;
	else
		return 0;
}



int CountChars_PrintChar( int ch )
{
	if ( ch > 0xffff )
		return ch;

	if ( iswprint( ch ) && !iswcntrl( ch ) )
		return ch;

	return 0xb7;
}



LPTSTR CountChars_nSpaces( unsigned int n )
{
	if ( n > 20 )
		n = 20;
	
	return (LPTSTR) CountChars_nSpacesText + 20 - n;
}



int CountChars_AllocCounts( _CountCharsRange *Ranges, unsigned int n )
{
	if ( !Ranges || !n )
		return -1;


	for ( unsigned int i = 0; i < n; i++ ) {
		Ranges[i].n = Ranges[i].LastChar - Ranges[i].FirstChar + 1;
		Ranges[i].Counts = (_CountCharsDuo*) malloc( Ranges[i].n * sizeof( _CountCharsDuo ) );
		if ( Ranges[i].Counts == NULL )
			return ERROR_NOT_ENOUGH_MEMORY;

		memset( Ranges[i].Counts, 0x00, Ranges[i].n * sizeof( _CountCharsDuo ) );
	}

	return 0;
}



int CountChars_FreeCounts( _CountCharsRange *Ranges, unsigned int n )
{
	if ( !Ranges || !n )
		return -1;


	int k		= 0;


	for ( unsigned int i = 0; i < n; i++ ) {
		if ( Ranges[i].Counts ) {
			free( Ranges[i].Counts );
			Ranges[i].Counts = NULL;
			k++;
		}
	}

	return k;
}



int CountChars_Cmp$( const void *Left, const void *Right )
{
	_CountCharsSortEntry *left		= (_CountCharsSortEntry*) Left;
	_CountCharsSortEntry *right		= (_CountCharsSortEntry*) Right;


	if ( left->Count < right->Count )
		return 1;
	else if ( left->Count > right->Count )
		return -1;
	else if ( left->ch < right->ch )
		return -1;
	else if ( left->ch > right->ch )
		return 1;
	else
		return 0;
}



int CountChars_MakeHexString( int ch, LPTSTR outHexString, LPTSTR outCharChar )
{
	if ( !outHexString )
		return -1;

	if ( ( ch < 0 ) || ( ch > 0x10ffff ) )
		return -2;


	if ( ch <= 0xffff )
		wsprintf( outHexString, L"  %04X", ch );
	else if ( ch <= 0xfffff )
		wsprintf( outHexString, L" %05X", ch );
	else
		wsprintf( outHexString, L"%06X", ch );


	if ( outCharChar ) {
		if ( ch <= 0xffff ) {
			ch = CountChars_PrintChar( ch );
			outCharChar[0] = ch;
			outCharChar[1] = '\0';
		}
		else {
			MM_CreateSurrogatePair( ch, outCharChar );
			outCharChar[2] = '\0';
		}
	}
	
	return 0;
}



int CountChars_DumpCountsSorted( _CountCharsArgs *Args, BOOL GrandTotal, int Flags )
{
	if ( !Args )
		return -1;


	wchar_t CharHex1[16]			= L"";
	wchar_t CharHex2[16]			= L"";
	wchar_t CharChar[3]				= L"";

	_CountCharsRange *Ranges		= Args->Ranges;
	unsigned int n					= Args->NumRanges;
	unsigned __int64 OthersCount	= GrandTotal ? Args->Other.Total : Args->Other.InFile;
	unsigned __int64 TotalChars		= GrandTotal ? Args->TotalChars.Total : Args->TotalChars.InFile;

	_CountCharsSortEntry *SortEntries	= NULL;

	unsigned int ArraySize				= 0;
	unsigned int k						= 0;

	unsigned int AddSpaces				= 0;


	if ( Flags & COUNTCHARS_FLAG_RANGE_TOTALS )
		AddSpaces += 6;


	if ( ( Flags & COUNTCHARS_FLAG_NO_CHAR_COUNTS ) == 0 ) {

		for ( unsigned int i = 0; i < n; i++ )
			ArraySize += Ranges[i].n;

		SortEntries = (_CountCharsSortEntry*) malloc( ArraySize * sizeof( _CountCharsSortEntry ) );
		if ( !SortEntries )
			return ERROR_NOT_ENOUGH_MEMORY;

		memset( SortEntries, 0x00, ArraySize * sizeof( _CountCharsSortEntry ) );

		k = 0;

		for ( unsigned int i = 0; i < n; i++ ) {
			for ( int j = 0; j < Ranges[i].n; j++ ) {
				SortEntries[k].ch = Ranges[i].FirstChar + j;
				SortEntries[k].Count = GrandTotal ? Ranges[i].Counts[j].Total : Ranges[i].Counts[j].InFile;
				k++;
			}
		}


		if ( Flags & COUNTCHARS_FLAG_CONSOLE )
			_cputws( L"\rSorting: " );

		qsort( SortEntries, ArraySize, sizeof( _CountCharsSortEntry ), CountChars_Cmp$ );

		if ( Flags & COUNTCHARS_FLAG_CONSOLE )
			_cputws( L"\r                    \r" );


		for ( unsigned int i = 0; i < ArraySize; i++ ) {
			if ( ( Flags & COUNTCHARS_FLAG_ZEROS ) || SortEntries[i].Count ) {
				CountChars_MakeHexString( SortEntries[i].ch, CharHex1, CharChar );

				Printf( L"  %s  %s %s:  %I64u", CharHex1, CharChar, CountChars_nSpaces( AddSpaces ), SortEntries[i].Count );

				if ( CL_EndLine() == USER_ABORT )
					goto UserAbort;
			}
		}
	}


	if ( Flags & COUNTCHARS_FLAG_RANGE_TOTALS ) {
		_CountChars_Ernie Ernie[COUNTCHARS_MAX_RANGES];
		memset( Ernie, 0x00, sizeof( Ernie ) );

		for ( unsigned int i = 0; i < n; i++ ) {
			Ernie[i].FirstChar	= Ranges[i].FirstChar;
			Ernie[i].LastChar	= Ranges[i].LastChar;
			Ernie[i].Count		= GrandTotal ? Ranges[i].RangeTotal.Total : Ranges[i].RangeTotal.InFile;
		}

		qsort( Ernie, n, sizeof( _CountChars_Ernie ), CountChars_ErnieCmp$ );

		for ( unsigned int i = 0; i < n; i++ ) {
			CountChars_MakeHexString( Ernie[i].FirstChar, CharHex1, NULL );
			CountChars_MakeHexString( Ernie[i].LastChar, CharHex2, NULL );
			Printf( L"  %s - %s :  %I64u", CharHex1, CharHex2, Ernie[i].Count );

			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;
		}
	}


	if ( ( Flags & COUNTCHARS_FLAG_NO_OTHERS_COUNT ) == 0 ) {
		if ( OthersCount || ( Flags & COUNTCHARS_FLAG_ZEROS ) )  {
			Printf( L"    Other  %s :  %I64u", CountChars_nSpaces( AddSpaces ), OthersCount );
			if ( CL_EndLine() == USER_ABORT )
				goto UserAbort;
		}
	}

	if ( ( Flags & COUNTCHARS_FLAG_NO_TOTAL_COUNT ) == 0 ) {
		Printf( L"    TOTAL  %s :  %I64u", CountChars_nSpaces( AddSpaces ), TotalChars );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}

	free( SortEntries );
	return 0;

UserAbort:

	free( SortEntries );
	return USER_ABORT;
}



int CountChars_DumpCounts( _CountCharsArgs *Args, BOOL GrandTotal, int Flags )
{
	if ( !Args )
		return -1;


	if ( Flags & COUNTCHARS_FLAG_SORT )
		return CountChars_DumpCountsSorted( Args, GrandTotal, Flags );


	wchar_t CharHex1[16]			= L"";
	wchar_t CharHex2[16]			= L"";
	wchar_t CharChar[3]				= L"";

	_CountCharsRange *Ranges		= Args->Ranges;
	unsigned int n					= Args->NumRanges;
	unsigned __int64 OthersCount	= GrandTotal ? Args->Other.Total : Args->Other.InFile;
	unsigned __int64 TotalChars		= GrandTotal ? Args->TotalChars.Total : Args->TotalChars.InFile;

	unsigned __int64 Count			= 0;
	int ch							= '\0';

	unsigned int AddSpaces			= 0;


	if ( Flags & COUNTCHARS_FLAG_RANGE_TOTALS )
		AddSpaces += 6;


	if ( ( Flags & COUNTCHARS_FLAG_NO_CHAR_COUNTS ) == 0 ) {

		for ( unsigned int i = 0; i < n; i++ ) {
			for ( int j = 0; j < Ranges[i].n; j++ ) {
				Count = GrandTotal ? Ranges[i].Counts[j].Total : Ranges[i].Counts[j].InFile;

				if ( ( Flags & COUNTCHARS_FLAG_ZEROS ) || Count ) {
					ch = Ranges[i].FirstChar + j;
					CountChars_MakeHexString( ch, CharHex1, CharChar );

					Printf( L"  %s  %s %s:  %I64u", CharHex1, CharChar, CountChars_nSpaces( AddSpaces ), Count );

					if ( CL_EndLine() == USER_ABORT )
						return USER_ABORT;
				}
			}
		}
	}


	if ( Flags & COUNTCHARS_FLAG_RANGE_TOTALS ) {
		for ( unsigned int i = 0; i < n; i++ ) {
			Count = GrandTotal ? Ranges[i].RangeTotal.Total : Ranges[i].RangeTotal.InFile;

			CountChars_MakeHexString( Ranges[i].FirstChar, CharHex1, NULL );
			CountChars_MakeHexString( Ranges[i].LastChar, CharHex2, NULL );
			Printf( L"  %s - %s :  %I64u", CharHex1, CharHex2, Count );

			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;
		}
	}


	if ( ( Flags & COUNTCHARS_FLAG_NO_OTHERS_COUNT ) == 0 ) {
		if ( OthersCount || ( Flags & COUNTCHARS_FLAG_ZEROS ) )  {
			Printf( L"    Other  %s :  %I64u", CountChars_nSpaces( AddSpaces ), OthersCount );
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;
		}
	}

	if ( ( Flags & COUNTCHARS_FLAG_NO_TOTAL_COUNT ) == 0 ) {
		Printf( L"    TOTAL  %s :  %I64u", CountChars_nSpaces( AddSpaces ), TotalChars );
		if ( CL_EndLine() == USER_ABORT )
			return USER_ABORT;
	}

	return 0;
}



int CountChars_RangesOverlap( _CountCharsRange *A, _CountCharsRange *B )
{
	/*		Returns:

			0		ranges are disjoint and not consecutive
			1		ranges are consecutive
			2		ranges overlap
			3		ranges are identical
	*/

	if ( !A || !B )
		return 0;


	if ( ( A->FirstChar == B->FirstChar ) && ( A->LastChar == B->LastChar ) )
		return 3;

	if ( ( B->FirstChar >= A->FirstChar ) && ( B->FirstChar <= A->LastChar ) )
		return 2;

	if ( ( B->LastChar >= A->FirstChar ) && ( B->LastChar <= A->LastChar ) )
		return 2;

	if ( ( A->FirstChar >= B->FirstChar ) && ( A->FirstChar <= B->LastChar ) )
		return 2;

	if ( ( A->LastChar >= B->FirstChar ) && ( A->LastChar <= B->LastChar ) )
		return 2;

	if ( A->FirstChar == ( B->LastChar + 1 ) )
		return 1;

	if ( B->FirstChar == ( A->LastChar + 1 ) )
		return 1;

	return 0;
}



int CountChars_MergeRanges( _CountCharsRange *A, _CountCharsRange *B )
{
	if ( !A || !B )
		return -1;


	if ( A->FirstChar > B->FirstChar )
		A->FirstChar = B->FirstChar;

	if ( A->LastChar < B->LastChar )
		A->LastChar = B->LastChar;

	return 0;
}



int CountChars_CombineRanges( _CountCharsRange *Ranges, unsigned int *n )
{
	if ( !Ranges || !n )
		return -1;

	if ( *n < 2 )
		return 0;


	int First			= 0;
	int Second			= 0;
	BOOL Change			= FALSE;
	int rv				= 0;


	do {
		Change = FALSE;

		for ( unsigned int i = 0; i < ( *n - 1 ); i++ ) {
			for ( unsigned int j = i + 1; j < *n; j++ ) {
				//	wprintf( L"   * Compare Ranges[%u] against Ranges[%u] :  %d\n", i, j, CountChars_RangesOverlap( Ranges + i, Ranges + j ) );

				if ( CountChars_RangesOverlap( Ranges + i, Ranges + j ) > 1 ) {
					First = i;
					Second = j;
					Change = TRUE;
					break;
				}
			}

			if ( Change )
				break;
		}


		if ( Change ) {
			//	wprintf( L"   * Need to combine range #%u ( %04X - %04X ) with range #%u ( %04X - %04X ) :\n", First, Ranges[First].FirstChar, Ranges[First].LastChar, Second, Ranges[Second].FirstChar, Ranges[Second].LastChar );

			CountChars_MergeRanges( Ranges + First, Ranges + Second );
			//	wprintf( L"   * Range[%u] is now %04X - %04X.\n", First, Ranges[First].FirstChar, Ranges[First].LastChar );

			if ( Second == ( *n - 1 ) ) {
				*n -= 1 ;
				//	wprintf( L"   * Removed final range; n is now %u.\n", *n );
			}
			else {
				//	wprintf( L"   * Move ranges %u to %u down:\n", Second, *n - 1 );

				for ( unsigned int i = Second; i < ( *n - 1 ); i++ )
					Ranges[i] = Ranges[i + 1];

				*n -= 1;
				//	wprintf( L"   * n is now %u.\n", *n );
			}
		}

	}
	while ( Change );

	return rv;
}



int CountChars_DumpRanges( _CountCharsRange *Ranges, unsigned int n )
{
	if ( !Ranges || !n )
		return -1;


	for ( unsigned int i = 0; i < n; i++ ) {
		Printf( L"   Range %2u:  %04X - %04X", i, Ranges[i].FirstChar, Ranges[i].LastChar );
		if ( CL_EndLine() == USER_ABORT )
			return USER_ABORT;
	}

	return 0;
}



int CountChars_ParseRange( LPTSTR inString, int *FirstChar, int *LastChar )
{
	if ( !inString || !FirstChar || !LastChar )
		return -1;
	else if ( inString[0] == '\0' )
		return -1;


	unsigned int First			= 0;
	unsigned int Last			= 0;
	wchar_t *sep1				= NULL;
	wchar_t *sep2				= NULL;


	if ( ( towupper( inString[0] ) == 'U' ) && ( inString[1] == '+' ) && iswxdigit( inString[2] ) ) {
		First = wcstoul( inString + 2, &sep1, 16 );

		if ( First > 0x10ffff ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"FirstChar must be U+0000 - U+10FFFF: U+%04X\r\n", First );
			return 1;
		}
	}
	else if ( ( inString[0] == APOSTROPHE ) && ( iswprint( inString[1] ) && !iswcntrl( inString[1] ) ) && ( inString[2] == APOSTROPHE ) ) {
		First = inString[1];
		sep1 = inString + 3;
	}
	else if ( ( inString[0] == APOSTROPHE ) && MM_DecodeSurrogatePair( inString[1], inString[2] ) && ( inString[3] == APOSTROPHE ) ) {
		First = MM_DecodeSurrogatePair( inString[1], inString[2] );
		sep1 = inString + 4;
	}
	else if ( ( iswprint( inString[0] ) && !iswcntrl( inString[0] ) && !iswdigit( inString[0] ) ) && ( ( inString[1] == '-' ) || ( inString[1] == '\0' ) ) ) {
		First = inString[0];
		sep1 = inString + 1;
	}
	else if ( MM_DecodeSurrogatePair( inString[0], inString[1] ) && ( ( inString[2] == '-' ) || ( inString[2] == '\0' ) ) ) {
		First = MM_DecodeSurrogatePair( inString[0], inString[1] );
		sep1 = inString + 2;
	}
	else if ( ParseInt( inString, &First, &sep1 ) ) {
		_wcsupr_s( inString, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad FirstChar value in character range: \"%s\"\r\n", inString );
		return 1;
	}

	if ( First > 0x10ffff ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"FirstChar must be 0 - 0x10FFFF: 0x%04X\r\n", First );
		return 1;
	}

	if ( *sep1 == '\0' ) {
		*FirstChar = First;
		*LastChar = First;
		return 0;
	}

	if ( *sep1 != '-' ) {
		_wcsupr_s( sep1, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error in character range: \"%s\"\r\n", sep1 );
		return 2;
	}

	sep1++;

	if ( ( towupper( sep1[0] ) == 'U' ) && ( sep1[1] == '+' ) && iswxdigit( sep1[2] ) ) {
		Last = wcstoul( sep1 + 2, &sep2, 16 );

		if ( *sep2 != '\0' ) {
			_wcsupr_s( inString, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad LastChar in character range: \"%s\"\r\n", sep1 );
			return 1;
		}

		if ( Last > 0x10ffff ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"LastChar must be U+0000 - U+10FFFF: U+%04X\r\n", Last );
			return 1;
		}
	}
	else if ( ( sep1[0] == APOSTROPHE ) && ( iswprint( sep1[1] ) && !iswcntrl( sep1[1] ) ) && ( sep1[2] == APOSTROPHE ) && ( sep1[3] == '\0' ) ) {
		Last = sep1[1];
	}
	else if ( ( sep1[0] == APOSTROPHE ) && MM_DecodeSurrogatePair( sep1[1], sep1[2] ) && ( sep1[3] == APOSTROPHE ) && ( sep1[4] == '\0' ) ) {
		Last = MM_DecodeSurrogatePair( sep1[1], sep1[2] );
	}
	else if ( ( iswprint( sep1[0] ) && !iswcntrl( sep1[0] ) && !iswdigit( sep1[0] ) ) && ( sep1[1] == '\0' ) ) {
		Last = sep1[0];
	}
	else if ( MM_DecodeSurrogatePair( sep1[0], sep1[1] ) && ( sep1[2] == '\0' ) ) {
		Last = MM_DecodeSurrogatePair( sep1[0], sep1[2] );
	}
	else if ( ParseInt( sep1, &Last, NULL ) ) {
		_wcsupr_s( inString, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad LastChar in character range: \"%s\"\r\n", sep1 );
		return 3;
	}

	if ( Last > 0x10ffff ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"LastChar must be 0 - 0x10FFFF: 0x%04X\r\n", Last );
		return 3;
	}

	if ( First > Last ) {
		unsigned int t = First;
		First = Last;
		Last = t;
	}

	*FirstChar = (int) First;
	*LastChar = (int) Last;
	return 0;
}



int CountChars_ResetFileCounts( _CountCharsArgs *Args )
{
	if ( !Args )
		return -1;


	for ( unsigned int i = 0; i < Args->NumRanges; i++ ) {
		for ( int j = 0; j < Args->Ranges[i].n; j++ )
			Args->Ranges[i].Counts[j].InFile = 0;

		Args->Ranges[i].RangeTotal.InFile = 0;
	}

	Args->Other.InFile		= 0;
	Args->TotalChars.InFile	= 0;

	return Args->NumRanges;
}



int CountCharsInFile( LPCTSTR Filename, int Flags, void *Args )
{
	if ( !Filename || !Args )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	_MMFile MMFile;

	_CountCharsArgs *args	= (_CountCharsArgs*) Args;

	int ch					= 0;
	int rv					= 0;
	BOOL Found				= FALSE;
	BOOL ForceUppercase		= Flags & COUNTCHARS_FLAG_UPPERCASE;
	BOOL StripAccents		= Flags & COUNTCHARS_FLAG_STRIP_ACCENTS;
	unsigned int j			= 0;


	//	Open file:

	rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC | MM_FLAG_HIGH_CHARS, args->CodePage, args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}

	CL_HookBreak();


	if ( ShowFilenameHeader( &MMFile, ( Flags & COUNTCHARS_FLAG_VERBOSE_HEADER ) > 0 ) == USER_ABORT )
		goto UserAbort;


	//	Count characters in file:

	CountChars_ResetFileCounts( args );

	if ( Flags & COUNTCHARS_FLAG_CONSOLE )
		_cputws( L"\rReading file: " );


	while ( MMFile.Index < MMFile.FileSize ) {

		ch = MMGetChar( &MMFile );
		if ( ch < 0 )
			goto ErrorAbort;

		if ( ch <= 0xffff ) {
			if ( StripAccents )
				StripAccent( ch );

			if ( ForceUppercase )
				ch = towupper( ch );
		}

		Found = FALSE;

		for ( unsigned int i = 0; i < args->NumRanges; i++ ) {
			if ( ( ch >= args->Ranges[i].FirstChar ) && ( ch <= args->Ranges[i].LastChar ) ) {
				Found = TRUE;
				j = ch - args->Ranges[i].FirstChar;
				args->Ranges[i].Counts[j].InFile	+= 1;
				args->Ranges[i].Counts[j].Total		+= 1;
				args->Ranges[i].RangeTotal.InFile	+= 1;
				args->Ranges[i].RangeTotal.Total	+= 1;
			}
		}
		
		if ( !Found ) {
			args->Other.InFile	+= 1;
			args->Other.Total	+= 1;
		}

		args->TotalChars.InFile	+= 1;
		args->TotalChars.Total	+= 1;
	}

	CloseMMFile( &MMFile );

	args->FilesProcessed += 1;

	if ( Flags & COUNTCHARS_FLAG_CONSOLE )
		_cputws( L"\r                    \r" );

	
	//	Report counts:

	if ( CountChars_DumpCounts( args, FALSE, Flags ) == USER_ABORT )
		goto UserAbort;


	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;

ErrorAbort:

	if ( Flags & COUNTCHARS_FLAG_CONSOLE )
		_cputws( L"\r              \r" );

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return MMFile.ErrorCode;
}



DLLExports int WINAPI countchars( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	int rv					= 0;
	int fhrv				= 0;

	_CountCharsRange Ranges[COUNTCHARS_MAX_RANGES];
	unsigned int NumRanges	= 0;

	BOOL CombineRanges		= TRUE;
	BOOL Debug				= FALSE;

	_CountCharsArgs Args;


	CL_NoPausing();

	memset( &Ranges, 0x00, sizeof( Ranges ) );
	memset( &Args, 0x00, sizeof( _CountCharsArgs ) );


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( CountCharsHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( COUNTCHARS_COMMAND_VER ) );
			}
			else if ( !wcscmp( arg[i], L"/$" ) )
				Debug = TRUE;
			else if ( !_wcsicmp( arg[i], L"/ASCII" ) ) {
				if ( NumRanges >= COUNTCHARS_MAX_RANGES ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Too many character ranges (%u max)\r\n", COUNTCHARS_MAX_RANGES );
					goto ErrorExit;
				}

				Ranges[NumRanges].FirstChar	= 0;
				Ranges[NumRanges].LastChar	= 0x7f;
				NumRanges++;
			}
			else if ( !_wcsicmp( arg[i], L"/BMP" ) ) {
				if ( NumRanges >= COUNTCHARS_MAX_RANGES ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Too many character ranges (%u max)\r\n", COUNTCHARS_MAX_RANGES );
					goto ErrorExit;
				}

				Ranges[NumRanges].FirstChar	= 0;
				Ranges[NumRanges].LastChar	= 0xffff;
				NumRanges++;
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( NumRanges >= COUNTCHARS_MAX_RANGES ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Too many character ranges (%u max)\r\n", COUNTCHARS_MAX_RANGES );
					goto ErrorExit;
				}

				if ( CountChars_ParseRange( arg[i] + j, &( Ranges[NumRanges].FirstChar ), &( Ranges[NumRanges].LastChar ) ) )
					goto ErrorExit;

				NumRanges++;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = TRUE;
			else if ( !_wcsicmp( arg[i], L"/G" ) )
				Flags |= COUNTCHARS_FLAG_VERBOSE_HEADER;
			else if ( !_wcsicmp( arg[i], L"/HI" ) ) {
				if ( NumRanges >= COUNTCHARS_MAX_RANGES ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Too many character ranges (%u max)\r\n", COUNTCHARS_MAX_RANGES );
					goto ErrorExit;
				}

				Ranges[NumRanges].FirstChar	= 0x10000;
				Ranges[NumRanges].LastChar	= 0x10ffff;
				NumRanges++;
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, NULL ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/O" ) )
				Flags |= COUNTCHARS_FLAG_SORT;
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Flags = ( Flags | COUNTCHARS_FLAG_RANGE_TOTALS ) & ~COUNTCHARS_FLAG_NO_CHAR_COUNTS;
			else if ( !_wcsicmp( arg[i], L"/RO" ) )
				Flags |= ( COUNTCHARS_FLAG_RANGE_TOTALS | COUNTCHARS_FLAG_NO_CHAR_COUNTS );
			else if ( !_wcsicmp( arg[i], L"/SA" ) )
				Flags |= COUNTCHARS_FLAG_STRIP_ACCENTS;
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				Flags |= COUNTCHARS_FLAG_UPPERCASE;
			else if ( !_wcsicmp( arg[i], L"/V" ) )
				CombineRanges = FALSE;
			else if ( !_wcsicmp( arg[i], L"/W" ) )
				Flags |= COUNTCHARS_FLAG_NO_OTHERS_COUNT;
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				Flags |= COUNTCHARS_FLAG_NO_TOTAL_COUNT;
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				Flags |= COUNTCHARS_FLAG_ZEROS;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );

	if ( ConsoleHandle( STD_OUTPUT_HANDLE ) )
		Flags |= COUNTCHARS_FLAG_CONSOLE;



	if ( !NumRanges ) {
		Ranges[0].FirstChar	= 0;
		Ranges[0].LastChar	= 127;
		NumRanges++;
	}


	if ( CombineRanges )
		CountChars_CombineRanges( Ranges, &NumRanges );

	qsort( Ranges, NumRanges, sizeof( _CountCharsRange ), CountChars_RangeCmp$ );

	if ( Debug )
		if ( CountChars_DumpRanges( Ranges, NumRanges ) == USER_ABORT )
			goto UserAbort;

	CountChars_AllocCounts( Ranges, NumRanges );


	Args.Ranges		= Ranges;
	Args.NumRanges	= NumRanges;



	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = CountCharsInFile( (LPTSTR) StdinFilename, Flags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, CountCharsInFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	if ( Args.FilesProcessed > 1 ) {

		if ( HLprintf( L"\n\x12Grand total for %u files:\n\n", Args.FilesProcessed ) == USER_ABORT )
			goto UserAbort;

		if ( CountChars_DumpCounts( &Args, TRUE, Flags ) == USER_ABORT )
			goto UserAbort;
	}


	FH_FreeLinkedList( &FirstRange );
	CountChars_FreeCounts( Ranges, NumRanges );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	CountChars_FreeCounts( Ranges, NumRanges );
	return 1;

UserAbort :

	FH_FreeLinkedList( &FirstRange );
	CountChars_FreeCounts( Ranges, NumRanges );
	return 3;
}


#endif	//	COUNTCHARS_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-28
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				CountCharsInFile() filename is defined as a constant:  fix for mmfiles 1.4.4.2

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
