#ifndef EXPAND_ESCAPES_CPP
#define EXPAND_ESCAPES_CPP			1.0.4.1


#ifndef MAX_CODEPOINT
#define MAX_CODEPOINT				0x10ffff
#endif	//	MAX_CODEPOINT

#ifndef SURROGATE_MASK
#define SURROGATE_MASK				0xf800
#endif	//	SURROGATE_MASK

#ifndef HIGH_SURROGATE
#define HIGH_SURROGATE				0xd800
#define LOW_SURROGATE				0xdc00
#endif	//	HIGH_SURROGATE



int ParseOctalEscape( LPCTSTR inValue, LPTSTR outBuffer, size_t BufSize )
{
	if ( !inValue || !outBuffer || !BufSize )
		return 0;

	if ( !iswdigit( inValue[0] ) )
		return 0;


	unsigned int Char			= 0;
	int digcount				= 0;
	size_t j					= wcslen( outBuffer );


	for ( int i = 0; i < 3; i++ ) {
		if ( ( inValue[i] < '0' ) || ( inValue[i] > '7' ) )
			break;

		Char = ( Char << 3 ) | ( inValue[i] - '0' );
		digcount++;
	}

	if ( !Char )
		return 0;

	if ( j < ( BufSize - 1 ) ) {
		outBuffer[j++]	= Char;
		outBuffer[j]	= '\0';
	}
	else
		return 0;

	return digcount;
}



int ParseHexEscape( LPCTSTR inValue, LPTSTR outBuffer, size_t BufSize )
{
	if ( !inValue || !outBuffer || !BufSize )
		return 0;

	if ( !iswxdigit( inValue[0] ) )
		return 0;

	
	unsigned int Char			= 0;
	unsigned int digval			= 0;
	int digcount				= 0;
	wchar_t ch					= '\0';
	size_t j					= wcslen( outBuffer );


	for ( int i = 0; i < 4; i++ ) {
		ch = towupper( inValue[i] );

		if ( ( ch >= '0' ) && ( ch <= '9' ) )
			digval = ch - '0';
		else if ( ( ch >= 'A' ) && ( ch <= 'F' ) )
			digval = ch - 'A' + 10;
		else
			break;

		Char = ( Char << 4 ) | digval;
		digcount++;
	}

	if ( !Char )
		return 0;

	if ( j < ( BufSize - 1 ) ) {
		outBuffer[j++]	= Char;
		outBuffer[j]	= '\0';
	}
	else
		return 0;

	return digcount;
}



int ParseDecimalEscape( LPCTSTR inValue, LPTSTR outBuffer, size_t BufSize )
{
	if ( !inValue || !outBuffer || !BufSize )
		return 0;

	if ( !iswdigit( inValue[0] ) )
		return 0;


	unsigned int Char			= 0;
	int digcount				= 0;
	size_t j					= wcslen( outBuffer );


	for ( int i = 0; i < 5; i++ ) {
		if ( ( inValue[i] < '0' ) || ( inValue[i] > '9' ) )
			break;

		Char = ( Char * 10 ) + ( inValue[i] - '0' );
		digcount++;
	}

	if ( !Char || ( Char > 65535 ) )
		return 0;

	if ( j < ( BufSize - 1 ) ) {
		outBuffer[j++]	= Char;
		outBuffer[j]	= '\0';
	}
	else
		return 0;

	return digcount;
}



int ParseUnicodeEscape( LPCTSTR inValue, int MaxDigits, LPTSTR outBuffer, size_t BufSize )
{
	if ( !inValue || !MaxDigits || !outBuffer || !BufSize )
		return 0;

	if ( !iswxdigit( inValue[0] ) )
		return 0;


	unsigned int CodePoint		= 0;
	unsigned int digval			= 0;
	int digcount				= 0;
	wchar_t ch					= '\0';
	wchar_t LowSurrogate		= '\0';
	wchar_t HighSurrogate		= '\0';
	size_t j					= wcslen( outBuffer );


	for ( int i = 0; i < MaxDigits; i++ ) {
		ch = towupper( inValue[i] );

		if ( ( ch >= '0' ) && ( ch <= '9' ) )
			digval = ch - '0';
		else if ( ( ch >= 'A' ) && ( ch <= 'F' ) )
			digval = ch - 'A' + 10;
		else
			break;

		CodePoint = ( CodePoint << 4 ) | digval;
		if ( CodePoint > MAX_CODEPOINT )
			return 0;

		digcount++;
	}


	if ( !CodePoint || ( ( CodePoint & SURROGATE_MASK ) == HIGH_SURROGATE ) )
		return 0;


	if ( CodePoint < 0x10000 ) {
		if ( j < ( BufSize - 1 ) ) {
			outBuffer[j++]	= CodePoint;
			outBuffer[j]	= '\0';
		}
		else
			return 0;
	}
	else {
		CodePoint -= 0x10000;
		LowSurrogate	= ( CodePoint & 0x03ff ) | LOW_SURROGATE;
		HighSurrogate	= ( CodePoint >> 10 ) | HIGH_SURROGATE;

		if ( j < ( BufSize - 2 ) ) {
			outBuffer[j++]	= HighSurrogate;
			outBuffer[j++]	= LowSurrogate;
			outBuffer[j]	= '\0';
		}
		else
			return 0;
	}

	return digcount;
}



int ExpandEscapes( LPTSTR String, size_t buflen )
{
	if ( !String || ( buflen < 20 ) )
		return -1;

	if ( String[0] == '\0' )
		return 0;


	wchar_t *temp		= NULL;
	wchar_t ch2			= '\0';
	size_t j			= 0;
	int k				= 0;
	int trv				= 0;


	temp = (wchar_t*) malloc( buflen * sizeof( wchar_t ) );
	if ( !temp )
		return -1;

	temp[0] = '\0';


	for ( wchar_t *ch = String; *ch; ch++ ) {

		if ( *ch == '\\' ) {
			ch++;
			ch2 = '\0';

			switch ( *ch ) {

			case 'a' :

				ch2 = '\a', k++;
				break;

			case 'b' :

#ifdef EXPAND_ESCAPES_REAL_BACKSPACE

				ch2 = '\b', k++;
				break;

#else	//	EXPAND_ESCAPES_REAL_BACKSPACE

				if ( j ) {
					if ( ( temp[j - 1] != '\n' ) && ( temp[j - 1] != '\r' ) ) {
						j--;
						temp[j] = '\0';
					}
				}
				k++;
				break;

#endif	//	EXPAND_ESCAPES_REAL_BACKSPACE

			case 'c' :

				ch2 = ',', k++;
				break;

			case 'e' :

				ch2 = 0x1b, k++;
				break;

			case 'f' :

				ch2 = '\f', k++;
				break;

			case 'k' :

				ch2 = '`', k++;
				break;

			case 'n' :

				ch2 = '\n', k++;
				break;

			case 'p' :

				ch2 = '%', k++;
				break;

			case 'q' :

				ch2 = '"', k++;
				break;

			case 'r' :

				ch2 = '\r', k++;
				break;

			case 's' :

				ch2 = 0x20, k++;
				break;

			case 't' :

				ch2 = '\t', k++;
				break;

			case 'u' :

				if ( trv = ParseUnicodeEscape( ch + 1, 4, temp, buflen ) ) {
					ch += trv;
					j = wcslen( temp );
					k++;
				}
				else
					ch2 = *ch;

				break;

			case 'U' :

				if ( trv = ParseUnicodeEscape( ch + 1, 8, temp, buflen ) ) {
					ch += trv;
					j = wcslen( temp );
					k++;
				}
				else
					ch2 = *ch;

				break;

			case 'v' :

				ch2 = '\v', k++;
				break;

			case 'x' :

				if ( trv = ParseHexEscape( ch + 1, temp, buflen ) ) {
					ch += trv;
					j = wcslen( temp );
					k++;
				}
				else
					ch2 = *ch;

				break;

			case '#' :

				if ( trv = ParseDecimalEscape( ch + 1, temp, buflen ) ) {
					ch += trv;
					j = wcslen( temp );
					k++;
				}
				else
					ch2 = *ch;

				break;

			case '0' :
			case '1' :
			case '2' :
			case '3' :
			case '4' :
			case '5' :
			case '6' :
			case '7' :

				if ( trv = ParseOctalEscape( ch, temp, buflen ) ) {
					ch = ch + trv - 1;
					j = wcslen( temp );
					k++;
				}
				else
					ch2 = *ch;

				break;

			default :

				ch2 = *ch;
			}

			if ( ch2 && ( j < ( buflen - 1 ) ) ) {
				temp[j++]	= ch2;
				temp[j]		= '\0';
			}
		}
		else if ( j < ( buflen - 1 ) ) {
			temp[j++]	= *ch;
			temp[j]		= '\0';
		}

		if ( *ch == '\0' )
			break;
	}
	

	wcscpy_s( String, buflen, temp );
	free( temp );
	return k;
}


#endif	//	EXPAND_ESCAPES_CPP


/*

	1.0.3		2021-10-25

	1.0.4.0		2022-11-10
				Added an option for \b to expand to a regular ASCII backspace
				character, instead of deleting the previous character.  Define
				EXPAND_ESCAPES_REAL_BACKSPACE to enable this behavior.

	1.0.4.1		2022-11-16
				Minor tweak to macro definitions; no functional changes.

*/
