#ifndef FILE_HANDLER_VERSION
#define FILE_HANDLER_VERSION			1.0.15.0


#include <shlobj.h>


//			Functions to parse attribute and range options, and to enumerate files:



#define FH_TIME_STAMP_WRITE					0
#define FH_TIME_STAMP_LAST_ACCESS			1
#define FH_TIME_STAMP_CREATION				2

#define FH_DATE_FORMAT_MDY					0
#define FH_DATE_FORMAT_DMY					1
#define FH_DATE_FORMAT_YMD					2

#define FH_EPOCH							1601
#define FH_ULTIMA							9999

#define FH_WEEKDAYS_MASK					0x3e
#define FH_WEEKENDS_MASK					0x41

#define FH_FLAG_RECURSE						0x0001
#define FH_FLAG_NO_JUNCTIONS				0x0002
#define FH_FLAG_NO_HIDDEN_DIRS				0x0004
#define FH_FLAG_NO_SYSTEM_DIRS				0x0008
#define FH_FLAG_RECURSE_BOTTOM_UP			0x0010
#define FH_FLAG_SUPPRESS_FNF_ERROR			0x0100
#define FH_FLAG_CANONICALIZED				0x0200
#define FH_FLAG_INCLUDE_LISTS				0x0400
#define FH_FLAG_MINUS_IS_SPECIAL			0x0800
#define FH_FLAG_ALLOW_DOTS					0x1000
#define FH_FLAG_ALLOW_CLIPBOARD				0x2000
#define FH_FLAG_ALLOW_FILE_LISTS			0x4000
#define FH_FLAG_ALLOW_URLS					0x8000
#define FH_FLAG_EXCEPTIONS_DO_NOT_ABORT		0x800000

#define TRUENAME_FLAG_INCLUDE_LISTS			0x0001
#define TRUENAME_FLAG_HOME_PATHS			0x0002
#define TRUENAME_FLAG_REPARSE_POINTS		0x0004

#define FH_TAG_MOUNT_POINT					0xA0000003
#define FH_TAG_SYMLINK						0xA000000C

#define FH_USER_ABORT						-3
#define FH_OTHER_ABORT						-100
#define FH_EXCEPTION_RETURN					-200


#ifndef ERRHANDLE
#define ERRHANDLE GetStdHandle( STD_ERROR_HANDLE )
#endif // ERRHANDLE


#define FH_EXP( arg )						#arg
#define FH_STR( arg )						FH_EXP( arg )
#define FH_VER_STRING						FH_STR( FILE_HANDLER_VERSION )


#include "mmfiles.cpp"



struct _SizeRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	unsigned __int64 Smallest;
	unsigned __int64 Largest;
};

struct _TimeRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	unsigned short int WhichStamp;
	unsigned int EarliestTime;
	unsigned int LatestTime;
};

struct _DateRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	unsigned short int WhichStamp;
	unsigned int EarliestJulian;
	unsigned int LatestJulian;
	unsigned int EarliestTime;
	unsigned int LatestTime;
};

struct _ExclusionRange {
	void *Next;
	wchar_t Type;
	wchar_t Filespec;
};

struct _DOWRange {
	void *Next;
	wchar_t Type;
	unsigned short int WhichStamp;
	unsigned short int DOWMask;
};

struct _DescriptionRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	wchar_t Description;
};

struct _OwnerRange {
	void *Next;
	wchar_t Type;
	BOOL Reverse;
	wchar_t Owner;
};


const unsigned int _FHMonthLength[13][2] = { {NULL, NULL}, {31, 31}, {28, 29}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31} };
const unsigned int _FHKalends[13][2] = { {NULL, NULL}, {0, 0}, {31, 31}, {59, 60}, {90, 91}, {120, 121}, {151, 152}, {181, 182}, {212, 213}, {243, 244}, {273, 274}, {304, 305}, {334, 335} };
const wchar_t _FHDowName[7][4] = { L"Sun", L"Mon", L"Tue", L"Wed", L"Thu", L"Fri", L"Sat" };



struct FH_FilenameItem {
	FH_FilenameItem *Next;
	int Flags;
	wchar_t Filename;
};

struct FH_FilenamesLinkedList {
	FH_FilenameItem *First;
	FH_FilenameItem *Last;
	unsigned int Count;
};



int FH_AddFileToLinkedList( LPCTSTR Filename, int Flags, void *FilesList )
{
	if ( !Filename || !FilesList )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	//	wprintf( L"   * Add file '%s' to linked list:  ", Filename );

	FH_FilenamesLinkedList *List	= (FH_FilenamesLinkedList*) FilesList;
	
	
	FH_FilenameItem	*New = (FH_FilenameItem*) malloc( sizeof( FH_FilenameItem ) + ( wcslen( Filename ) * sizeof( wchar_t ) ) );
	if ( !New )
		return ERROR_NOT_ENOUGH_MEMORY;

	New->Next		= NULL;
	New->Flags		= Flags;
	wcscpy_s( &(New->Filename), wcslen( Filename ) + 1, Filename );

	if ( List->First ) {
		(List->Last)->Next	= New;
		List->Last			= New;
		//	wprintf( L"Added to end; " );
	}
	else {
		List->First			= New;
		List->Last			= New;
		List->Count			= 0;
		//	wprintf( L"First item; " );
	}

	List->Count += 1;
	//	wprintf( L" %u items.\n", List->Count );
	return 0;
}



//	Deletes all items in a linked list, and zeros the first-item pointer:

int FH_FreeLinkedList( void **FirstItem )
{
	if ( !FirstItem )
		return -1;

	size_t *This = (size_t*) *FirstItem, Next = NULL;

	while ( This ) {
		//	Printf( L"   * Freeing memory at 0x%08x\r\n", This );
		Next = *This;
		free( This );
		This = (size_t*) Next;
	}

	*FirstItem = NULL;
	return 0;
}



int FH_FreeLinkedList( FH_FilenameItem **FirstItem )
{
	return FH_FreeLinkedList( (void**) FirstItem );
}



int _FH_ExpandVars( LPTSTR Buffer )
{
	if ( !Buffer )
		return -1;


	HANDLE OldStdErr	= NULL;
	HANDLE DevNull		= NULL;
	int rv				= 0;


	OldStdErr	= GetStdHandle( STD_ERROR_HANDLE );
	DevNull		= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL );
	if ( DevNull != INVALID_HANDLE_VALUE )
		SetStdHandle( STD_ERROR_HANDLE, DevNull );

	rv = ExpandVariables( Buffer, 1 );

	if ( DevNull != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStdErr );
		CloseHandle( DevNull );
	}

	return rv;
}



BOOL FH_IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



int FH_StripQuotes( LPTSTR String )
{
	if ( !String )
		return -1;

	int n = 0;

	size_t j = 0;
	for ( size_t i = 0; ; i++ ) {
		if ( String[i] == '"' )
			n++;
		else {
			String[j++] = String[i];
			if ( String[i] == '\0' )
				break;
		}
	}

	return n;
}



int FH_TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( iswspace( ch ) == 0 ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int FH_AppendFilename( LPTSTR Path, LPCTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



int FH_AssembleFilename( LPTSTR outBuffer, LPCTSTR inPath, LPCTSTR inFilename, size_t BufferSize )
{
	if ( !outBuffer || !inFilename )
		return -1;


	if ( inPath )
		wcscpy_s( outBuffer, BufferSize, inPath );
	else {
		wcscpy_s( outBuffer, BufferSize, L"%_cwds" );
		ExpandVariables( outBuffer, 1 );
	}

	FH_AppendFilename( outBuffer, inFilename, BufferSize );
	return 0;
}



int FH_GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



wchar_t FH_LastChar( LPCTSTR inString )
{
	if ( !inString )
		return '\0';
	else if ( inString[0] == '\0' )
		return '\0';


	return inString[wcslen( inString ) - 1];
}



struct _FH_ReparseDataBuffer {
	ULONG  ReparseTag;
	USHORT ReparseDataLength;
	USHORT Reserved;
	union {
		struct {
			USHORT SubstituteNameOffset;
			USHORT SubstituteNameLength;
			USHORT PrintNameOffset;
			USHORT PrintNameLength;
			ULONG  Flags;
			WCHAR  PathBuffer[1];
		} SymbolicLinkReparseBuffer;
		struct {
			USHORT SubstituteNameOffset;
			USHORT SubstituteNameLength;
			USHORT PrintNameOffset;
			USHORT PrintNameLength;
			WCHAR  PathBuffer[1];
		} MountPointReparseBuffer;
		struct {
			UCHAR DataBuffer[1];
		} GenericReparseBuffer;
	};
};



int _FH_ParseInt( LPCTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;


	unsigned int value		= 0;
	wchar_t *sep1			= NULL;


	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return errno;
		else if ( !EndChar && *sep1 )
			return 1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return errno;
		else if ( !EndChar && *sep1 )
			return 1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return 1;
}



int ParseAttributesOption( LPTSTR inString, BOOL AllowDirectories, DWORD *AttribMask, DWORD *AttribValues )
{
	if ( !inString )
		return -1;
	else if ( inString[0] == '\0' )
		return -1;

	int j = 0;
	if ( inString[j] == '/' )
		j++;

	if ( iswalpha( inString[j] ) )
		j++;
	
	if ( FH_IsColonOrEquals( inString[j] ) )
		j++;


	*AttribMask = AllowDirectories ? 0 : FILE_ATTRIBUTE_DIRECTORY;
	*AttribValues = 0;


	if ( inString[j] == '\0' )
		return 0;


	BOOL Minus = FALSE;
	const wchar_t ErrorString[] = L"Bad attributes: \"%s\"\r\n";


	while ( inString[j] ) {

		switch ( towupper( inString[j] ) ) {

			case 'A' :
				*AttribMask |= FILE_ATTRIBUTE_ARCHIVE;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_ARCHIVE;
				else
					*AttribValues |= FILE_ATTRIBUTE_ARCHIVE;

				Minus = FALSE;
				break;

			case 'C' :
				*AttribMask |= FILE_ATTRIBUTE_COMPRESSED;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_COMPRESSED;
				else
					*AttribValues |= FILE_ATTRIBUTE_COMPRESSED;

				Minus = FALSE;
				break;

			case 'D' :
				if ( !AllowDirectories ) {
					DisplayErrorHeader();
					_wcsupr_s( inString, MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Directories not allowed: \"%s\"\r\n", inString );
					return 1;
				}

				*AttribMask |= FILE_ATTRIBUTE_DIRECTORY;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_DIRECTORY;
				else
					*AttribValues |= FILE_ATTRIBUTE_DIRECTORY;

				Minus = FALSE;
				break;

			case 'E' :
				*AttribMask |= FILE_ATTRIBUTE_ENCRYPTED;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_ENCRYPTED;
				else
					*AttribValues |= FILE_ATTRIBUTE_ENCRYPTED;

				Minus = FALSE;
				break;

			case 'H' :
				*AttribMask |= FILE_ATTRIBUTE_HIDDEN;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_HIDDEN;
				else
					*AttribValues |= FILE_ATTRIBUTE_HIDDEN;

				Minus = FALSE;
				break;

			case 'I' :
				*AttribMask |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;
				else
					*AttribValues |= FILE_ATTRIBUTE_NOT_CONTENT_INDEXED;

				Minus = FALSE;
				break;

			case 'O' :
				*AttribMask |= FILE_ATTRIBUTE_OFFLINE;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_OFFLINE;
				else
					*AttribValues |= FILE_ATTRIBUTE_OFFLINE;

				Minus = FALSE;
				break;

			case 'R' :
				*AttribMask |= FILE_ATTRIBUTE_READONLY;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_READONLY;
				else
					*AttribValues |= FILE_ATTRIBUTE_READONLY;

				Minus = FALSE;
				break;

			case 'S' :
				*AttribMask |= FILE_ATTRIBUTE_SYSTEM;

				if ( Minus )
					*AttribValues &= ~FILE_ATTRIBUTE_SYSTEM;
				else
					*AttribValues |= FILE_ATTRIBUTE_SYSTEM;

				Minus = FALSE;
				break;

			case '-' :
				if ( iswalpha( inString[j + 1] ) == 0 ) {
					DisplayErrorHeader();
					_wcsupr_s( inString, MAX_ARG_LEN );
					Qprintf( ERRHANDLE, ErrorString, inString );
					return 1;
				}

				Minus = TRUE;
				break;

			default :
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( ERRHANDLE, ErrorString, inString );
				return 1;
		}

		j++;
	}

	return 0;
}



int ParseNOption( LPTSTR inString, int *Recurse, BOOL *NoColor )
{
	if ( !inString )
		return -1;
	else if ( inString[0] == '\0' )
		return -1;

	int j = 0;
	wchar_t optchar	= 'N';


	if ( ( inString[j] == '/' ) || ( inString[j] == '-' ) )
		j++;

	if ( iswalpha( inString[j] ) ) {
		optchar = towupper( inString[j] );
		j++;
	}

	if ( FH_IsColonOrEquals( inString[j] ) )
		j++;

	while ( inString[j] ) {
		switch ( towupper( inString[j] ) ) {
		case 'D' :
			*Recurse |= FH_FLAG_NO_HIDDEN_DIRS;
			break;

		case 'Z' :
			*Recurse |= FH_FLAG_NO_SYSTEM_DIRS;
			break;

		case 'J' :
			*Recurse |= FH_FLAG_NO_JUNCTIONS;
			break;

		case 'F' :
			*Recurse |= FH_FLAG_SUPPRESS_FNF_ERROR;
			break;

		case 'C' :
			if ( NoColor ) {
				*NoColor = TRUE;
				break;
			}
			//  else fall through to ....

		default :
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown suboption: /%c:%c\r\n", optchar, towupper( inString[j] ) );
			return 1;
		}

		j++;
	}

	return 0;
}



int ParseSOption( LPTSTR inString, int *Recurse, unsigned int *MinDepth, unsigned int *MaxDepth )
{
	if ( !inString  || !Recurse || !MinDepth || !MaxDepth )
		return -1;


	wchar_t *ch				= NULL;
	unsigned int temp		= 0;


	if ( _wcsnicmp( inString, L"/S", 2 ) )
		return 0;

	*Recurse |= FH_FLAG_RECURSE;

	if ( !_wcsicmp( inString, L"/S" ) ) {
		*MinDepth = 0;
		*MaxDepth = UINT_MAX;
		return 0;
	}

	ch = inString + 2;
	if ( FH_IsColonOrEquals( *ch ) )
		ch++;

	if ( iswdigit( *ch ) ) {
		if ( _FH_ParseInt( ch, &temp, &ch ) ) {
			_wcsupr_s( inString, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad MaxDepth value: \"%s\"\r\n", inString );
			return 1;
		}

		*MaxDepth = temp;
	}

	if ( ( *ch == '+' ) && ( iswdigit( ch[1] ) ) ) {
		if ( _FH_ParseInt( ch + 1, &temp, NULL ) ) {
			_wcsupr_s( inString, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad MinDepth value: \"%s\"\r\n", inString );
			return 1;
		}

		*MinDepth = temp;
		return 0;
	}

	if ( *ch ) {
		_wcsupr_s( inString, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", inString );
		return 1;
	}

	return 0;
}



size_t _FHTrimString( LPTSTR String )
{
	if ( !String )
		return -1;

	size_t j = 0, terminus = 0;
	wchar_t ch = '\0';

	for ( size_t i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( ch == '"' )
			continue;

		if ( iswspace( ch ) ) {
			if ( j == 0 )
				continue;
		}
		else
			terminus = j + 1;
		
		String[j++] = ch;
	}

	String[terminus] = '\0';

	return terminus;
}



int ParseSize2( wchar_t **ch, unsigned __int64 *Value )
{
	if ( !ch || !Value )
		return -1;
	else if ( !iswdigit( **ch ) )
		return 1;

	*Value = 0;

	while ( iswdigit( **ch ) ) {
		*Value = ( *Value * 10 ) + ( **ch - '0' );
		(*ch)++;
	}

	switch ( **ch ) {
		case 'k' :
			*Value *= 1000LL;
			(*ch)++;
			break;

		case 'm' :
			*Value *= 1000000LL;
			(*ch)++;
			break;

		case 'g' :
			*Value *= 1000000000LL;
			(*ch)++;
			break;

		case 't' :
			*Value *= 1000000000000LL;
			(*ch)++;
			break;

		case 'p' :
			*Value *= 1000000000000000LL;
			(*ch)++;
			break;

		case 'K' :
			*Value = *Value << 10;
			(*ch)++;
			break;

		case 'M' :
			*Value = *Value << 20;
			(*ch)++;
			break;

		case 'G' :
			*Value = *Value << 30;
			(*ch)++;
			break;

		case 'T' :
			*Value = *Value << 40;
			(*ch)++;
			break;

		case 'P' :
			*Value = *Value << 50;
			(*ch)++;
			break;
	}

	return 0;
}



int ParseSizeRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	wchar_t *sep1 = inString;
	unsigned __int64 Smallest = 0, Largest = ULLONG_MAX;
	BOOL Add = FALSE;

	if ( FH_IsColonOrEquals( *sep1 ) )
		sep1++;

	if ( wcscmp( sep1, L"]" ) == 0 )
		return 0;

	if ( iswdigit( *sep1 ) )
		ParseSize2( &sep1, &Smallest );

	if ( *sep1 == ',' )
		sep1++;

	if ( ( *sep1 == '+' ) && iswdigit( *(sep1 + 1) ) ) {
		Add = TRUE;
		sep1++;
	}

	if ( iswdigit( *sep1 ) )
		ParseSize2( &sep1, &Largest );

	if ( Add )
		Largest += Smallest;

	if ( Smallest > Largest ) {
		__int64 t = Smallest;
		Smallest = Largest;
		Largest = t;
	}

	if ( wcscmp( sep1, L"]" ) ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( GetStdHandle( STD_ERROR_HANDLE), L"Bad size range: \"%s\"\r\n", inString );
		return 1;
	}

	// Printf( L"   * Size range = %q - %q %s\r\n", Smallest, Largest, Reverse ? L"(reversed)" : L"" );

	_SizeRange *SizeRange = (_SizeRange*) malloc( sizeof( _SizeRange ) );

	if ( !SizeRange ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t allocate size range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", SizeRange );

	SizeRange->Next		= *outPointer;
	SizeRange->Type		= 'S';
	SizeRange->Reverse	= Reverse;
	SizeRange->Smallest	= Smallest;
	SizeRange->Largest	= Largest;

	*outPointer = SizeRange;
	return 0;
}



int IsTimeSep( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '.' ) );
}



int ParseTime2( wchar_t **ch, unsigned int *Value, BOOL *HasSeconds )
{
	if ( !ch || !Value )
		return -1;
	else if ( !iswdigit( **ch ) )
		return 1;

	unsigned int Hour = 0, Minute = 0, Second = 0;

	Hour = wcstoul( *ch, ch, 10 );

	if ( Hour > 23 )
		return 1;

	if ( !IsTimeSep( **ch ) )
		return 1;

	(*ch)++;

	if ( iswdigit( **ch ) == 0 )
		return 1;

	Minute = wcstoul( *ch, ch, 10 );

	if ( Minute > 59 )
		return 1;

	if ( IsTimeSep( **ch ) ) {
		(*ch)++;

		if ( iswdigit( **ch ) == 0 )
			return 1;

		Second = wcstoul( *ch, ch, 10 );

		if ( Second > 59 )
			return 1;

		if ( HasSeconds != NULL )
			*HasSeconds = TRUE;
	}
	else if ( HasSeconds != NULL )
		*HasSeconds = FALSE;

	if ( ( towupper( **ch ) == 'A' ) || ( towupper( **ch ) == 'P' ) ) {
		if ( ( Hour == 0 ) || ( Hour > 12 ) )
			return 1;

		if ( towupper( **ch ) == 'A' )
			Hour %= 12;
		else
			Hour = ( Hour % 12 ) + 12;

		(*ch)++;
		if ( towupper( **ch ) == 'M' )
			(*ch)++;
	}

	*Value = Second + ( 60 * Minute ) + ( 60 * 60 * Hour );
	return 0;
}



int ParseTimeRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t *sep1 = inString;
	unsigned int EarliestTime = 0, LatestTime = 0, Offset = 0;
	unsigned short int WhichStamp = FH_TIME_STAMP_WRITE;
	BOOL EarliestHasSeconds = FALSE, LatestHasSeconds = FALSE, Add = FALSE;

	SYSTEMTIME Now;
	GetLocalTime( &Now );


	if ( towupper( *sep1 ) == 'W' ) {
		WhichStamp = FH_TIME_STAMP_WRITE;
		sep1++;
	}
	else if ( towupper( *sep1 ) == 'A' ) {
		WhichStamp = FH_TIME_STAMP_LAST_ACCESS;
		sep1++;
	}
	else if ( towupper( *sep1 ) == 'C' ) {
		WhichStamp = FH_TIME_STAMP_CREATION;
		sep1++;
	}

	if ( FH_IsColonOrEquals( *sep1 ) )
		sep1++;

	if ( iswdigit( *sep1 ) ) {
		if ( ParseTime2( &sep1, &EarliestTime, &EarliestHasSeconds ) ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad time range: \"%s\"\r\n", inString );
			return 1;
		}
	}
	else if ( ( ( *sep1 == '+' ) || ( *sep1 == '-' ) ) && iswdigit( *(sep1 + 1) ) ) {
		Add = ( *sep1 == '+' );
		Offset = wcstoul( sep1 + 1, &sep1, 10 );
		EarliestTime = ( 60 * Now.wMinute ) + ( 60 * 60 * Now.wHour );

		if ( Add ) {
			EarliestTime += ( 60 * Offset );
			if ( EarliestTime >= ( 24 * 60 * 60 ) )
				EarliestTime %= ( 24 * 60 * 60 );
		}
		else {
			if ( EarliestTime > Offset )
				EarliestTime -= Offset;
			else
				EarliestTime = 0;
		}
	}
	else if ( ( *sep1 == ',' ) || ( *sep1 == ']' ) )
		EarliestTime = ( 60 * Now.wMinute ) + ( 60 * 60 * Now.wHour );

	if ( *sep1 == ',' )
		sep1++;

	if ( iswdigit( *sep1 ) ) {
		if ( ParseTime2( &sep1, &LatestTime, &LatestHasSeconds ) ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad time range: \"%s\"\r\n", inString );
			return 1;
		}
	}
	else if ( ( ( *sep1 == '+' ) || ( *sep1 == '-' ) ) && iswdigit( *(sep1 + 1) ) ) {
		Add = ( *sep1 == '+' );
		Offset = wcstoul( sep1 + 1, &sep1, 10 );

		if ( Add ) {
			LatestTime = EarliestTime + ( 60 * Offset );
			if ( LatestTime >= ( 24 * 60 * 60 ) )
				LatestTime %= ( 24 * 60 * 60 );
		}
		else {
			if ( EarliestTime > Offset )
				LatestTime = EarliestTime - Offset;
			else
				LatestTime = 0;
		}
	}
	else if ( *sep1 == ']' )
		LatestTime = ( 60 * Now.wMinute ) + ( 60 * 60 * Now.wHour );

	if ( EarliestTime > LatestTime ) {
		unsigned int t = EarliestTime;
		EarliestTime = LatestTime;
		LatestTime = t;
		BOOL s = EarliestHasSeconds;
		EarliestHasSeconds = LatestHasSeconds;
		LatestHasSeconds = s;
	}

	if ( !LatestHasSeconds )
		LatestTime += 59;

	if ( wcscmp( sep1, L"]" ) ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Bad time range: \"%s\"\r\n", inString );
		return 1;
	}

	// Printf( L"   * Time range = %u - %u %s\r\n", EarliestTime, LatestTime, Reverse ? L"(reversed)" : L"" );

	_TimeRange *TimeRange = (_TimeRange*) malloc( sizeof( _TimeRange ) );

	if ( !TimeRange ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t allocate time range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", TimeRange );

	TimeRange->Next			= *outPointer;
	TimeRange->Type			= 'T';
	TimeRange->Reverse		= Reverse;
	TimeRange->WhichStamp	= WhichStamp;
	TimeRange->EarliestTime	= EarliestTime;
	TimeRange->LatestTime	= LatestTime;

	*outPointer = TimeRange;
	return 0;
}



BOOL _FHLeapness( int Year )
{
	if ( ( Year % 400 ) == 0 ) return TRUE;
	if ( ( Year % 100 ) == 0 ) return FALSE;
	return ( ( Year % 4 ) == 0 );
}



unsigned int _FHJulian( unsigned int Year, unsigned int Month, unsigned int Day )
{
	return ( ( Year - FH_EPOCH ) * 365 ) + ( ( Year - FH_EPOCH ) / 4 ) - ( ( Year - FH_EPOCH ) / 100 ) + ( ( Year - FH_EPOCH ) / 400 ) + _FHKalends[Month][_FHLeapness( Year )] + Day;
}



int _FHIsDateSep( wchar_t ch )
{
	return ( ( ch == '/' ) || ( ch == '-' ) || ( ch == '.' ) );
}



unsigned int GetLocalDateFormat()
{
	wchar_t temp[SHORT_BUF_LEN] = L"";
	unsigned int n = 0xffff;

	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_IDATE, temp, SHORT_BUF_LEN ) == 0 )
		return FH_DATE_FORMAT_MDY;

	n = wcstoul( temp, NULL, 10 );

	if ( n < 3 )
		return n;
	else
		return FH_DATE_FORMAT_MDY;
}



int ParseDate2( wchar_t **ch, unsigned int *Value )
{
	if ( ( ch == NULL ) || ( Value == NULL ) )
		return -1;
	else if ( iswdigit( **ch ) == 0 )
		return 1;

	unsigned int Value1 = 0, Value2 = 0, Value3 = 0;
	unsigned int Year = 0, Month = 0, Day = 0;
	unsigned int DateFormat = FH_DATE_FORMAT_MDY;

	Value1 = wcstoul( *ch, ch, 10 );

	if ( !_FHIsDateSep( **ch ) )
		return 1;

	(*ch)++;

	if ( iswdigit( **ch ) == 0 )
		return 1;

	Value2 = wcstoul( *ch, ch, 10 );

	if ( !_FHIsDateSep( **ch ) )
		return 1;

	(*ch)++;

	if ( iswdigit( **ch ) == 0 )
		return 1;

	Value3 = wcstoul( *ch, ch, 10 );

	if ( ( Value1 >= FH_EPOCH ) && ( Value1 <= FH_ULTIMA ) )
		DateFormat = FH_DATE_FORMAT_YMD;
	else
		DateFormat = GetLocalDateFormat();

	switch ( DateFormat ) {
		case FH_DATE_FORMAT_MDY :
			Year	= Value3;
			Month	= Value1;
			Day		= Value2;
			break;

		case FH_DATE_FORMAT_DMY :
			Year	= Value3;
			Month	= Value2;
			Day		= Value1;
			break;

		case FH_DATE_FORMAT_YMD :
			Year	= Value1;
			Month	= Value2;
			Day		= Value3;
			break;
	}

	if ( ( Year >= 80 ) && ( Year <= 99 ) )
		Year += 1900;
	else if ( Year <= 99 )
		Year += 2000;

	if ( ( Year < FH_EPOCH ) || ( Year > FH_ULTIMA ) )
		return 1;
	else if ( ( Month < 1 ) || ( Month > 12 ) )
		return 1;
	else if ( ( Day < 1 ) || ( Day > _FHMonthLength[Month][_FHLeapness( Year )] ) )
		return 1;

	*Value = _FHJulian( Year, Month, Day );
	return 0;
}



int ParseDateRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t *sep1 = inString;
	unsigned int EarliestJulian = 0, LatestJulian = 0, Offset = 0;
	unsigned int EarliestTime = UINT_MAX, LatestTime = UINT_MAX;
	unsigned short int WhichStamp = FH_TIME_STAMP_WRITE;
	BOOL EarliestHasSeconds = FALSE, LatestHasSeconds = FALSE, Add = FALSE;

	SYSTEMTIME Now;
	GetLocalTime( &Now );


	if ( towupper( *sep1 ) == 'W' ) {
		WhichStamp = FH_TIME_STAMP_WRITE;
		sep1++;
	}
	else if ( towupper( *sep1 ) == 'A' ) {
		WhichStamp = FH_TIME_STAMP_LAST_ACCESS;
		sep1++;
	}
	else if ( towupper( *sep1 ) == 'C' ) {
		WhichStamp = FH_TIME_STAMP_CREATION;
		sep1++;
	}

	if ( FH_IsColonOrEquals( *sep1 ) )
		sep1++;

	if ( iswdigit( *sep1 ) ) {
		if ( ParseDate2( &sep1, &EarliestJulian ) ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad date1 in range: \"%s\"\r\n", inString );
			return 1;
		}

		if ( *sep1 == '@' ) {
			sep1++;

			if ( ParseTime2( &sep1, &EarliestTime, &EarliestHasSeconds ) ) {
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Bad date1 (time) in range: \"%s\"\r\n", inString );
				return 1;
			}
		}
	}
	else if ( ( ( *sep1 == '+' ) || ( *sep1 == '-' ) ) && iswdigit( *(sep1 + 1) ) ) {
		Add = ( *sep1 == '+' );
		Offset = wcstoul( sep1 + 1, &sep1, 10 );
		EarliestJulian = _FHJulian( Now.wYear, Now.wMonth, Now.wDay );

		if ( Add )
			EarliestJulian += Offset;
		else {
			if ( EarliestJulian > Offset )
				EarliestJulian -= Offset;
			else
				EarliestJulian = 0;
		}

		if ( *sep1 == '@' ) {
			sep1++;

			if ( ParseTime2( &sep1, &EarliestTime, &EarliestHasSeconds ) ) {
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Bad date1 (time) in range: \"%s\"\r\n", inString );
				return 1;
			}
		}
	}
	else if ( ( *sep1 == ',' ) || ( *sep1 == ']' ) )
		EarliestJulian = _FHJulian( Now.wYear, Now.wMonth, Now.wDay );

	if ( *sep1 == ',' )
		sep1++;

	if ( iswdigit( *sep1 ) ) {
		if ( ParseDate2( &sep1, &LatestJulian ) ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad date2 in range: \"%s\"\r\n", inString );
			return 1;
		}

		if ( *sep1 == '@' ) {
			sep1++;

			if ( ParseTime2( &sep1, &LatestTime, &LatestHasSeconds ) ) {
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Bad date2 (time) in range: \"%s\"\r\n", inString );
				return 1;
			}
		}
	}
	else if ( ( ( *sep1 == '+' ) || ( *sep1 == '-' ) ) && iswdigit( *(sep1 + 1) ) ) {
		Add = ( *sep1 == '+' );
		Offset = wcstoul( sep1 + 1, &sep1, 10 );

		if ( Add )
			LatestJulian = EarliestJulian + Offset;
		else {
			if ( EarliestJulian > Offset )
				LatestJulian = EarliestJulian - Offset;
			else
				LatestJulian = 0;
		}

		if ( *sep1 == '@' ) {
			sep1++;

			if ( ParseTime2( &sep1, &LatestTime, &LatestHasSeconds ) ) {
				DisplayErrorHeader();
				_wcsupr_s( inString, MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Bad date2 (time) in range: \"%s\"\r\n", inString );
				return 1;
			}
		}
	}
	else if ( *sep1 == ']' )
		LatestJulian = _FHJulian( Now.wYear, Now.wMonth, Now.wDay );

	if ( ( EarliestJulian > LatestJulian ) || ( ( EarliestJulian == LatestJulian ) && ( EarliestTime > LatestTime ) ) ) {
		unsigned int t = EarliestJulian;
		EarliestJulian = LatestJulian;
		LatestJulian = t;
		BOOL s = EarliestHasSeconds;
		EarliestHasSeconds = LatestHasSeconds;
		LatestHasSeconds = s;
	}

	if ( EarliestTime == UINT_MAX )
		EarliestTime = 0;

	if ( ( LatestTime != UINT_MAX ) && !LatestHasSeconds )
		LatestTime += 59;

	if ( wcscmp( sep1, L"]" ) ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Bad date range: \"%s\"\r\n", inString );
		return 1;
	}

	// Printf( L"   * Date range = %u - %u %s\r\n", EarliestJulian, LatestJulian, Reverse ? L"(reversed)" : L"" );

	_DateRange *DateRange = (_DateRange*) malloc( sizeof( _DateRange ) );

	if ( DateRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t allocate date range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", DateRange );

	DateRange->Next				= *outPointer;
	DateRange->Type				= 'D';
	DateRange->Reverse			= Reverse;
	DateRange->WhichStamp		= WhichStamp;
	DateRange->EarliestJulian	= EarliestJulian;
	DateRange->LatestJulian		= LatestJulian;
	DateRange->EarliestTime		= EarliestTime;
	DateRange->LatestTime		= LatestTime;

	*outPointer = DateRange;
	return 0;
}



int AddExclusionRange( LPCTSTR Filespec, void **outPointer )
{
	_ExclusionRange *ExclusionRange = (_ExclusionRange*) malloc( sizeof( _ExclusionRange ) + ( wcslen( Filespec ) * sizeof( wchar_t ) ) );

	if ( ExclusionRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t allocate exclusion range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", DateRange );

	ExclusionRange->Next		= *outPointer;
	ExclusionRange->Type		= '!';
	wcscpy_s( &(ExclusionRange->Filespec), wcslen( Filespec ) + 1, Filespec );

	*outPointer = ExclusionRange;
	return 0;
}



int ParseExclusionRange( LPTSTR inString, void **outPointer )
{
	if ( wcscmp( inString, L"]" ) == 0 )
		return 0;


	wchar_t Filespec[MAX_ARG_LEN] = L"";
	int j = 0;
	wchar_t ch = '\0';
	BOOL quote = FALSE;

	for ( int i = 0; inString[i]; i++ ) {
		ch = inString[i];

		if ( iswspace( ch ) && ( j == 0 ) )
			continue;

		if ( ch == '"' ) {
			quote = !quote;
			continue;
		}

		if ( !quote && iswspace( ch ) ) {
			if ( _FHTrimString( Filespec ) > 0 )
				AddExclusionRange( Filespec, outPointer );

			j = 0;
			Filespec[j] = '\0';
		}
		
		if ( ( ch == ']' ) && ( inString[i + 1] == '\0' ) ) {
			if ( _FHTrimString( Filespec ) > 0 ) {
				AddExclusionRange( Filespec, outPointer );
				return 0;
			}
		}

		if ( j < ( MAX_ARG_LEN - 1 ) ) {
			Filespec[j++] = ch;
			Filespec[j] = '\0';
		}
	}

	return 0;
}



int ParseDOW( LPTSTR inString )
{
	if ( inString == NULL )
		return -1;

	int num = -1;
	wchar_t *sep1 = '\0';

	if ( iswdigit( inString[0] ) ) {
		num = wcstol( inString, &sep1, 10 );
		if ( *sep1 != '\0' )
			return -1;

		if ( ( num >= 0 ) && ( num <= 7 ) )
			return num % 7;
		else
			return -1;
	}

	for ( int i = 0; i < 7; i++ )
		if ( _wcsnicmp( inString, _FHDowName[i], 2 ) == 0 )
			return i;

	return -1;
}



int ParseDOW2( LPTSTR inString, int *len )
{
	if ( inString == NULL )
		return -1;

	wchar_t shortbuf[SHORT_BUF_LEN] = L"";
	int i = 0;
	wchar_t ch = '\0';

	do {
		ch = inString[i];
		if ( iswspace( ch ) || ( ch == ';' ) || ( ch == '-' ) || ( ch == '.' ) || ( ch == ',' ) || ( ch == ']' ) || ( i >= ( SHORT_BUF_LEN - 1 ) ) )
			ch = '\0';

		shortbuf[i++] = ch;

	} while ( ch );

	if ( len )
		*len = i - 1;

	if ( ( _wcsicmp( shortbuf, L"weekday" ) == 0 ) || ( _wcsicmp( shortbuf, L"weekdays" ) == 0 ) )
		return FH_WEEKDAYS_MASK;
	else if ( ( _wcsicmp( shortbuf, L"weekend" ) == 0 ) || ( _wcsicmp( shortbuf, L"weekends" ) == 0 ) )
		return FH_WEEKENDS_MASK;
	else
		return ParseDOW( shortbuf );
}



int ParseDOWRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t *ch = inString;
	int i = 0;
	unsigned short int DOWMask = 0;
	unsigned short int WhichStamp = FH_TIME_STAMP_WRITE;
	int dow = 0, dow2 = 0;
	int len = 0;

	const wchar_t ErrorStr[] = L"Bad day-of-the-week range: \"%s\"\r\n";

	
	if ( towupper( inString[i] ) == 'W' ) {
		WhichStamp = FH_TIME_STAMP_WRITE;
		i++;
	}
	else if ( towupper( inString[i] ) == 'A' ) {
		WhichStamp = FH_TIME_STAMP_LAST_ACCESS;
		i++;
	}
	else if ( towupper( inString[i] ) == 'C' ) {
		WhichStamp = FH_TIME_STAMP_CREATION;
		i++;
	}

	if ( FH_IsColonOrEquals( inString[i] ) )
		i++;
	else {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, ErrorStr, inString );
		return 1;
	}


	while ( inString[i] ) {
		dow = ParseDOW2( inString + i, &len );
		if ( dow < 0 ) {
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( ERRHANDLE, ErrorStr, inString );
			return 1;
		}
		else
			i += len;

		dow2 = -1;
		if ( ( i < ( MAX_ARG_LEN - 4 ) ) && ( dow <= 6 ) ) {
			if ( ( inString[i] == '-' ) || ( ( inString[i] == '.' ) && ( inString[i + 1] == '.' ) ) ) {
				// parse dow range:

				i += ( inString[i] == '.' ) ? 2 : 1;

				dow2 = ParseDOW2( inString + i, &len );
				if ( ( dow2 < 0 ) || ( dow2 > 6 ) ) {
					DisplayErrorHeader();
					_wcsupr_s( inString, MAX_ARG_LEN );
					Qprintf( ERRHANDLE, ErrorStr, inString );
					return 1;
				}
				else
					i += len;

				if ( ( inString[i] == '-' ) || ( inString[i] == '.' ) ) {
					DisplayErrorHeader();
					_wcsupr_s( inString, MAX_ARG_LEN );
					Qprintf( ERRHANDLE, ErrorStr, inString );
					return 1;
				}
			}
		}

		while ( iswspace( inString[i] ) || ( inString[i] == ',' ) || ( inString[i] == ';' ) )
			i++;

		if ( dow <= 6 )
			DOWMask |= ( 1 << dow );
		else
			DOWMask |= dow;

		while ( ( dow2 >= 0 ) && ( dow != dow2 ) ) {
			dow = ( ( dow + 1 ) % 7 );
			DOWMask |= ( 1 << dow );
		}

		if ( wcscmp( inString + i, L"]" ) == 0 )
			break;
	}

	if ( Reverse )
		DOWMask ^= 0x7f;

	if ( DOWMask == 0 ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Empty day-of-the-week range: \"%s\"\r\n", inString );
		return 1;
	}


	// Printf( L"   * DOW mask = %u\r\n", DOWMask );

	_DOWRange *DOWRange = (_DOWRange*) malloc( sizeof( _DOWRange ) );

	if ( DOWRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t allocate DOW range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", DOWRange );

	DOWRange->Next			= *outPointer;
	DOWRange->Type			= 'W';
	DOWRange->WhichStamp	= WhichStamp;
	DOWRange->DOWMask		= DOWMask;

	*outPointer = DOWRange;
	return 0;
}



int ParseDescriptionRange( LPTSTR inString, BOOL Reverse, BOOL SlashI, void **outPointer )
{
	if ( !inString )
		return -1;
	else if ( inString[0] == '\0' )
		return 0;
	else if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t Description[MAX_ARG_LEN] = L"";
	int i = 0, len = 0;
	
	if ( FH_IsColonOrEquals( inString[i] ) )
		i++;

	wcscpy_s( Description, MAX_ARG_LEN, inString + i );

	if ( !SlashI ) {
		len = (int) wcslen( Description );

		if ( len )
			if ( Description[len - 1] == ']' )
				Description[len - 1] = '\0';
	}

	// Printf( L"   * Description range = '%s' %s\r\n", Description, Reverse ? L"(reversed)" : L"" );

	_DescriptionRange *DescriptionRange = (_DescriptionRange*) malloc( sizeof( _DescriptionRange ) + ( wcslen( Description ) * sizeof( wchar_t ) ) );

	if ( DescriptionRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t allocate description range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", DescriptionRange );

	DescriptionRange->Next			= *outPointer;
	DescriptionRange->Type			= 'I';
	DescriptionRange->Reverse		= Reverse;
	wcscpy_s( &(DescriptionRange->Description), wcslen( Description ) + 1, Description );

	*outPointer = DescriptionRange;
	return 0;
}



int ParseOwnerRange( LPTSTR inString, BOOL Reverse, void **outPointer )
{
	if ( inString == NULL )
		return -1;
	else if ( wcscmp( inString, L"]" ) == 0 )
		return 0;

	wchar_t Owner[MAX_ARG_LEN] = L"";
	int i = 0, len = 0;
	
	if ( FH_IsColonOrEquals( inString[i] ) )
		i++;

	wcscpy_s( Owner, MAX_ARG_LEN, inString + i );

	len = (int) wcslen( Owner );

	if ( len )
		if ( Owner[len - 1] == ']' )
			Owner[len - 1] = '\0';

	// Printf( L"   * Owner range = '%s' %s\r\n", Owner, Reverse ? L"(reversed)" : L"" );

	_OwnerRange *OwnerRange = (_OwnerRange*) malloc( sizeof( _OwnerRange ) + ( wcslen( Owner ) * sizeof( wchar_t ) ) );

	if ( OwnerRange == NULL ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t allocate owner range structure\r\n" );
		return 1;
	}

	// Printf( L"   * Allocated memory at 0x%08x\r\n", OwnerRange );

	OwnerRange->Next			= *outPointer;
	OwnerRange->Type			= 'O';
	OwnerRange->Reverse			= Reverse;
	wcscpy_s( &(OwnerRange->Owner), wcslen( Owner ) + 1, Owner );

	*outPointer = OwnerRange;
	return 0;
}



int ParseRangeOption( LPTSTR inString, void **outPointer )
{
	if ( !inString || !outPointer )
		return -1;
	else if ( inString[0] == '\0' )
		return -1;


	if ( !_wcsnicmp( inString, L"/I", 2 ) && inString[2] )
		return ParseDescriptionRange( inString + 2, FALSE, TRUE, outPointer );
	else if ( !_wcsnicmp( inString, L"/!I", 3 ) && inString[3] )
		return ParseDescriptionRange( inString + 3, TRUE, TRUE, outPointer );


	wchar_t *ch		= inString;
	BOOL Reverse	= FALSE;

	if ( *ch != '/' )
		return 1;

	ch++;
	if ( *ch == '!' ) {
		Reverse = TRUE;
		ch++;
	}

	if ( *ch != '[' )
		return 1;

	ch++;

	if ( FH_LastChar( inString ) != ']' ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Range option must end with ']': \"%s\"\r\n", inString );
		return 1;
	}

	if ( Reverse && ( *ch == '!' ) ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Reverse exclusion range not supported: \"%s\"\r\n", inString );
		return 1;
	}

	switch ( towupper( *ch ) ) {

		case 'S' :
			return ParseSizeRange( ch + 1, Reverse, outPointer );

		case 'T' :
			return ParseTimeRange( ch + 1, Reverse, outPointer );

		case 'D' :
			return ParseDateRange( ch + 1, Reverse, outPointer );

		case '!' :
			return ParseExclusionRange( ch + 1, outPointer );

		case 'W' :
			return ParseDOWRange( ch + 1, Reverse, outPointer );

		case 'I' :
			return ParseDescriptionRange( ch + 1, Reverse, FALSE, outPointer );

		case 'O' :
			return ParseOwnerRange( ch + 1, Reverse, outPointer );

		default :
			DisplayErrorHeader();
			_wcsupr_s( inString, MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Unknown range type: \"%s\"\r\n", inString );
			return 1;
	}

	return 0;
}



BOOL LooksLikeARangeOption( LPTSTR inString )
{
	if ( !inString )
		return FALSE;


	if ( !wcsncmp( inString, L"/[", 2 ) || !wcsncmp( inString, L"/![", 3 ) )
		return TRUE;

	if ( !_wcsnicmp( inString, L"/I", 2 ) && inString[2] )
		return TRUE;

	if ( !_wcsnicmp( inString, L"/!I", 3 ) && inString[3] )
		return TRUE;

	return FALSE;
}



int ParseCommonOption( LPTSTR Option, BOOL AllowDirectories, DWORD *AttribMask, DWORD *AttribValues, int *Recurse, unsigned int *MinDepth, unsigned int *MaxDepth, BOOL *NoHighlight, void **FirstRange )
{

/**************************************************************************

	This function combines parsers for several common file-filter options:

		/A:			attributes
		/S			recursion
		/N			disable features
		/[] /![]	ranges
		/I /!I		file description filters

	Return value:

		  0			not a common option
		> 0			valid common option has been handled
		< 0			error

**************************************************************************/

	if ( !Option || !AttribMask || !AttribValues || !Recurse || !MinDepth || !MaxDepth || !FirstRange )
		return -1;


	if ( !wcsncmp( Option, L"/[", 2 ) || !wcsncmp( Option, L"/![", 3 ) ) {
		if ( ParseRangeOption( Option, FirstRange ) )
			return -10;
		else
			return 10;
	}
	else if ( !_wcsnicmp( Option, L"/A", 2 ) ) {
		if ( ParseAttributesOption( Option, FALSE, AttribMask, AttribValues ) )
			return -11;
		else
			return 11;
	}
	else if ( !_wcsnicmp( Option, L"/I", 2 ) ) {
		if ( ParseDescriptionRange( Option + 2, FALSE, TRUE, FirstRange ) )
			return -12;
		else
			return 12;
	}
	else if ( !_wcsnicmp( Option, L"/!I", 3 ) ) {
		if ( ParseDescriptionRange( Option + 3, TRUE, TRUE, FirstRange ) )
			return -12;
		else
			return 12;
	}
	else if ( !_wcsnicmp( Option, L"/N", 2 ) ) {
		if ( ParseNOption( Option, Recurse, NoHighlight ) )
			return -14;
		else
			return 14;
	}
	else if ( !_wcsnicmp( Option, L"/S", 2 ) ) {
		if ( ParseSOption( Option, Recurse, MinDepth, MaxDepth ) )
			return -15;
		else
			return 15;
	}

	return 0;
}



LPTSTR FH_PathPart( LPCTSTR Filename, LPTSTR outPath )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 )
		return NULL;

	if ( outPath ) {
		for ( size_t i = 0; i <= EndOfPathPart; i++ )
			outPath[i] = Filename[i];

		outPath[EndOfPathPart + 1] = '\0';
		return outPath;
	}

	return (LPTSTR) Filename;
}



LPTSTR FH_FilenamePart( LPCTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return (LPTSTR) Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return (LPTSTR) Filename + EndOfPathPart + 1;
}



LPTSTR FH_ExtensionPart( LPCTSTR Filename, LPTSTR outExt )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	LPCTSTR Ext			= NULL;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) || ( Filename[i] == ':' ) )
			Ext = NULL;
		else if ( Filename[i] == '.' )
			Ext = Filename + i;
	}

	if ( !Ext )
		return NULL;

	if ( outExt ) {
		wcscpy_s( outExt, MAX_FILENAME_LEN, Ext );
		return outExt;
	}

	return (LPTSTR) Ext;
}



int _FH_CopyDown( LPTSTR String, size_t Offset )
{
	if ( !String || !Offset )
		return -1;


	BOOL Done			= FALSE;


	for ( size_t i = 0; !Done; i++ ) {
		String[i] = String[i + Offset];
		Done = ( String[i] == '\0' );
	}

	return 0;
}



int _FH_GetReparsePointTarget( LPCTSTR Filename, LPTSTR Target )
{
	if ( !Filename || !Target )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]			= L"";

	DWORD Attribs							= 0;
	DWORD BufferSize						= sizeof( _FH_ReparseDataBuffer ) + ( MAX_FILENAME_LEN * sizeof( wchar_t ) );
	DWORD BytesReturned						= 0;
	_FH_ReparseDataBuffer *ReparseData		= NULL;
	unsigned int Tag						= 0;
	
	HANDLE FileHandle						= NULL;

	int rv				= 0;


	Attribs = GetFileAttributes( Filename );
	if ( ( Attribs == INVALID_FILE_ATTRIBUTES ) || ( ( Attribs & FILE_ATTRIBUTE_REPARSE_POINT ) == 0 ) )
		return 0;

	//	Printf( L"   * '%s' is a reparse point!\r\n", Filename );

	ReparseData = (_FH_ReparseDataBuffer*) malloc( BufferSize );

	if ( ReparseData ) {
		memset( ReparseData, 0x00, BufferSize );

		FileHandle = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT, NULL );
		if ( FileHandle != INVALID_HANDLE_VALUE ) {

			//	Printf( L"   * Opened file handle!\r\n" );

			if ( DeviceIoControl( FileHandle, FSCTL_GET_REPARSE_POINT, NULL, 0, ReparseData, BufferSize, &BytesReturned, NULL ) ) {

				//	Printf( L"   * DeviceIoControl() succeeded!  ReparseData->ReparseTag = 0x%08X\r\n", ReparseData->ReparseTag );
				Tag = ReparseData->ReparseTag;

				if ( Tag == FH_TAG_SYMLINK ) {
					memcpy( temp, &(ReparseData->SymbolicLinkReparseBuffer.PathBuffer) + ReparseData->SymbolicLinkReparseBuffer.PrintNameOffset, ReparseData->SymbolicLinkReparseBuffer.PrintNameLength );
					temp[ReparseData->SymbolicLinkReparseBuffer.PrintNameLength] = '\0';

					if ( temp[0] == '\0' ) {
						memcpy( temp, &(ReparseData->SymbolicLinkReparseBuffer.PathBuffer) + ReparseData->SymbolicLinkReparseBuffer.SubstituteNameOffset, ReparseData->SymbolicLinkReparseBuffer.SubstituteNameLength );
						temp[ReparseData->SymbolicLinkReparseBuffer.SubstituteNameLength / sizeof( wchar_t )] = '\0';
					}
				}
				else if ( Tag == FH_TAG_MOUNT_POINT ) {
					memcpy( temp, &(ReparseData->MountPointReparseBuffer.PathBuffer) + ReparseData->MountPointReparseBuffer.SubstituteNameOffset, ReparseData->MountPointReparseBuffer.SubstituteNameLength );
					temp[ReparseData->MountPointReparseBuffer.SubstituteNameLength / sizeof( wchar_t )] = '\0';

					if ( !wcsncmp( temp, L"\\??\\Volume{", 11 ) && iswxdigit( temp[11] ) ) {		//	virtual disk mounted to a junction??
						temp[1] = '\\';
						temp[2] = '.';
					}
				}

				if ( !wcsncmp( temp, L"\\??\\", 4 ) )
					_FH_CopyDown( temp, 4 );

				//	Printf( L"   * temp = '%s'\r\n", temp );

				if ( temp[0] ) {

					if ( Tag == FH_TAG_MOUNT_POINT )
						wcscpy_s( Target, MAX_FILENAME_LEN, temp );
					else if ( ( Tag == FH_TAG_SYMLINK ) && ( ( ReparseData->SymbolicLinkReparseBuffer.Flags ) & 1 ) ) {
						FH_PathPart( Filename, Target );
						FH_AppendFilename( Target, temp, MAX_FILENAME_LEN );
					}
					else if ( Tag == FH_TAG_SYMLINK )
						wcscpy_s( Target, MAX_FILENAME_LEN, temp );

					rv = 1;
				}
			}

			CloseHandle( FileHandle );
		}

		free( ReparseData );
	}

	return rv;
}



int _FHGetFileOwner( LPCTSTR Filename, LPTSTR Owner )
{
	if ( !Filename || !Owner )
		return -1;

	SECURITY_INFORMATION *FileSecurity = NULL;
	DWORD FileSecSize = 0;

	SID *OwnerSid = NULL;
	BOOL Dummy;

	wchar_t OwnerName[MAX_ARG_LEN]		= L"";
	wchar_t OwnerDomain[MAX_ARG_LEN]	= L"";
	DWORD OwnerNameSize = MAX_ARG_LEN, OwnerDomainSize = MAX_ARG_LEN;

	SID_NAME_USE SidNameUse = SidTypeUnknown;

	int rv = 1;


	GetFileSecurity( Filename, OWNER_SECURITY_INFORMATION, NULL, 0, &FileSecSize );
	if ( FileSecSize == 0 )
		return 1;

	FileSecurity = (SECURITY_INFORMATION*) malloc( FileSecSize );

	if ( FileSecurity ) {
		if ( GetFileSecurity( Filename, OWNER_SECURITY_INFORMATION, FileSecurity, FileSecSize, &FileSecSize ) )
			if ( GetSecurityDescriptorOwner( FileSecurity, (PSID*) &OwnerSid, &Dummy ) )
				if ( LookupAccountSid( NULL, OwnerSid, OwnerName, &OwnerNameSize, OwnerDomain, &OwnerDomainSize, &SidNameUse ) )
					rv = 0;

		free( FileSecurity );
	}

	if ( OwnerName[0] && OwnerDomain[0] ) {
		wcscpy_s( Owner, MAX_ARG_LEN, OwnerDomain );
		wcscat_s( Owner, MAX_ARG_LEN, L"\\" );
		wcscat_s( Owner, MAX_ARG_LEN, OwnerName );
	}

	return rv;
}



int _FHGetFileDescription( LPCTSTR Filename, LPTSTR Description, unsigned int BufferLen )
{
	if ( !Filename || !Description )
		return -1;

	if ( Filename == Description )
		return -1;

	if ( Filename[0] == '\0' )
		return ERROR_FILE_NOT_FOUND;

	if ( !QueryIsFileOrDirectory( (LPTSTR) Filename ) )
		return ERROR_FILE_NOT_FOUND;


	wchar_t temp[MAX_ARG_LEN]		= L"%@descript[";


	wcscat_s( temp, MAX_ARG_LEN, Filename );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	_FH_ExpandVars( temp );				//	ExpandVariables( temp, 1 );

	if ( (unsigned int) wcslen( temp ) >= BufferLen )
		return ERROR_BUFFER_OVERFLOW;

	wcscpy_s( Description, BufferLen, temp );
	return 0;
}



int CheckRanges( WIN32_FIND_DATA *FoundItem, LPCTSTR PathName, void *FirstRange )
{
	if ( !FoundItem || !PathName )
		return -1;

	if ( !FirstRange )
		return 0;


	_SizeRange *SizeRange				= (_SizeRange*) FirstRange;
	_TimeRange *TimeRange				= NULL;
	_DateRange *DateRange				= NULL;
	_ExclusionRange *ExclusionRange		= NULL;
	_DOWRange *DOWRange					= NULL;
	_DescriptionRange *DescriptionRange	= NULL;
	_OwnerRange *OwnerRange				= NULL;

	unsigned __int64 FileSize = 0;
	FILETIME *FileStamp = NULL;
	FILETIME LocalFileTime;
	SYSTEMTIME FileTimeUnpacked;
	unsigned int FileTimeAsSeconds = 0;
	unsigned int FileDateAsJulian = 0;
	wchar_t FullName[MAX_FILENAME_LEN]	= L"";
	wchar_t StringBuffer[MAX_ARG_LEN]	= L"";


	BOOL Okay = FALSE;


	while ( SizeRange ) {

		switch ( towupper( SizeRange->Type ) ) {

			case 'S' :
				FileSize = ( (__int64) FoundItem->nFileSizeHigh << 32 ) | FoundItem->nFileSizeLow;

				Okay = ( ( FileSize >= SizeRange->Smallest ) && ( FileSize <= SizeRange->Largest ) );
				if ( SizeRange->Reverse )
					Okay = !Okay;

				break;

			case 'T' :
				TimeRange = (_TimeRange*) SizeRange;

				if ( TimeRange->WhichStamp == FH_TIME_STAMP_LAST_ACCESS )
					FileStamp = &FoundItem->ftLastAccessTime;
				else if ( TimeRange->WhichStamp == FH_TIME_STAMP_CREATION )
					FileStamp = &FoundItem->ftCreationTime;
				else
					FileStamp = &FoundItem->ftLastWriteTime;

				FileTimeToLocalFileTime( FileStamp, &LocalFileTime );
				FileTimeToSystemTime( &LocalFileTime, &FileTimeUnpacked );
				FileTimeAsSeconds = FileTimeUnpacked.wSecond + ( 60 * FileTimeUnpacked.wMinute ) + ( 60 * 60 * FileTimeUnpacked.wHour );

				Okay = ( ( FileTimeAsSeconds >= TimeRange->EarliestTime ) && ( FileTimeAsSeconds <= TimeRange->LatestTime ) );
				if ( TimeRange->Reverse )
					Okay = !Okay;

				break;

			case 'D' :
				DateRange = (_DateRange*) SizeRange;

				if ( DateRange->WhichStamp == FH_TIME_STAMP_LAST_ACCESS )
					FileStamp = &FoundItem->ftLastAccessTime;
				else if ( DateRange->WhichStamp == FH_TIME_STAMP_CREATION )
					FileStamp = &FoundItem->ftCreationTime;
				else
					FileStamp = &FoundItem->ftLastWriteTime;

				FileTimeToLocalFileTime( FileStamp, &LocalFileTime );
				FileTimeToSystemTime( &LocalFileTime, &FileTimeUnpacked );
				FileDateAsJulian = _FHJulian( FileTimeUnpacked.wYear, FileTimeUnpacked.wMonth, FileTimeUnpacked.wDay );
				FileTimeAsSeconds = FileTimeUnpacked.wSecond + ( 60 * FileTimeUnpacked.wMinute ) + ( 60 * 60 * FileTimeUnpacked.wHour );

				Okay = ( ( FileDateAsJulian > DateRange->EarliestJulian ) || ( ( FileDateAsJulian == DateRange->EarliestJulian ) && ( FileTimeAsSeconds >= DateRange->EarliestTime ) ) );
				Okay &= ( ( FileDateAsJulian < DateRange->LatestJulian ) || ( ( FileDateAsJulian == DateRange->LatestJulian ) && ( FileTimeAsSeconds <= DateRange->LatestTime ) ) );
				if ( DateRange->Reverse )
					Okay = !Okay;

				break;

			case '!' :
				ExclusionRange = (_ExclusionRange*) SizeRange;
				Okay = ( WildcardComparison( &(ExclusionRange->Filespec), FoundItem->cFileName, 1, 1 ) != 0 );
				break;

			case 'W' :
				DOWRange = (_DOWRange*) SizeRange;

				if ( DOWRange->WhichStamp == FH_TIME_STAMP_LAST_ACCESS )
					FileStamp = &FoundItem->ftLastAccessTime;
				else if ( DOWRange->WhichStamp == FH_TIME_STAMP_CREATION )
					FileStamp = &FoundItem->ftCreationTime;
				else
					FileStamp = &FoundItem->ftLastWriteTime;

				FileTimeToLocalFileTime( FileStamp, &LocalFileTime );
				FileTimeToSystemTime( &LocalFileTime, &FileTimeUnpacked );
				FileDateAsJulian = _FHJulian( FileTimeUnpacked.wYear, FileTimeUnpacked.wMonth, FileTimeUnpacked.wDay );

				Okay = DOWRange->DOWMask & ( 1 << ( FileDateAsJulian % 7 ) );
				break;

			case 'I' :
				DescriptionRange = (_DescriptionRange*) SizeRange;

				wcscpy_s( FullName, MAX_FILENAME_LEN, PathName );
				FH_AppendFilename( FullName, FoundItem->cFileName, MAX_FILENAME_LEN );		//	MakeDirectoryName( FullName, FoundItem->cFileName );
				StringBuffer[0] = '\0';
				_FHGetFileDescription( FullName, StringBuffer, MAX_ARG_LEN );				//	ReadDescriptions( FullName, StringBuffer );

				Okay = ( WildcardComparison( &(DescriptionRange->Description), StringBuffer, 0, 1 ) == 0 );
				if ( DescriptionRange->Reverse )
					Okay = !Okay;

				break;

			case 'O' :
				OwnerRange = (_OwnerRange*) SizeRange;

				wcscpy_s( FullName, MAX_FILENAME_LEN, PathName );
				FH_AppendFilename( FullName, FoundItem->cFileName, MAX_FILENAME_LEN );		//	MakeDirectoryName( FullName, FoundItem->cFileName );
				StringBuffer[0] = '\0';

				_FHGetFileOwner( FullName, StringBuffer );

				Okay = ( WildcardComparison( &(OwnerRange->Owner), StringBuffer, 0, 1 ) == 0 );
				if ( OwnerRange->Reverse )
					Okay = !Okay;

				break;
		}

		if ( !Okay )
			return 1;
		
		SizeRange = (_SizeRange*) SizeRange->Next;
	}


	return 0;
}



BOOL _FH_IsSlash( wchar_t ch )
{
	return ( ( ch == '/' ) || ( ch == '\\' ) );
}



BOOL _FH_AlNum( wchar_t ch )
{
	if ( ( ch >= 'A' ) && ( ch <= 'Z' ) )
		return TRUE;

	if ( ( ch >= 'a' ) && ( ch <= 'z' ) )
		return TRUE;

	if ( ( ch >= '0' ) && ( ch <= '9' ) )
		return TRUE;

	return FALSE;
}



BOOL _FH_HasDriveLetter( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;

	if ( _FH_AlNum( Filename[0] ) )
		if ( Filename[1] == ':' )
			return TRUE;

	return FALSE;
}



BOOL _FH_HasDoubleSlash( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;

	if ( _FH_IsSlash( Filename[0] ) && _FH_IsSlash( Filename[1] ) )
		return TRUE;

	return FALSE;
}



BOOL _FH_HasDriveOrDoubleSlash( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;

	if ( _FH_AlNum( Filename[0] ) )
		if ( Filename[1] == ':' )
			return TRUE;

	if ( _FH_IsSlash( Filename[0] ) && _FH_IsSlash( Filename[1] ) )
		return TRUE;

	return FALSE;
}



int _FH_ReplaceChar( LPTSTR String, wchar_t From, wchar_t To )
{
	if ( !String )
		return -1;


	int Count		= 0;


	for ( size_t i = 0; String[i]; i++ )
		if ( String[i] == From )
			String[i] = To, Count++;

	return Count;
}



int _FH_GetHomePath( LPTSTR outPath, size_t Len )
{
	if ( !outPath )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	wchar_t temp2[MAX_FILENAME_LEN]		= L"";


	temp[0] = '\0';
	GetEnvironmentVariable( L"HOME", temp, MAX_FILENAME_LEN );
	if ( temp[0] ) {
		_FH_ReplaceChar( temp, '/', '\\' );
		if ( QueryIsDirectory( temp ) ) {
			//	Printf( L"   * Got directory from HOME variable:  '%s'\r\n", temp );
			wcscpy_s( outPath, Len, temp );
			return 1;
		}
	}

	temp[0]		= '\0';
	temp2[0]	= '\0';
	GetEnvironmentVariable( L"HOMEDRIVE", temp, MAX_FILENAME_LEN );
	GetEnvironmentVariable( L"HOMEPATH", temp2, MAX_FILENAME_LEN );
	if ( iswalpha( temp[0] ) && ( temp[1] == ':' ) && _FH_IsSlash( temp2[0] ) ) {
		wcscat_s( temp, MAX_FILENAME_LEN, temp2 );
		_FH_ReplaceChar( temp, '/', '\\' );
		if ( QueryIsDirectory( temp ) ) {
			//	Printf( L"   * Got directory from HOMEDRIVE and HOMEPATH variables:  '%s'\r\n", temp );
			wcscpy_s( outPath, Len, temp );
			return 1;
		}
	}

	temp[0] = '\0';
	SHGetFolderPath( NULL, CSIDL_PROFILE, NULL, SHGFP_TYPE_CURRENT, temp );
	if ( temp[0] ) {
		_FH_ReplaceChar( temp, '/', '\\' );
		if ( QueryIsDirectory( temp ) ) {
			//	Printf( L"   * Got directory from SHGetFolderPath():  '%s'\r\n", temp );
			wcscpy_s( outPath, Len, temp );
			return 1;
		}
	}

	return 0;
}



BOOL _FH_FixTildePath( LPTSTR String, size_t BufLen, BOOL MightBeFile )
{
	if ( !String )
		return FALSE;


	wchar_t temp[MAX_FILENAME_LEN]			= L"";
	wchar_t NewPath[MAX_FILENAME_LEN]		= L"";
	wchar_t HomePath[MAX_PATH]				= L"";


	wcscpy_s( temp, MAX_FILENAME_LEN, String );
	_FH_ReplaceChar( temp, '/', '\\' );

	if ( temp[0] != '~' )
		return FALSE;

	if ( temp[1] && ( temp[1] != '\\' ) )
		return FALSE;


	if ( MightBeFile ) {
		if ( temp[1] && QueryIsDirectory( L".\\~" ) )
			return FALSE;

		if ( ( temp[1] == '\0' ) && QueryIsFileOrDirectory( L".\\~" ) )
			return FALSE;
	}
	else {
		wchar_t temp2[MAX_FILENAME_LEN]		= L".\\";

		wcscat_s( temp2, MAX_FILENAME_LEN, temp );
		if ( QueryIsDirectory( temp2 ) )
			return FALSE;
	}
	

	_FH_GetHomePath( HomePath, MAX_FILENAME_LEN );
	if ( !HomePath[0] )
		return FALSE;

	wcscpy_s( NewPath, MAX_FILENAME_LEN, HomePath );

	if ( temp[1] && temp[2] )
		FH_AppendFilename( NewPath, temp + 2, MAX_FILENAME_LEN );

	wcscpy_s( String, BufLen, NewPath );
	return TRUE;
}



unsigned int _FHHasWildcards( LPCTSTR Filename )
{
	if ( !Filename )
		return 0;


	unsigned int Wildcards			= 0;


	for ( const wchar_t *ch = Filename; *ch; ch++ ) {
		if ( ( *ch == '\\' ) || ( *ch == '/' ) )
			Wildcards = 0;
		else if ( ( *ch == '*' ) || ( *ch == '?' ) )
			Wildcards++;
	}

	return Wildcards;
}



int _FH_SemiFix( LPTSTR Filename, wchar_t From, wchar_t To )
{
	if ( !Filename )
		return -1;


	int Count		= 0;
	size_t last		= 0;


	if ( iswalnum( Filename[0] ) && ( Filename[1] == ':' ) )
		last = 2;

	for ( size_t i = last + 1; Filename[i]; i++ )
		if ( _FH_IsSlash( Filename[i] ) )
			last = i;

	for ( size_t i = last + 1; Filename[i]; i++ )
		if ( Filename[i] == From )
			Filename[i] = To, Count++;

	return Count;
}



BOOL _FH_AbsPath( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	if ( iswalnum( Filename[0] ) && ( Filename[1] == ':' ) )
		return TRUE;

	if ( _FH_IsSlash( Filename[0] ) && _FH_IsSlash( Filename[1] ) )
		return TRUE;

	return FALSE;
}



int _FH_ParsePathPart( wchar_t **CharPtr, wchar_t **StartChar, LPTSTR outPart )
{
	if ( !CharPtr )
		return 0;
	else if ( *CharPtr == NULL )
		return 0;
	else if ( **CharPtr == '\0' )
		return 0;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	int j								= 0;
	wchar_t *ch							= *CharPtr;


	while ( _FH_IsSlash( *ch ) )
		ch++;

	if ( StartChar && *ch )
		*StartChar = ch;

	while ( *ch && ( !_FH_IsSlash( *ch ) ) ) {
		if ( j < ( MAX_FILENAME_LEN - 1 ) ) {
			temp[j++] = *ch;
			temp[j] = '\0';
		}
		ch++;
	}

	if ( temp[0] == '\0' )
		return 0;

	*CharPtr = ch;

	if ( outPart )
		wcscpy_s( outPart, MAX_FILENAME_LEN, temp );

	return j;
}



int _FH_RootPtr( LPTSTR Filename )
{
	if ( !Filename )
		return -1;


	int i			= 0;


	if ( _FH_HasDriveLetter( Filename ) )
		return 2;

	if ( _FH_HasDoubleSlash( Filename ) ) {
		i = 2;

		if ( Filename[i] && ( !_FH_IsSlash( Filename[i] ) ) ) {
			while ( Filename[i] && ( !_FH_IsSlash( Filename[i] ) ) )
				i++;				//	skip past hostname

			if ( _FH_IsSlash( Filename[i] ) ) {
				i++;				//	skip over hostname/sharename separator

				if ( Filename[i] && ( !_FH_IsSlash( Filename[i] ) ) ) {
					while ( Filename[i] && ( !_FH_IsSlash( Filename[i] ) ) )
						i++;		//	skip past sharename

					return i;
				}
			}
		}
	}

	return 0;
}



int _FH_GetPortion( LPTSTR inFilename, LPTSTR outPortion, unsigned int *Index )
{
	if ( !inFilename || !outPortion || !Index )
		return -1;


	int Ptr				= _FH_RootPtr( inFilename );
	wchar_t ch			= 0;


	outPortion[0]	= '\0';

	if ( (int) *Index > Ptr )
		Ptr = *Index;

	if ( inFilename[Ptr] == '\0' )
		return 0;


	for ( int i	= 0; ; i++ ) {
		ch = inFilename[i];
		outPortion[i]	= ch;

		if ( ch == '\0' ) {
			*Index = i;
			return i;
		}
		else if ( _FH_IsSlash( ch ) && ( i > Ptr ) ) {
			outPortion[i] = '\0';
			*Index = i;
			return i;
		}
	}

	return 0;
}



int _FH_ResolveReparsePoints( LPTSTR Filename )
{
	if ( !Filename )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	wchar_t Target[MAX_FILENAME_LEN]	= L"";

	unsigned int Index					= 0;

	int Count			= 0;


	while ( _FH_GetPortion( Filename, temp, &Index ) ) {
		if ( _FH_GetReparsePointTarget( temp, Target ) > 0 ) {

			if ( Filename[Index] )
				wcscpy_s( temp, MAX_FILENAME_LEN, Filename + Index );
			else
				temp[0] = '\0';

			wcscpy_s( Filename, MAX_FILENAME_LEN, Target );

			if ( temp[0] ) {
				if ( ( temp[0] == '\\' ) || ( temp[0] == '/' ) )
					wcscat_s( Filename, MAX_FILENAME_LEN, temp + 1 );
				else
					wcscat_s( Filename, MAX_FILENAME_LEN, temp );
			}

			Index = 0;
		}
	}


	return Count;
}



int FH_Truename( LPCTSTR inFilename, LPTSTR outTruename, size_t BufLen, unsigned int Flags )
{
	if ( !inFilename || !outTruename )
		return -1;
	
	if ( inFilename[0] == '\0' )
		return -1;


	wchar_t Truename[MAX_FILENAME_LEN]		= L"";
	wchar_t Truename2[MAX_FILENAME_LEN]		= L"";


	wcscpy_s( Truename, MAX_FILENAME_LEN, inFilename );

	_FH_ReplaceChar( Truename, '/', '\\' );

	if ( Flags & TRUENAME_FLAG_INCLUDE_LISTS )
		_FH_SemiFix( Truename, ';', '|' );

	if ( ( Flags & TRUENAME_FLAG_HOME_PATHS ) && ( Truename[0] == '~' ) )
		_FH_FixTildePath( Truename, MAX_FILENAME_LEN, TRUE );

	if ( Flags & TRUENAME_FLAG_REPARSE_POINTS )
		_FH_ResolveReparsePoints( Truename );

	if ( _FH_HasDriveLetter( Truename ) )
		Truename[0] = towupper( Truename[0] );

	Truename2[0] = '\0';
	QueryTrueName( Truename, Truename2 );

	if ( Truename2[0] && _wcsicmp( Truename, Truename2 ) )
		wcscpy_s( Truename, MAX_FILENAME_LEN, Truename2 );

	if ( wcslen( Truename ) >= BufLen )
		return ERROR_BUFFER_OVERFLOW;

	wcscpy_s( outTruename, BufLen, Truename );
	return 0;
}



int FH_Truename( LPCTSTR inFilename, LPTSTR outTruename )
{
	return FH_Truename( inFilename, outTruename, MAX_FILENAME_LEN, TRUENAME_FLAG_HOME_PATHS );
}



int FH_WildcardComparison2( LPCTSTR Wildspec, LPCTSTR Filename, int Ext, int Brackets, BOOL IncludeLists )
{
	if ( !Wildspec || !Filename )
		return -1;


	if ( !IncludeLists )
		return WildcardComparison( (LPTSTR) Wildspec, (LPTSTR) Filename, Ext, Brackets );


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	size_t j							= 0;
	BOOL Done							= FALSE;
	int rv								= -2;


	for ( size_t i = 0; !Done; i++ ) {

		if ( ( Wildspec[i] == '|' ) || ( Wildspec[i] == '\0' ) ) {
			if ( Wildspec[i] == '\0' )
				Done = TRUE;

			j = 0;

			if ( temp[0] ) {
				rv = WildcardComparison( temp, (LPTSTR) Filename, Ext, Brackets );
				if ( !rv )
					return rv;
			}
		}
		else if ( j < ( MAX_FILENAME_LEN - 1 ) ) {
			temp[j++] = Wildspec[i];
			temp[j] = '\0';
		}
	}

	return rv;
}



BOOL FH_IsPhule( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;

	if ( !wcscmp( Filename, L"." ) || !wcscmp( Filename, L".." ) )
		return TRUE;
	else
		return FALSE;
}



WORD _FH_HighlightAttr()
{
	BYTE FirstColor[16]		= { 0x0f, 0x1f, 0x2f, 0x3f, 0x4f, 0x5f, 0x6f, 0x70, 0x8f, 0x9f, 0xa0, 0xb0, 0xcf, 0xd0, 0xe0, 0xf2 };
	BYTE SecondColor[16]	= { 0x0e, 0x1b, 0x2b, 0x3e, 0x4b, 0x5b, 0x6b, 0x79, 0x8b, 0x9b, 0xa9, 0xb9, 0xc0, 0xde, 0xec, 0xf0 };

	HANDLE ConsoleHandle	= NULL;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	WORD NormalColor		= 0;
	WORD Highlight			= 0;
	int fg					= 0;
	int bg					= 0;


	ConsoleHandle = CreateFile( L"conout$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );
	if ( ConsoleHandle == INVALID_HANDLE_VALUE )
		return 0x0007;

	memset( &ConsoleInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
	GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
	NormalColor = ConsoleInfo.wAttributes;


	fg = ( NormalColor & 0x000f );
	bg = ( NormalColor & 0x00f0 ) >> 4;


	if ( fg == bg )
		Highlight = 0x000f;
	else if ( fg && !( fg & 0x0008 ) && !( bg & 0x0008 ) )
		Highlight = NormalColor | 0x0008;
	else if ( fg == FirstColor[bg] )
		Highlight = SecondColor[bg];
	else
		Highlight = FirstColor[bg];

	SetConsoleTextAttribute( ConsoleHandle, Highlight );
	CloseHandle( ConsoleHandle );
	return NormalColor;
}



int _FH_RestoreAttr( WORD Attr )
{
	HANDLE ConsoleHandle		= NULL;


	ConsoleHandle = CreateFile( L"conout$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );
	if ( ConsoleHandle == INVALID_HANDLE_VALUE )
		return 1;

	SetConsoleTextAttribute( ConsoleHandle, Attr );
	CloseHandle( ConsoleHandle );
	return 0;
}



int _FH_DisplayFileException( LPCTSTR Filename, int Flags, int rv )
{
	WORD NormalColor		= _FH_HighlightAttr();


	_cwprintf( L"\n\u2022  EXCEPTION in FileHandler():   Filename = \"%s\"   Flags = 0x%04X\n", Filename ? Filename : L"<NULL>", Flags );
	_cwprintf( L"   RETURNING :  %d\n\n", rv );
	_FH_RestoreAttr( NormalColor );
	return 0;
}



int _FH_HandleListFile( LPCTSTR ListFilename, int Flags, void *MoreArgs, int (*FileHandler) ( LPCTSTR Filename, int Flags, void *MoreArgs ), int Recurse )
{
	if ( !ListFilename )
		return -1;
	else if ( ListFilename[0] == '\0' )
		return -1;


	wchar_t Line[MAX_FILENAME_LEN]				= L"";
	wchar_t CanonicalName[MAX_FILENAME_LEN]		= L"";

	_MMFile MMFile;
	BOOL Complain		= !( Flags & FH_FLAG_SUPPRESS_FNF_ERROR );
	int Unicode			= 0;
	int fhrv			= 0;
	int rv				= 0;
	int erv				= ( Recurse & FH_FLAG_EXCEPTIONS_DO_NOT_ABORT ) ? FH_EXCEPTION_RETURN : FH_USER_ABORT;



	if ( OpenMMFile( &MMFile, ListFilename, 0, GetOEMCP(), 0 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening list file \"%s\"\r\n", ListFilename );
		return 2;
	}

	GuessFileEncoding( &MMFile, FALSE );


	while ( MMFile.Index < MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, Line, MAX_FILENAME_LEN, MMGL_FLAG_SKIP_BLANKS ) )
			break;

		FH_TrimLine( Line );
		if ( Line[0] == ':' )
			continue;

		FH_StripQuotes( Line );
		if ( Line[0] == '\0' )
			continue;

		CanonicalName[0] = '\0';
		FH_Truename( Line, CanonicalName, MAX_FILENAME_LEN, TRUENAME_FLAG_HOME_PATHS );

		if ( CanonicalName[0] ) {
			if ( QueryIsFileOrDirectory( CanonicalName ) ) {

				try {
					fhrv = FileHandler( CanonicalName, Flags, MoreArgs );
				}
				catch ( ... ) {
					fhrv = erv;
					_FH_DisplayFileException( CanonicalName, Flags, fhrv );
				}

				if ( fhrv == FH_USER_ABORT ) {
					rv = FH_USER_ABORT;
					break;
				}
				else if ( fhrv )
					rv = fhrv;
			}
			else if ( Complain ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"File not found: %s\r\n", CanonicalName );
				rv = 2;
			}
		}
	}

	CloseMMFile( &MMFile );
	return rv;
}



int HandleMatchingFiles( LPCTSTR Wildspec, int Flags, void *MoreArgs, int (*FileHandler) ( LPCTSTR Filename, int Flags, void *MoreArgs ), int Recurse, DWORD AttribMask, DWORD AttribValues, void *FirstRange, unsigned int MinDepth, unsigned int MaxDepth )
{
	if ( !Wildspec )
		return -1;
	else if ( Wildspec[0] == '\0' )
		return -1;


	int rv		= 0;
	int erv		= ( Recurse & FH_FLAG_EXCEPTIONS_DO_NOT_ABORT ) ? FH_EXCEPTION_RETURN : FH_USER_ABORT;

	//	Printf( L"   * MinDepth = %u   MaxDepth = %u\r\n", MinDepth, MaxDepth );
	

	if ( Recurse & FH_FLAG_ALLOW_CLIPBOARD ) {
		if ( _wcsicmp( Wildspec, L"clip:" ) == 0 ) {
			try {
				rv = FileHandler( Wildspec, Flags, MoreArgs ) ? -1 : 0;
			}
			catch ( ... ) {
				rv = erv;
				_FH_DisplayFileException( Wildspec, Flags, erv );
			}

			return rv;
		}
	}


	if ( ( Recurse & FH_FLAG_ALLOW_URLS ) && QueryIsURL( (LPTSTR) Wildspec ) ) {
		try {
			rv = FileHandler( Wildspec, Flags, MoreArgs ) ? -1 : 0;
		}
		catch ( ... ) {
			rv = erv;
			_FH_DisplayFileException( Wildspec, Flags, erv );
		}

		return rv;
	}


	if ( ( Recurse & FH_FLAG_ALLOW_FILE_LISTS ) && ( Wildspec[0] == '@' ) && Wildspec[1] )
		if ( !_FHHasWildcards( Wildspec ) )
			if ( !QueryIsFile( (LPTSTR) Wildspec ) )
				if ( QueryIsFile( (LPTSTR) Wildspec + 1 ) )
					return _FH_HandleListFile( Wildspec + 1, Flags, MoreArgs, FileHandler, Recurse );


	if ( Recurse & FH_FLAG_MINUS_IS_SPECIAL ) {
		if ( !wcscmp( Wildspec, L"-" ) ) {
			try {
				rv = FileHandler( Wildspec, Flags, MoreArgs ) ? -1 : 0;
			}
			catch ( ... ) {
				rv = erv;
				_FH_DisplayFileException( Wildspec, Flags, erv );
			}

			return rv;
		}
	}
	

	wchar_t TrueName[MAX_FILENAME_LEN]	= L"";
	wchar_t Filename[MAX_FILENAME_LEN]  = L"";
	wchar_t Pathname[MAX_FILENAME_LEN]  = L"";
	wchar_t WildPart[MAX_FILENAME_LEN]  = L"";
	//	wchar_t Extension[MAX_FILENAME_LEN]	= L"";

	unsigned int TruenameFlags			= TRUENAME_FLAG_HOME_PATHS;

	HANDLE FindHandle = NULL;
	WIN32_FIND_DATA FoundItem;
	wchar_t ErrorText[SHORT_BUF_LEN] = L"";
	BOOL FoundAny = FALSE, RangeOkay = TRUE, Okay = TRUE, ErrorShown = FALSE;
	int rv2 = 0, err = 0;


	if ( Recurse & FH_FLAG_INCLUDE_LISTS )
		TruenameFlags |= TRUENAME_FLAG_INCLUDE_LISTS;

	if ( Recurse & FH_FLAG_CANONICALIZED )
		wcscpy_s( TrueName, MAX_FILENAME_LEN, Wildspec );
	else {
		FH_Truename( Wildspec, TrueName, MAX_FILENAME_LEN, TruenameFlags );
		Recurse |= FH_FLAG_CANONICALIZED;
	}


	FH_PathPart( TrueName, Pathname );
	FH_FilenamePart( TrueName, WildPart );
	//	FH_ExtensionPart( WildPart, Extension );



	if ( ( ( Recurse & ( FH_FLAG_RECURSE | FH_FLAG_RECURSE_BOTTOM_UP ) ) == ( FH_FLAG_RECURSE | FH_FLAG_RECURSE_BOTTOM_UP ) ) && MaxDepth ) {
		wcscpy_s( Filename, MAX_FILENAME_LEN, Pathname );
		wcscat_s( Filename, MAX_FILENAME_LEN, L"*" );

		FindHandle = FindFirstFile( Filename, &FoundItem );

		if ( FindHandle != INVALID_HANDLE_VALUE ) {

			do {
				if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
					continue;

				if ( FoundItem.dwFileAttributes & ( FILE_ATTRIBUTE_DEVICE | FILE_ATTRIBUTE_VIRTUAL ) )
					continue;

				if ( FH_IsPhule( FoundItem.cFileName ) )
					continue;

				if ( Recurse & FH_FLAG_NO_JUNCTIONS )
					if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT )
						continue;

				if ( Recurse & FH_FLAG_NO_HIDDEN_DIRS )
					if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN )
						continue;

				if ( Recurse & FH_FLAG_NO_SYSTEM_DIRS )
					if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM )
						continue;

				wcscpy_s( Filename, MAX_FILENAME_LEN, Pathname );
				wcscat_s( Filename, MAX_FILENAME_LEN, FoundItem.cFileName );
				wcscat_s( Filename, MAX_FILENAME_LEN, L"\\" );
				wcscat_s( Filename, MAX_FILENAME_LEN, WildPart );

				rv2 = HandleMatchingFiles( Filename, Flags, MoreArgs, FileHandler, Recurse | FH_FLAG_SUPPRESS_FNF_ERROR | FH_FLAG_CANONICALIZED, AttribMask, AttribValues, FirstRange, MinDepth ? MinDepth - 1 : 0, ( MaxDepth == UINT_MAX ) ? MaxDepth : MaxDepth - 1 );

				if ( rv2 == FH_USER_ABORT ) {
					rv = FH_USER_ABORT;
					break;
				}
				else if ( rv2 < 0 )
					rv = rv2;
				else if ( rv2 == 0 )
					FoundAny = TRUE;

			} while ( FindNextFile( FindHandle, &FoundItem ) );

			FindClose( FindHandle );
		}
	}

	if ( rv == FH_USER_ABORT )
		return rv;


	if ( !MinDepth ) {

		wcscpy_s( Filename, MAX_FILENAME_LEN, Pathname );
		wcscat_s( Filename, MAX_FILENAME_LEN, L"*" );

		FindHandle = FindFirstFile( Filename, &FoundItem );

		if ( FindHandle == INVALID_HANDLE_VALUE ) {
			err = GetLastError();
			if ( err != ERROR_FILE_NOT_FOUND ) {
				DisplayErrorHeader();
				FH_GetErrorText( err, ErrorText );
				Qprintf( ERRHANDLE, L"Error matching \"%s\" :\r\n", TrueName );
				Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
				//	Qprintf( ERRHANDLE, L"Error matching \"%s\": %s\r\n", TrueName, ErrorText );
				ErrorShown = TRUE;
				rv = -1;
			}
		}
		else {
			do {
				if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DEVICE )
					continue;

				if ( ( FoundItem.ftCreationTime.dwHighDateTime | FoundItem.ftCreationTime.dwLowDateTime | FoundItem.ftLastWriteTime.dwHighDateTime | FoundItem.ftLastWriteTime.dwLowDateTime ) == 0 )
					continue;

				if ( ( Recurse & FH_FLAG_ALLOW_DOTS ) == 0 )
					if ( FH_IsPhule( FoundItem.cFileName ) )
						continue;

				if ( FH_WildcardComparison2( WildPart, FoundItem.cFileName, 1, 1, ( Recurse & FH_FLAG_INCLUDE_LISTS ) != 0 ) )
					continue;

				if ( ( FoundItem.dwFileAttributes & AttribMask ) != ( AttribValues & AttribMask ) )
					continue;

				if ( CheckRanges( &FoundItem, Pathname, FirstRange ) )
					continue;

				FoundAny = TRUE;

				wcscpy_s( Filename, MAX_FILENAME_LEN, Pathname );
				wcscat_s( Filename, MAX_FILENAME_LEN, FoundItem.cFileName );

				try {
					rv2 = FileHandler( Filename, Flags, MoreArgs );
				}
				catch ( ... ) {
					rv2 = erv;
					_FH_DisplayFileException( Filename, Flags, erv );
				}

				if ( rv2 == FH_USER_ABORT ) {
					rv = FH_USER_ABORT;
					break;
				}
				else if ( rv2 )
					rv = rv2;

			} while ( FindNextFile( FindHandle, &FoundItem ) );

			FindClose( FindHandle );
			FindHandle = NULL;
		}
	}

	if ( rv == FH_USER_ABORT )
		return rv;


	if ( ( ( Recurse & ( FH_FLAG_RECURSE | FH_FLAG_RECURSE_BOTTOM_UP ) ) == FH_FLAG_RECURSE ) && MaxDepth ) {
		wcscpy_s( Filename, MAX_FILENAME_LEN, Pathname );
		wcscat_s( Filename, MAX_FILENAME_LEN, L"*" );

		FindHandle = FindFirstFile( Filename, &FoundItem );

		if ( FindHandle != INVALID_HANDLE_VALUE ) {

			do {
				if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
					continue;

				if ( FoundItem.dwFileAttributes & ( FILE_ATTRIBUTE_DEVICE | FILE_ATTRIBUTE_VIRTUAL ) )
					continue;

				if ( FH_IsPhule( FoundItem.cFileName ) )
					continue;

				if ( Recurse & FH_FLAG_NO_JUNCTIONS )
					if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT )
						continue;

				if ( Recurse & FH_FLAG_NO_HIDDEN_DIRS )
					if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN )
						continue;

				if ( Recurse & FH_FLAG_NO_SYSTEM_DIRS )
					if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM )
						continue;

				wcscpy_s( Filename, MAX_FILENAME_LEN, Pathname );
				wcscat_s( Filename, MAX_FILENAME_LEN, FoundItem.cFileName );
				wcscat_s( Filename, MAX_FILENAME_LEN, L"\\" );
				wcscat_s( Filename, MAX_FILENAME_LEN, WildPart );

				rv2 = HandleMatchingFiles( Filename, Flags, MoreArgs, FileHandler, Recurse | FH_FLAG_SUPPRESS_FNF_ERROR | FH_FLAG_CANONICALIZED, AttribMask, AttribValues, FirstRange, MinDepth ? MinDepth - 1 : 0, ( MaxDepth == UINT_MAX ) ? MaxDepth : MaxDepth - 1 );

				if ( rv2 == FH_USER_ABORT ) {
					rv = FH_USER_ABORT;
					break;
				}
				else if ( rv2 < 0 )
					rv = rv2;
				else if ( rv2 == 0 )
					FoundAny = TRUE;

			} while ( FindNextFile( FindHandle, &FoundItem ) );

			FindClose( FindHandle );
		}
	}

	if ( rv == FH_USER_ABORT )
		return rv;


	if ( !FoundAny ) {
		if ( !rv )
			rv = 1;

		if ( ( ( Recurse & FH_FLAG_SUPPRESS_FNF_ERROR ) == 0 ) && !ErrorShown ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No matching files: \"%s\"\r\n", TrueName );
		}
	}

	return rv;
}



LPCTSTR FH_RootSlash( LPCTSTR Filename )
{
	const wchar_t *ch				= NULL;
	int k							= 0;


	if ( _FH_IsSlash( Filename[0] ) && _FH_IsSlash( Filename[1] && iswalpha( Filename[2] ) ) ) {
		for ( ch = Filename; *ch; ch++ ) {
			if ( _FH_IsSlash( *ch ) ) {
				k++;
				if ( k == 4 )
					return ch;
			}
		}
	}


	ch = Filename;

	if ( _FH_IsSlash( ch[0] ) && ( ch[1] == '?' ) && ( ch[2] == '?' ) && _FH_IsSlash( ch[3] ) )
		ch += 4;

	if ( ( ( *ch >= 'A' ) && ( *ch <= 'Z' ) ) || ( ( *ch >= 'a' ) && ( *ch <= 'z' ) ) || ( ( *ch >= '0' ) && ( *ch <= '9' ) ) )
		if ( ( ch[1] == ':' ) && _FH_IsSlash( ch[2] ) )
			return ch + 2;

	return NULL;
}



int FH_TrimTrailingSlash( LPTSTR Filename )
{
	if ( !Filename )
		return -1;


	size_t len					= wcslen( Filename );
	int rv						= 0;


	if ( len > 1 )
		if ( _FH_IsSlash( Filename[len - 1] ) )
			if ( !_FH_IsSlash( Filename[len - 2] ) && ( Filename[len - 2] != ':' ) )
				if ( ( Filename + len - 1 ) != FH_RootSlash( Filename ) )
					Filename[len - 1] = '\0', rv++;

	return rv;
}



DLLExports int WINAPI _fhver$( LPTSTR outString )
{
	if ( !outString )
		return -1;

	MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, FH_VER_STRING, -1, outString, SHORT_BUF_LEN );
	return 0;
}


#endif		//	FILE_HANDLER_VERSION


/*

	1.0.12.0	2020-02-21

	1.0.13.0	2022-03-05
				Fix for ~ home paths not being expanded correctly.

	1.0.14.0	2023-08-23
				Made filename args to many functions constant.
				Added FH_TrimTrailingSlash().

	1.0.14.1	2023-09-20
				Broke out FH_RootSlash() as a separate function.

	1.0.14.2	2023-12-15
				Removed dependency on #define QUOTE.

	1.0.14.3	2024-03-24
				Tweaked FH_Truename() to support nested directory aliases.

	1.0.15.0	2024-04-01
				Rewrote FH_Truename() to use QueryTrueName() to handle as much as
				it can -- dots fix, directory aliases, doubled slashes, etc.
				Should help with odd stuff like nested directory aliases,
				directory aliases containing non-alphanumeric characters, etc.

*/
