
//			@OBSCURE and @CLARIFY, @BINTOB85 and @B85TOBIN functions:



#define OBSCURE_HEADER_SIZE		8
#define OBSCURE_HEADER_CHARS	10
#define OBSCURE_BUFFER_SIZE		( 4 * MAX_ARG_LEN )
#define OBSCURE_KEY				0x0fd6614a
#define KEY_2					0x0130

#define HEADER_KEY_2		0
#define HEADER_LENGTH_LO	1
#define HEADER_KEY_0		2
#define HEADER_CHECK_HI		3
#define HEADER_KEY_3		4
#define HEADER_LENGTH_HI	5
#define HEADER_CHECK_LO		6
#define HEADER_KEY_1		7


const wchar_t Base85Digits[86] = L"!#$'()*+,-./0123456789:;=?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\_abcdefghijklmnopqrstuvwxyz{}~";



int UTF16toUTF8( LPTSTR inString, BYTE *outBytes, int MaxBytes )
{
	if ( !inString || !outBytes )
		return -1;

	if ( MaxBytes < 1024 )
		return -1;

	if ( inString[0] == '\0' )
		return 0;

	int cb = 0;
	wchar_t ch = '\0';

	for ( int i = 0; ch = inString[i]; i++ ) {
		if ( ch < 0x80 ) {
			if ( cb > ( MaxBytes - 1 ) )
				return -2;

			outBytes[cb++] = (BYTE) ch;
		}
		else if ( ch < 0x0800 ) {
			if ( cb > ( MaxBytes - 2 ) )
				return -2;

			outBytes[cb++] = ( ch >> 6 ) | 0xc0;
			outBytes[cb++] = ( ch & 0x3f ) | 0x80;
		}
		else {
			if ( cb > ( MaxBytes - 3 ) )
				return -2;

			outBytes[cb++] = ( ch >> 12 ) | 0xe0;
			outBytes[cb++] = ( ( ch >> 6 ) & 0x3f ) | 0x80;
			outBytes[cb++] = ( ch & 0x3f ) | 0x80;
		}
	}

	return cb;
}


int UTF8toUTF16( BYTE *inBytes, LPTSTR outString, int CountBytes )
{
	if ( !inBytes || !outString || ( CountBytes < 0 ) )
		return -1;

	int j = 0;
	BYTE B1 = 0, B2 = 0, B3 = 0;

	for ( int i = 0; i < CountBytes; i++ ) {
		B1 = inBytes[i];

		if ( B1 < 0x80 )
			outString[j++] = B1;
		else if ( B1 < 0xc0 )
			return -1;
		else if ( B1 < 0xe0 ) {
			if ( i >= ( CountBytes - 1 ) )
				return -1;

			i++;
			B2 = inBytes[i];
			if ( ( B2 & 0xc0 ) != 0x80 )
				return -1;
			else
				outString[j++] = ( ( B1 & 0x1f ) << 6 ) | ( B2 & 0x3f );
		}
		else if ( B1 < 0xf0 ) {
			if ( i >= ( CountBytes - 2 ) )
				return -1;

			i++;
			B2 = inBytes[i++];
			B3 = inBytes[i];
			if ( ( ( B2 & 0xc0 ) != 0x80 ) || ( ( B3 & 0xc0 ) != 0x80 ) )
				return -1;
			else
				outString[j++] = ( ( B1 & 0x0f ) << 12 ) | ( ( B2 & 0x3f ) << 6 ) | ( B3 & 0x3f );
		}
		else
			return -1;
	}

	outString[j] = 0x00;
	return j;
}


int Checksum( BYTE *inBytes, int CountBytes )
{
	if ( !inBytes || ( CountBytes < 0 ) )
		return -1;

	int a = 0;

	for ( int i = 0; i < CountBytes; i++ )
		a = ( ( a << 1 ) + inBytes[i] ) & 0xffff;

	return a;
}


unsigned long int Reverse( unsigned long int Value, int Bits )
{
	unsigned long int rv = 0;

	for ( int i = 0; i < Bits; i++ ) {
		rv = ( rv << 1 ) | ( Value & 1 );
		Value = Value >> 1;
	}

	return rv;
}


DWORD Rol( DWORD Value, int Bits )
{
	for ( int i = 0; i < Bits; i++ )
		Value = ( ( Value << 1 ) & UINT_MAX ) | ( Value >> 31 );

	return Value;
}


DWORD Rol( DWORD Value )
{
	return ( ( Value << 1 ) & UINT_MAX ) | ( Value >> 31 );
}


int DumpBufferHex( BYTE *inBytes, int CountBytes )
{
	for ( int i = 0; i < CountBytes; i++ ) {
		Printf( L"%02x ", inBytes[i] );
		if ( ( i & 0x0f ) == 0x0f )
			CrLf();
	}

	CrLf();
	CrLf();
	return 0;
}


int BinToBase85( BYTE *inBytes, int CountBytes, LPTSTR outString )
{
	if ( !inBytes || !outString || ( CountBytes < 0 ) )
		return -1;

	int d[5];
	unsigned int a = 0;
	int j = 0;


	for ( int i = 0; i < CountBytes; i += 4 ) {
		a = 0;

		for ( int k = 0; k < 4; k++ ) {
			a = a << 8;

			if ( ( i + k ) < CountBytes )
				a |= inBytes[i + k];
		}

		for ( int k = 0; k < 5; k++ ) {
			d[k] = a % 85;
			a /= 85;
		}

		for ( int k = 4; k >= 0; k-- )
			outString[j++] = Base85Digits[d[k]];
	}

	outString[j] = 0x00;
	return j;
}


int Base85Value( wchar_t ch )
{
	for ( int i = 0; i < 85; i++ )
		if ( Base85Digits[i] == ch )
			return i;

	return -1;
}


int Base85ToBin( LPTSTR inString, BYTE *outBytes, int MaxBytes )
{
	if ( !inString || !outBytes || ( MaxBytes < 0 ) )
		return -1;

	int j = 0;
	unsigned long int a = 0;
	int v = 0;

	if ( MaxBytes == 0 )
		MaxBytes = INT_MAX;

	for ( int i = 0; inString[i]; ) {

		a = 0;
		
		for ( int k = 0; k < 5; k++ ) {
			v = Base85Value( inString[i++] );

			if ( v < 0 )
				return -2;
			else
				a = ( a * 85 ) + v;
		}

		if ( a > UINT_MAX )
			return -2;

		for ( int k = 0; k < 4; k++ ) {
			if ( j < MaxBytes ) {
				outBytes[j++] = ( a >> 24 );
				a = ( a << 8 ) & UINT_MAX;
			}
		}

		if ( j >= MaxBytes )
			return j;
	}

	return j;
}


int Mangle( BYTE *Buffer, int CountBytes, DWORD Key, WORD Key2 )
{
	if ( !Buffer )
		return -1;

	for ( int i = 0; i < CountBytes; i++ ) {
		Buffer[i] ^= ( Key & 0xff );
		Key = Rol( Key ) ^ Key2;
	}

	return 0;
}


int Erase( LPTSTR inString )
{
	if ( !inString )
		return -1;

	memset( inString, 0x00, MAX_ARG_LEN * sizeof( wchar_t ) );
	return 0;
}


unsigned int CheckDebug()
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%$debug$";
	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '\0' )
		return 0;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;
	
	_set_errno( 0 );

	if ( ( shortbuf[0] == '0' ) && ( towupper( shortbuf[1] ) == 'X' ) && iswxdigit( shortbuf[2] ) ) {
		//		parse hexadecimal

		value = wcstoul( shortbuf + 2, &sep1, 16 );

		if ( errno )
			return 0;
		else if ( *sep1 != 0x00 )
			return 0;

		return value;
	}
	else if ( iswdigit( shortbuf[0] ) ) {
		//		parse unsigned decimal

		value = wcstoul( shortbuf, &sep1, 10 );

		if ( errno )
			return 0;
		else if ( *sep1 != 0x00 )
			return 0;

		return value;
	}

	return 0;
}


DLLExports int WINAPI f_obscure( LPTSTR inString )
{
	if ( !inString )
		return 1;

	unsigned int Debug = CheckDebug();

	if ( wcslen( inString ) >= MAX_ARG_LEN ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Argument too long\r\n" );
		return -1;
	}

	BYTE Header[OBSCURE_HEADER_SIZE], Buffer[OBSCURE_BUFFER_SIZE];
	
	DWORD Key = GetRand( 0xfffffffe ) + 1;

	if ( Debug )
		Printf( L"   * Key = 0x%08x\r\n", Key );

	DWORD Key2 = Reverse( Key ^ OBSCURE_KEY, 32 );
	Header[HEADER_KEY_0] = Key2 & 0xff;
	Header[HEADER_KEY_1] = ( Key2 >> 8 ) & 0xff;
	Header[HEADER_KEY_2] = ( Key2 >> 16 ) & 0xff;
	Header[HEADER_KEY_3] = ( Key2 >> 24 );


	int BytesInBuffer = UTF16toUTF8( inString, Buffer, OBSCURE_BUFFER_SIZE );
	Erase( inString );

	if ( BytesInBuffer == -2 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return -1;
	}

	if ( Debug )
		Printf( L"   * BytesInBuffer = %u\r\n", BytesInBuffer );

	int Check = Checksum( Buffer, BytesInBuffer );
	if ( Debug )
		Printf( L"   * Check = 0x%04x\r\n", Check );

	DWORD Check2 = Check ^ ( Key >> 16 );
	Header[HEADER_CHECK_LO] = Check2 & 0xff;
	Header[HEADER_CHECK_HI] = (BYTE) ( Check2 >> 8 );

	DWORD Length2 = Reverse( BytesInBuffer, 16 ) ^ Check ;
	Header[HEADER_LENGTH_LO] = Length2 & 0xff;
	Header[HEADER_LENGTH_HI] = (BYTE) ( Length2 >> 8 );

	BinToBase85( Header, OBSCURE_HEADER_SIZE, inString );

	if ( Debug & 2 ) {
		Printf( L"Header -----\r\n" );
		DumpBufferHex( Header, OBSCURE_HEADER_SIZE );
		Printf( L"Data -----\r\n" );
		DumpBufferHex( Buffer, BytesInBuffer );
	}

	Mangle( Buffer, BytesInBuffer, Key, KEY_2 );

	BinToBase85( Buffer, BytesInBuffer, inString + OBSCURE_HEADER_CHARS );

	return 0;
}


int BadStringError( LPTSTR inString, int Debug )
{
	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Bad string" );

	if ( Debug )
		Qprintf( ERRHANDLE, L" (%s)", inString );

	Qprintf( ERRHANDLE, L"\r\n" );
	return 0;
}


DLLExports int WINAPI f_clarify( LPTSTR inString )
{
	if ( !inString )
		return -1;

	unsigned int Debug = CheckDebug();

	if ( wcslen( inString ) < OBSCURE_HEADER_CHARS ) {
		BadStringError( L"too short", Debug );
		return -1;
	}

	BYTE Header[OBSCURE_HEADER_SIZE], Buffer[OBSCURE_BUFFER_SIZE];

	if ( Base85ToBin( inString, Header, OBSCURE_HEADER_SIZE ) < 0 ) {
		BadStringError( L"error decoding header", Debug );
		return -1;
	}

	if ( Debug & 2 ) {
		Printf( L"Header -----\r\n" );
		DumpBufferHex( Header, OBSCURE_HEADER_SIZE );
	}


	DWORD Key = Header[HEADER_KEY_0] | ( Header[HEADER_KEY_1] << 8 ) | ( Header[HEADER_KEY_2] << 16 ) | ( Header[HEADER_KEY_3] << 24 );
	Key = Reverse( Key, 32 ) ^ OBSCURE_KEY;
	if ( Debug )
		Printf( L"   * Key = 0x%08x\r\n", Key );

	DWORD Check = Header[HEADER_CHECK_LO] | ( Header[HEADER_CHECK_HI] << 8 );
	Check ^= ( Key >> 16 );

	DWORD BytesInBuffer = Header[HEADER_LENGTH_LO] | ( Header[HEADER_LENGTH_HI] << 8 );
	BytesInBuffer = Reverse( BytesInBuffer ^ Check, 16 );

	if ( Debug ) {
		Printf( L"   * BytesInBuffer = %u\r\n", BytesInBuffer );
		Printf( L"   * Check = 0x%04x\r\n", Check );
	}

	unsigned int nb = Base85ToBin( inString + OBSCURE_HEADER_CHARS, Buffer, OBSCURE_BUFFER_SIZE );
	if ( Debug )
		Printf( L"   * nb = %d\r\n", nb );

	if ( ( nb < BytesInBuffer ) || ( nb > ( BytesInBuffer + 3 ) ) ) {
		memset( Buffer, 0, sizeof( Buffer ) );
		BadStringError( L"error decoding data", Debug );
		return -1;
	}

	Mangle( Buffer, BytesInBuffer, Key, KEY_2 );

	if ( Debug & 2 ) {
		Printf( L"Data -----\r\n" );
		DumpBufferHex( Buffer, BytesInBuffer );
	}

	if ( Checksum( Buffer, BytesInBuffer ) != Check ) {
		memset( Buffer, 0, sizeof( Buffer ) );
		BadStringError( L"checksum", Debug );
		return -1;
	}

	if ( UTF8toUTF16( Buffer, inString, BytesInBuffer ) < 0 ) {
		memset( Buffer, 0, sizeof( Buffer ) );
		Erase( inString );
		BadStringError( L"invalid UTF-8", Debug );
		return -1;
	}

	return 0;
}


DLLExports int WINAPI f_bintob85( LPTSTR String )
{
	if ( !String )
		return 1;

	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );

	unsigned int BufferHandle = 0;
	ULONG_PTR BufferSize = 0;
	BYTE *Buffer = NULL;

	unsigned int Start = 0, Length = 128;
	const unsigned int MaxLength = ( MAX_ARG_LEN - 1 ) * 4 / 5;


	if ( arglen[0] ) {
		BufferHandle = wcstoul( arg[0], NULL, 16 );

		if ( ( BufferHandle == 0 ) || ( BufferHandle & 0x07 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad handle: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing handle\r\n" );
		return -1;
	}

	Buffer = (BYTE *) BufferHandle;

	BufferSize = QueryMemSize( Buffer );

	if ( ( BufferSize == 0 ) || ( BufferSize == MAXUINT_PTR ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad handle: \"%s\"\r\n", arg[0] );
		return -1;
	}
	

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Start, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad start address: \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( Start >= BufferSize ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Start address too high: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}
	else
		Start = 0;


	if ( arglen[2] ) {
		if ( ParseInt( arg[2], &Length, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad length: \"%s\"\r\n", arg[2] );
			return -1;
		}

		if ( ( Length == 0 ) || ( Length > MaxLength ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Length must be 0 to %u\r\n", MaxLength );
			return -1;
		}
	}
	else
		Length = 128;


	if ( ( Start + Length ) > BufferSize )
		Length = (unsigned int) BufferSize - Start;


	if ( BinToBase85( Buffer + Start, Length, String ) < 1 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error encoding binary to base 85\r\n" );
		return -1;
	}

	return 0;
}


DLLExports int WINAPI f_b85tobin( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned int BufferHandle = 0;
	ULONG_PTR BufferSize = 0;
	BYTE *Buffer = NULL;
	wchar_t *ch = NULL;

	unsigned int Start = 0, Length = 0;
	size_t StringLen = 0;

	BYTE *TempBuffer = NULL;


	if ( iswdigit( String[0] ) ) {
		BufferHandle = wcstoul( String, &ch, 16 );

		if ( ( BufferHandle == 0 ) || ( BufferHandle & 0x07 ) ) {
			DisplayErrorHeader();

			if ( ch )
				*ch = 0x00;

			Qprintf( ERRHANDLE, L"Bad handle: \"%s\"\r\n", String );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing handle\r\n" );
		return -1;
	}


	if ( ( *ch != ',' ) && ( *ch != ';' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error (missing comma)\r\n" );
		return -1;
	}

	ch++;


	Buffer = (BYTE *) BufferHandle;

	BufferSize = QueryMemSize( Buffer );

	if ( ( BufferSize == 0 ) || ( BufferSize == MAXUINT_PTR ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad handle: \"%s\"\r\n", arg[0] );
		return -1;
	}


	if ( iswdigit( *ch ) ) {
		if ( ParseInt( ch, &Start, &ch ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad start address\r\n" );
			return -1;
		}

		if ( Start >= BufferSize ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Start address too high: %u\r\n", Start );
			return -1;
		}
	}


	if ( ( *ch != ',' ) && ( *ch != ';' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error (missing comma)\r\n" );
		return -1;
	}

	ch++;


	StringLen = wcslen( ch );

	if ( StringLen == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing string\r\n" );
		return -1;
	}
	else if ( StringLen >= MAX_ARG_LEN ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"String too long\r\n" );
		return -1;
	}


	Length = ( (unsigned int) StringLen + 4 ) / 5 * 4;

	TempBuffer = (BYTE*) malloc( Length );

	if ( TempBuffer == NULL ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error allocating temp buffer\r\n" );
		return -1;
	}


	Length = Base85ToBin( ch, TempBuffer, Length );

	if ( Length < 1 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error decoding base 85 to binary\r\n" );
		free( TempBuffer );
		return -1;
	}

	if ( Length > ( BufferSize - Start ) )
		Length = (unsigned int) BufferSize - Start;


	memcpy( Buffer + Start, TempBuffer, Length );

	free( TempBuffer );
	wsprintf( String, L"%u", Length );
	return 0;
}

