
//			PASSWORD command and _PASSWORD variable:


#ifndef PASSWORD_COMMAND_VER
#define PASSWORD_COMMAND_VER		1.0.0.0


#define PWD_MIN_LENGTH				4
#define PWD_MAX_LENGTH				1024
#define PWD_MAX_ALPHA				512
#define PWD_MAX_DIGITS				128
#define PWD_MAX_PUNCT				64
#define PWD_MAX_EXTENDED			64
#define PWD_MAX_SYLLABLES			64
#define PWD_MAX_PASSWORDS			1024
#define PWD_CASE_RANDOM				0
#define PWD_CASE_LOWER				1
#define PWD_CASE_UPPER				2
#define PWD_CASE_WORD				3
#define PWD_CASE_ALTERNATE			4
#define PWD_CASE_ALTERNATE_2		5
#define PWD_CASE_LEET				6
#define PWD_CASE_UNLEET				7

#define PWD_NUM_NAUGHTY_WORDS		25

#define PWD_NUM_1CONS_STARTS		19
#define PWD_NUM_2CONS_STARTS		17
#define PWD_NUM_1VOWELS				5
#define PWD_NUM_2VOWELS				15
#define PWD_NUM_1CONS_ENDS			16
#define PWD_NUM_2CONS_ENDS			29
#define PWD_NUM_PLURAL_ENDS			14



const wchar_t PasswordHelpText[] = L"Generate random passwords.\n\
\n\
PASSWORD /A:m,n /C:n /D:m,n /E:m,n /F /L:m,n /N:n /P:m,n /S:m,n /Y\r\n\
\n\
\t/A:  Alpha characters (min/max)\n\
\t/C:  Case:  0 random, 1 lower, 2 upper, 3 word, 5 alternate\n\
\t/D:  Digits (min/max)\n\
\t/E:  Extended characters (min/max)\n\
\t/F   make First character a letter if possible\n\
\t/L:  total Length (min/max)\n\
\t/N:  Number of passwords to generate\n\
\t/P:  Punctuation characters (min/max)\n\
\t/S:  Syllables (min/max)\n\
\t/Y   copy password to clipboard\n\
\n\
By default, passwords 7-10 characters long will be generated.  This\n\
command also saves its parameters for use by the _PASSWORD variable. \n";



const wchar_t Naughty[PWD_NUM_NAUGHTY_WORDS][5] = { L"cock", L"coon", L"crap", L"cumm", L"cunt", L"dick", L"dyke",
	L"fags", L"fuck", L"gook", L"jizz", L"kike", L"kock", L"koon", L"kumm", L"kunt", L"nigs", L"phag",
	L"phuk", L"piss", L"rape", L"shit", L"slut", L"spik", L"twat" };



const wchar_t Syl1ConsStart[PWD_NUM_1CONS_STARTS][2]	= { L"b", L"c", L"d", L"f", L"g", L"h", L"j", L"k", L"l", L"m", L"n", L"p", L"r", L"s", L"t", L"v", L"w", L"y", L"z" };
const wchar_t Syl2ConsStart[PWD_NUM_2CONS_STARTS][3]	= { L"br", L"ch", L"cr", L"dr", L"fr", L"gr", L"kr", L"ph", L"pr", L"qu", L"sh", L"sl", L"sp", L"st", L"th", L"tr", L"wr" };
const wchar_t Syl1Vowel[PWD_NUM_1VOWELS][2]				= { L"a", L"e", L"i", L"o", L"u" };
const wchar_t Syl2Vowel[PWD_NUM_2VOWELS][3]				= { L"ai", L"ao", L"au", L"ea", L"ee", L"ei", L"eo", L"ia", L"ie", L"io", L"oa", L"oe", L"oi", L"oo", L"ou" };
const wchar_t Syl1ConsEnd[PWD_NUM_1CONS_ENDS][2]		= { L"b", L"d", L"f", L"g", L"k", L"l", L"m", L"n", L"p", L"r", L"s", L"t", L"v", L"w", L"x", L"z" };
const wchar_t Syl2ConsEnd[PWD_NUM_2CONS_ENDS][3]		= { L"bb", L"ck", L"dd", L"ft", L"gh", L"ld", L"lf", L"lk", L"ll", L"lm", L"lp", L"lt", L"mm", L"nt", L"rd", L"rf", L"rk", L"rl", L"rm", L"rn", L"rp", L"rr", L"rt", L"ss", L"st", L"th", L"tz", L"xx", L"zz" };
const wchar_t SylPluralEnd[PWD_NUM_PLURAL_ENDS][3]		= { L"bs", L"ds", L"gs", L"hs", L"ks", L"ls", L"ms", L"ns", L"ps", L"rs", L"ts", L"vs", L"ws", L"ys" };



int OffensiveSyllable( LPTSTR Syllable )
{
	if ( !Syllable )
		return 0;

	for ( int i = 0; i < PWD_NUM_NAUGHTY_WORDS; i++ )
		if ( !_wcsicmp( Syllable, Naughty[i] ) )
			return 1;

	return 0;
}



int MakeRandomSyllable( LPTSTR outString )
{
	if ( outString == NULL )
		return -1;

	do {
		switch ( GetRand( 10 ) ) {
			case 0 :								//  cvcc
			case 1 :
				wcscpy_s( outString, SHORT_BUF_LEN, Syl1ConsStart[ GetRand( PWD_NUM_1CONS_STARTS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl1Vowel[ GetRand( PWD_NUM_1VOWELS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl2ConsEnd[ GetRand( PWD_NUM_2CONS_ENDS ) ] );
				break;

			case 2 :								//  cvvc
			case 3 :
				wcscpy_s( outString, SHORT_BUF_LEN, Syl1ConsStart[ GetRand( PWD_NUM_1CONS_STARTS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl2Vowel[ GetRand( PWD_NUM_2VOWELS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl1ConsEnd[ GetRand( PWD_NUM_1CONS_ENDS ) ] );
				break;

			case 4 :								//  ccvc
			case 5 :
				wcscpy_s( outString, SHORT_BUF_LEN, Syl2ConsStart[ GetRand( PWD_NUM_2CONS_STARTS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl1Vowel[ GetRand( PWD_NUM_1VOWELS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl1ConsEnd[ GetRand( PWD_NUM_1CONS_ENDS ) ] );
				break;

			case 6 :								//  ccvv
				wcscpy_s( outString, SHORT_BUF_LEN, Syl2ConsStart[ GetRand( PWD_NUM_2CONS_STARTS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl2Vowel[ GetRand( PWD_NUM_2VOWELS ) ] );
				break;

			case 7 :								//  cvcs
				wcscpy_s( outString, SHORT_BUF_LEN, Syl1ConsStart[ GetRand( PWD_NUM_1CONS_STARTS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl1Vowel[ GetRand( PWD_NUM_1VOWELS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, SylPluralEnd[ GetRand( PWD_NUM_PLURAL_ENDS ) ] );
				break;

			case 8 :								//  cvce
				wcscpy_s( outString, SHORT_BUF_LEN, Syl1ConsStart[ GetRand( PWD_NUM_1CONS_STARTS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl1Vowel[ GetRand( PWD_NUM_1VOWELS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl1ConsEnd[ GetRand( PWD_NUM_1CONS_ENDS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, L"e" );
				break;

			case 9 :								//  ccvy
				wcscpy_s( outString, SHORT_BUF_LEN, Syl2ConsStart[ GetRand( PWD_NUM_2CONS_STARTS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, Syl1Vowel[ GetRand( PWD_NUM_1VOWELS ) ] );
				wcscat_s( outString, SHORT_BUF_LEN, L"y" );
				break;
		}

	} while ( OffensiveSyllable( outString ) );

	outString[0] = towupper( outString[0] );
	return 0;
}



unsigned int PwdMinLength = 7, PwdMaxLength = 10;
unsigned int PwdMinAlpha = 3, PwdMinDigits = 2, PwdMinPunct = 1, PwdMinExtended = 0, PwdMinSyllables = 0;
unsigned int PwdMaxAlpha = PWD_MAX_ALPHA, PwdMaxDigits = PWD_MAX_DIGITS, PwdMaxPunct = PWD_MAX_PUNCT, PwdMaxExtended = 0, PwdMaxSyllables = 0;
unsigned int PwdRecase = PWD_CASE_RANDOM;
BOOL PwdBeginsWithAlpha = FALSE;




int RandomPassword( LPTSTR outString )
{
	if ( !outString )
		return -1;


	wchar_t temp[PWD_MAX_LENGTH + 1]		= L"";

	wchar_t PunctChars[SHORT_BUF_LEN]		= L"!@#$*()-_=+;:,./?{}~";
	wchar_t ExtendedChars[SHORT_BUF_LEN]	= L"";
	
	int NumPunctChars						= 0;
	int NumExtendedChars					= 0;
	unsigned int Chars						= 0;
	unsigned int Alpha						= PwdMinAlpha;
	unsigned int Digits						= PwdMinDigits;
	unsigned int Punct						= PwdMinPunct;
	unsigned int Extended					= PwdMinExtended;
	unsigned int Syllables					= PwdMinSyllables;

	BOOL LastWasLetter						= FALSE;
	static BOOL Toggle						= FALSE;


	ExtendedChars[0] = '\0';
	GetEnvironmentVariable( L"PUNCTUATION_CHARACTERS", ExtendedChars, SHORT_BUF_LEN );
	if ( ExtendedChars[0] )
		wcscpy_s( PunctChars, SHORT_BUF_LEN, ExtendedChars );

	ExtendedChars[0] = '\0';
	GetEnvironmentVariable( L"EXTENDED_CHARACTERS", ExtendedChars, SHORT_BUF_LEN );

	NumPunctChars    = (int) wcslen( PunctChars );
	NumExtendedChars = (int) wcslen( ExtendedChars );


	outString[0] = '\0';


	Chars = PwdMinLength + GetRand( PwdMaxLength - PwdMinLength + 1 );

	if ( PwdMinSyllables && !PwdMaxAlpha && !PwdMaxDigits && !PwdMaxPunct && !PwdMaxExtended )
		Chars = ( 4 * PwdMinSyllables ) + ( 4 * GetRand( PwdMaxSyllables - PwdMinSyllables + 1 ) );


	while ( ( Alpha + Digits + Punct + Extended + ( 4 * Syllables ) ) < Chars ) {
		switch ( GetRand( 12 ) ) {
			case 0 :
				if ( Extended < PwdMaxExtended )
					Extended++;
				break;

			case 1 :
			case 2 :
			case 3 :
				if ( Digits < PwdMaxDigits )
					Digits++;
				break;

			case 4 :
			case 5 :
				if ( Punct < PwdMaxPunct )
					Punct++;
				break;

			case 6 :
				if ( ( Syllables < PwdMaxSyllables ) && ( ( Alpha + Digits + Punct + Extended + ( 4 * (  Syllables + 1 ) ) ) <= Chars ) )
					Syllables++;
				break;

			default :
				if ( Alpha < PwdMaxAlpha )
					Alpha++;
		}
	}


	unsigned int i = 0;

	for ( unsigned int j = 0; j < Alpha; j++ )
		temp[i++] = 'A';

	for ( unsigned int j = 0; j < Digits; j++ )
		temp[i++] = 'D';

	for ( unsigned int j = 0; j < Punct; j++ )
		temp[i++] = 'P';

	for ( unsigned int j = 0; j < Extended; j++ )
		temp[i++] = 'E';

	for ( unsigned int j = 0; j < Syllables; j++ )
		temp[i++] = 'S';

	temp[i] = '\0';

	for ( unsigned int j = 0; j < ( 4 * i ); j++ ) {		//  shuffle the characters a few times
		int m = j % i;
		int n = GetRand( i - 1 );
		if ( m == n )
			n = ( n + 1 ) % Chars;

		wchar_t ch = temp[m];
		temp[m] = temp[n];
		temp[n] = ch;
	}

	if ( PwdBeginsWithAlpha ) {
		int k = -1;
		for ( unsigned int j = 0; j < i; j++ )
			if ( ( k == -1 ) && ( ( temp[j] == 'A' ) || ( temp[j] == 'S' ) ) )
				k = j;

		if ( k > 0 ) {
			wchar_t ch = temp[0];
			temp[0] = temp[k];
			temp[k] = ch;
		}
	}


	i = 0;
	outString[i] = '\0';
	
	for ( unsigned int j = 0; temp[j]; j++ ) {
		switch ( temp[j] ) {

			case 'D' :
				outString[i++] = '0' + GetRand( 10 );
				break;

			case 'A' :
				outString[i++] = 'A' + GetRand( 26 );
				break;

			case 'P' :
				outString[i++] = PunctChars[ GetRand( NumPunctChars ) ];
				break;

			case 'E' :
				if ( NumExtendedChars )
					outString[i++] = ExtendedChars[ GetRand( NumExtendedChars ) ];
				else
					outString[i++] = 0x00c0 + GetRand( 0x0040 );

				break;

			case 'S' :
				MakeRandomSyllable( outString + i );
				i += 4;
				break;
		}

		outString[i] = '\0';
	}

	for ( int i = 0; outString[i]; i++ ) {
		if ( ( ( outString[i] >= 'A' ) && ( outString[i] <= 'Z' ) ) || ( ( outString[i] >= 'a' ) && ( outString[i] <= 'z' ) ) ) {
			switch ( PwdRecase ) {
				case PWD_CASE_RANDOM :
					if ( ( GetRand( 1024 ) & 15 ) < 10 )
						outString[i] = towlower( outString[i] );
					else
						outString[i] = towupper( outString[i] );
					break;

				case PWD_CASE_LOWER :
					outString[i] = towlower( outString[i] );
					break;

				case PWD_CASE_UPPER :
					outString[i] = towupper( outString[i] );
					break;

				case PWD_CASE_WORD :
					if ( LastWasLetter )
						outString[i] = towlower( outString[i] );
					else
						outString[i] = towupper( outString[i] );
					break;

				case PWD_CASE_ALTERNATE :
					if ( i & 1 )
						outString[i] = towlower( outString[i] );
					else
						outString[i] = towupper( outString[i] );
					break;

				case PWD_CASE_ALTERNATE_2 :
					if ( Toggle )
						outString[i] = towlower( outString[i] );
					else
						outString[i] = towupper( outString[i] );

					Toggle = !Toggle;
					break;

				case PWD_CASE_LEET :
					if ( IsVowel( outString[i] ) )
						outString[i] = towlower( outString[i] );
					else
						outString[i] = towupper( outString[i] );
					break;

				case PWD_CASE_UNLEET :
					if ( IsVowel( outString[i] ) )
						outString[i] = towupper( outString[i] );
					else
						outString[i] = towlower( outString[i] );
					break;
			}

			LastWasLetter = TRUE;
		}
		else
			LastWasLetter = FALSE;
	}


	return 0;
}



DLLExports int WINAPI password( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t PasswordBuffer[PWD_MAX_LENGTH + 1] = L"";

	unsigned int MinLength		= 0,	MaxLength		= 0;
	unsigned int MinAlpha		= 0,	MaxAlpha		= 0;
	unsigned int MinDigits		= 0,	MaxDigits		= 0;
	unsigned int MinPunct		= 0,	MaxPunct		= 0;
	unsigned int MinExtended	= 0,	MaxExtended		= 0;
	unsigned int MinSyllables	= 0,	MaxSyllables	= 0;

	BOOL BeginsWithAlpha		= FALSE;
	BOOL SpecifiedParts			= FALSE;
	BOOL SpecifiedAlpha			= FALSE;
	BOOL SpecifiedDigits		= FALSE;
	BOOL SpecifiedPunct			= FALSE;
	BOOL SpecifiedLength		= FALSE;
	BOOL CopyToClipboard		= FALSE;
	BOOL Debug					= FALSE;

	unsigned int NumPasswords	= 1;
	unsigned int Case			= PWD_CASE_RANDOM;

	wchar_t *ClipBuffer			= NULL;
	wchar_t *ch					= NULL;
	size_t BufferSize			= 0;

	int rv						= 0;




	ParseArgs( inString, PARSE_FORCE_UPPERCASE | PARSE_SLASHES_KLUDGE | PARSE_BREAK_SPACES );


	
	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( PasswordHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( PASSWORD_COMMAND_VER ) );
			else if ( !wcscmp( arg[i], L"/$" ) )
				Debug = TRUE;
			else if ( !wcsncmp( arg[i], L"/A", 2 ) ) {
				if ( ParseRange( arg[i], &MinAlpha, &MaxAlpha, 0, PWD_MAX_ALPHA, L"MinAlpha", L"MaxAlpha" ) )
					return 1;

				SpecifiedParts = SpecifiedAlpha = TRUE;
			}
			else if ( !wcsncmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Case, NULL ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad case: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/D", 2 ) ) {
				if ( ParseRange( arg[i], &MinDigits, &MaxDigits, 0, PWD_MAX_DIGITS, L"MinDigits", L"MaxDigits" ) )
					return 1;

				SpecifiedParts = SpecifiedDigits = TRUE;
			}
			else if ( !wcsncmp( arg[i], L"/E", 2 ) ) {
				if ( ParseRange( arg[i], &MinExtended, &MaxExtended, 0, PWD_MAX_EXTENDED, L"MinExtended", L"MaxExtended" ) )
					return 1;

				SpecifiedParts = TRUE;
			}
			else if ( !wcscmp( arg[i], L"/F" ) )
				BeginsWithAlpha = TRUE;
			else if ( !wcsncmp( arg[i], L"/L", 2 ) ) {
				if ( ParseRange( arg[i], &MinLength, &MaxLength, PWD_MIN_LENGTH, PWD_MAX_LENGTH, L"MinLength", L"MaxLength" ) )
					return 1;

				SpecifiedLength = TRUE;
			}
			else if ( !wcsncmp( arg[i], L"/N", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &NumPasswords, NULL ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad NumPasswords \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( NumPasswords > PWD_MAX_PASSWORDS ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"NumPasswords must be 0 - %u\r\n", PWD_MAX_PASSWORDS );
					return 1;
				}
			}
			else if ( !wcsncmp( arg[i], L"/P", 2 ) ) {
				if ( ParseRange( arg[i], &MinPunct, &MaxPunct, 0, PWD_MAX_PUNCT, L"MinPunct", L"MaxPunct" ) )
					return 1;

				SpecifiedParts = SpecifiedPunct = TRUE;
			}
			else if ( !wcsncmp( arg[i], L"/S", 2 ) ) {
				if ( ParseRange( arg[i], &MinSyllables, &MaxSyllables, 0, PWD_MAX_SYLLABLES, L"MinSyllables", L"MaxSyllables" ) )
					return 1;

				SpecifiedParts = TRUE;
			}
			else if ( !wcscmp( arg[i], L"/Y" ) )
				CopyToClipboard = TRUE;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( SpecifiedParts ) {
		if ( !MinLength )
			MinLength = 7;

		if ( MinLength < ( MinAlpha + MinDigits + MinPunct + MinExtended + ( 4 * MinSyllables ) ) )
			MinLength = MinAlpha + MinDigits + MinPunct + MinExtended + ( 4 * MinSyllables );

		if ( MinLength < PWD_MIN_LENGTH )
			MinLength = PWD_MIN_LENGTH;

		if ( MinLength > ( MaxAlpha + MaxDigits + MaxPunct + MaxExtended + ( 4 * MaxSyllables ) ) ) {
			int ld = MinLength - MaxAlpha - MaxDigits - MaxPunct - MaxExtended - ( 4 * MaxSyllables );

			if ( !SpecifiedAlpha && !SpecifiedDigits && !SpecifiedPunct ) {
				MinAlpha  = ld / 3;
				MinDigits = ld / 4;
				MinPunct  = ld / 5;
				MaxAlpha  = ( 2 * ld / 3 ) + 1;
				MaxDigits = ( ld / 3 ) + 1;
				MaxPunct  = ( ld / 4 ) + 1;
			}
			else if ( !SpecifiedAlpha && !SpecifiedDigits ) {
				MinAlpha  = ld / 2;
				MinDigits = ld / 3;
				MaxAlpha  = ( 3 * ld / 4 ) + 1;
				MaxDigits = ( ld / 2 ) + 1;
			}
			else if ( !SpecifiedAlpha && !SpecifiedPunct ) {
				MinAlpha  = ld / 2;
				MinPunct  = ld / 3;
				MaxAlpha  = ( 3 * ld / 4 ) + 1;
				MaxPunct  = ( ld / 2 ) + 1;
			}
			else if ( !SpecifiedDigits && !SpecifiedPunct ) {
				MinDigits  = ld / 2;
				MinPunct   = ld / 3;
				MaxDigits  = ( 3 * ld / 4 ) + 1;
				MaxPunct   = ( ld / 2 ) + 1;
			}
			else if ( !SpecifiedAlpha )
				MinAlpha = MaxAlpha = ld;
			else if ( !SpecifiedDigits )
				MinDigits = MaxDigits = ld;
			else if ( !SpecifiedPunct )
				MinPunct = MaxPunct = ld;
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Minimum length %u is greater than the sum of the maximums\r\n", MinLength );
				return -1;
			}

			if ( MaxAlpha > PWD_MAX_ALPHA )
				MaxAlpha = PWD_MAX_ALPHA;

			if ( MaxDigits > PWD_MAX_DIGITS )
				MaxDigits = PWD_MAX_DIGITS;

			if ( MaxPunct > PWD_MAX_PUNCT )
				MaxPunct = PWD_MAX_PUNCT;

			MaxLength = MinLength;
		}

		if ( !MaxLength )
			MaxLength = MinLength + ( MinLength / 2 );

		if ( MaxLength > ( MaxAlpha + MaxDigits + MaxPunct + MaxExtended + ( 4 * MaxSyllables ) ) )
			MaxLength = MaxAlpha + MaxDigits + MaxPunct + MaxExtended + ( 4 * MaxSyllables );

		if ( MinLength > MaxLength )
			MaxLength = MinLength;

		if ( MaxLength > PWD_MAX_LENGTH )
			MaxLength = PWD_MAX_LENGTH;
	}
	else {
		if ( !MinLength )
			MinLength = 7;

		if ( !MaxLength )
			MaxLength = MinLength + ( MinLength / 2 );

		if ( MaxLength > PWD_MAX_LENGTH )
			MaxLength = PWD_MAX_LENGTH;

		MinAlpha  = MinLength / 3;
		MinDigits = MinLength / 4;
		MinPunct  = MinLength / 5;

		MaxAlpha  = ( 2 * MaxLength / 3 ) + 1;
		if ( MaxAlpha > PWD_MAX_ALPHA )
			MaxAlpha = PWD_MAX_ALPHA;

		MaxDigits = ( MaxLength / 3 ) + 1;
		if ( MaxDigits > PWD_MAX_DIGITS )
			MaxDigits = PWD_MAX_DIGITS;

		MaxPunct  = ( MaxLength / 4 ) + 1;
		if ( MaxPunct > PWD_MAX_PUNCT )
			MaxPunct = PWD_MAX_PUNCT;

		MinExtended  = MaxExtended  = 0;
		MinSyllables = MaxSyllables = 0;

		if ( MinLength >= 7 ) {							//  let's do syllables!
			MaxSyllables = ( MinLength + 1 ) / 8;
			MinSyllables = MaxSyllables / 3;

			MinPunct  = ( MinLength - ( 4 * MaxSyllables ) ) / 3;
			MinDigits = ( MinLength - ( 4 * MaxSyllables ) - MinPunct ) / 2;
			MinAlpha  = ( MinLength - ( 4 * MaxSyllables ) - MinPunct - MinDigits );
		}
	}


	if ( Debug ) {
		Printf( L"   * Length    = %u - %u\r\n", MinLength, MaxLength );
		Printf( L"   * Alpha     = %u - %u\r\n", MinAlpha, MaxAlpha );
		Printf( L"   * Digits    = %u - %u\r\n", MinDigits, MaxDigits );
		Printf( L"   * Punct     = %u - %u\r\n", MinPunct, MaxPunct );
		Printf( L"   * Extended  = %u - %u\r\n", MinExtended, MaxExtended );
		Printf( L"   * Syllables = %u - %u\r\n\r\n", MinSyllables, MaxSyllables );
	}

	PwdMinLength		= MinLength;
	PwdMaxLength		= MaxLength;
	PwdMinAlpha			= MinAlpha;
	PwdMaxAlpha			= MaxAlpha;
	PwdMinDigits		= MinDigits;
	PwdMaxDigits		= MaxDigits;
	PwdMinPunct			= MinPunct;
	PwdMaxPunct			= MaxPunct;
	PwdMinExtended		= MinExtended;
	PwdMaxExtended		= MaxExtended;
	PwdMinSyllables		= MinSyllables;
	PwdMaxSyllables		= MaxSyllables;
	PwdRecase			= Case;
	PwdBeginsWithAlpha	= BeginsWithAlpha;


	if ( !NumPasswords )
		return 0;

	if ( CopyToClipboard ) {
		BufferSize = NumPasswords * ( MaxLength + 3 ) + 2;
		ClipBuffer = (wchar_t *) malloc( BufferSize * sizeof( wchar_t ) );

		if ( !ClipBuffer ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t allocate buffer for clipboard!\r\n" );
			rv = 2;
		}
		else
			ClipBuffer[0] = '\0';
	}


	for ( unsigned int i = 0; i < NumPasswords; i++ ) {
		RandomPassword( PasswordBuffer );

		Printf( L"%s\r\n", PasswordBuffer );

		if ( ClipBuffer ) {
			wcscat_s( ClipBuffer, BufferSize, PasswordBuffer );
			if ( NumPasswords > 1 )
				wcscat_s( ClipBuffer, BufferSize, L"\r\n" );
		}
	}


	if ( ClipBuffer ) {
		CopyTextToClipboard( ClipBuffer, (int) wcslen( ClipBuffer ) );
		free( ClipBuffer );
	}

	return rv;
}



DLLExports int WINAPI _password( LPTSTR outString )
{
	if ( !outString )
		return 1;

	outString[0] = '\0';
	RandomPassword( outString );
	return 0;
}


#endif	//	PASSWORD_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-28
				Adds a version number to this source file.  Help text tweaks.

*/
