
//			REPLACETEXT command:


#ifndef REPLACETEXT_COMMAND_VER
#define REPLACETEXT_COMMAND_VER		1.0.1.1


#define REPLACETEXT_HEADER			1
#define REPLACE_ENTRY_FIXCASE		1
#define REPLACE_ENTRY_WORDSONLY		2



const wchar_t ReplaceTextHelpText[] = L"Replace strings in text from a file.\n\
\n\
REPLACETEXT /C /CP:n /H /P /R:from:to /S /W /X:from:to filename....\r\n\
\n\
\t/C    character escapes (affects following /R: and /X:)\n\
\t/CP:  interpret OEM text as code page n\n\
\t/H    display filenames\n\
\t/P    page output\n\
\t/R:   old and replacement text\n\
\t/S    search in subdirectories for matching files\n\
\t/W    whole words only (affects following /R: and /X:)\n\
\t/X:   old and replacement text (do not capitalize)\n\
\n\
Input filenames may be specified on the command line, or text may\n\
be piped or redirected into REPLACETEXT.  You must specify at\n\
least one /R: or /X:; you may have as many as you like.\n";



struct ReplaceEntry {
	ReplaceEntry *Next;
	unsigned int Flags;
	wchar_t FromString;
};


struct _ReplaceWordsArgs {
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
	ReplaceEntry *FirstEntry;
};




int ReplaceWords( LPCTSTR Filename, int Flags, void *Args )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	_MMFile MMFile;
	wchar_t LineBuffer[BIG_LINE_LEN]	= L"";
	wchar_t temp[BIG_LINE_LEN]			= L"";

	_ReplaceWordsArgs *args				= (_ReplaceWordsArgs*) Args;
	ReplaceEntry *FirstEntry			= args->FirstEntry;

	wchar_t *OldText					= NULL;
	wchar_t *NewText					= NULL;
	int OldLen							= 0;
	int NewLen							= 0;
	wchar_t *SearchIndex				= NULL;
	wchar_t *Found						= NULL;
	int i								= 0;
	size_t j							= 0;
	wchar_t ch							= '\0';
	BOOL ForceFirst						= FALSE;
	BOOL ForceAll						= FALSE;
	int LetterCount						= 0;


	int rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, args->CodePage, args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	if ( Flags & REPLACETEXT_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	CL_HookBreak();


	while ( MMFile.Index <= MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, MMGL_FLAG_SPLIT_LONG_LINES ) )
			break;

		FirstEntry = args->FirstEntry;

		while ( FirstEntry ) {
			OldText	= &FirstEntry->FromString;
			OldLen	= (int) wcslen( OldText );
			NewText	= OldText + OldLen + 1;
			NewLen	= (int) wcslen( NewText );

			SearchIndex = LineBuffer;

			do {
				Found = Stristr( SearchIndex, OldText );
				if ( !Found )
					break;

				if ( FirstEntry->Flags & REPLACE_ENTRY_WORDSONLY ) {
					if ( Found > LineBuffer ) {
						if ( iswalnum( *( Found - 1 ) ) ) {
							SearchIndex = Found + 1;
							continue;
						}
					}

					if ( iswalnum( *( Found + OldLen ) ) ) {
						SearchIndex = Found + 1;
						continue;
					}
				}

				ForceFirst = ForceAll = FALSE;

				if ( FirstEntry->Flags & REPLACE_ENTRY_FIXCASE ) {
					wchar_t *ch = Found;

					while ( *ch && !iswalpha( *ch ) )
						ch++;

					if ( iswupper( *ch ) ) {
						ForceFirst = TRUE;

						ch++;

						while ( *ch && !iswalpha( *ch ) )
							ch++;

						if ( iswupper( *ch ) )
							ForceAll = TRUE;
					}
				}

				for ( i = 0; i < ( Found - LineBuffer ); i++ )
					temp[i] = LineBuffer[i];

				temp[i]		= '\0';
				LetterCount	= 0;

				for ( j = 0; NewText[j]; j++ ) {
					ch = NewText[j];

					if ( iswalpha( ch ) ) {
						LetterCount++;

						if ( ForceAll || ( ( LetterCount == 1 ) && ForceFirst ) )
							ch = towupper( ch );
					}

					if ( i < ( BIG_LINE_LEN - 1 ) ) {
						temp[i++] = ch;
						temp[i] = '\0';
					}
				}

				SearchIndex = Found + NewLen;

				for ( j = ( Found - LineBuffer ) + OldLen; LineBuffer[j]; j++ ) {
					if ( i < ( BIG_LINE_LEN - 1 ) ) {
						temp[i++] = LineBuffer[j];
						temp[i] = '\0';
					}
				}

				wcscpy_s( LineBuffer, BIG_LINE_LEN, temp );

			} while ( TRUE );

			FirstEntry = FirstEntry->Next;
		}

		QPuts( LineBuffer );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}



int hexdigval( wchar_t ch )
{
	if ( iswdigit( ch ) )
		return ( ch - '0' );
	else if ( ( towupper( ch ) >= 'A' ) && ( towupper( ch ) <= 'F' ) )
		return ( towupper(ch ) - 'A' + 10 );
	else
		return 0;
}


int ExpandCharEscapes( LPTSTR String )
{
	if ( !String )
		return -1;


	wchar_t temp[MAX_ARG_LEN]		= L"";

	int ch			= '\0';
	size_t j		= 0;
	size_t q		= 0;
	int k			= 0;
	long int Value	= 0;


	for ( size_t i = 0; String[i]; i++ ) {

		ch = String[i];

		if ( ch == '\\' ) {

			q = i + 1;

			if ( iswdigit( String[q] ) ) {

				while ( iswdigit( String[q] ) ) {
					Value = ( Value * 10 ) + ( String[q] - '0' );
					q++;
				}
			}
			else if ( ( towupper( String[q] ) == 'X' ) && iswxdigit( String[q + 1] ) ) {

				q++;

				while ( iswxdigit( String[q] ) ) {
					Value = ( Value * 16 ) + hexdigval( String[q] );
					q++;
				}
			}
			else if ( String[q] == '\\' )
				i++;


			if ( ( Value > 0 ) && ( Value <= 0xffff ) && q ) {
				ch = Value;
				i = q - 1;
				k++;
			}
		}


		if ( j < ( MAX_ARG_LEN - 1 ) )
			temp[j++] = ch;
	}

	temp[j] = '\0';


	if ( k )
		wcscpy_s( String, MAX_ARG_LEN, temp );

	return k;
}



DLLExports int WINAPI replacetext( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t OldText[MAX_ARG_LEN]		= L"";
	wchar_t NewText[MAX_ARG_LEN]		= L"";

	int Flags					= 0;
	int Recurse					= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth		= 0;
	unsigned int MaxDepth		= UINT_MAX;
	DWORD AttribMask			= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues			= 0;
	void *FirstRange			= NULL;
	BOOL NoHighlight			= FALSE;
	unsigned int Pause			= 0;

	BOOL FilenameSpecified		= FALSE;
	BOOL StdinSpecified			= FALSE;
	BOOL WriteBOM				= TRUE;
	BOOL WordsOnly				= FALSE;
	BOOL CharEscapes			= FALSE;
	ReplaceEntry *FirstEntry	= NULL;
	ReplaceEntry *LastEntry		= NULL;
	int rv						= 0;
	int fhrv					= 0;

	_ReplaceWordsArgs Args		= { 0, 0, NULL };

	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( (void**) &FirstEntry );
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( ReplaceTextHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( REPLACETEXT_COMMAND_VER ) );
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				CharEscapes = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				unsigned int Val = 0;

				if ( ParseInt( arg[i] + j, &Val, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				CharEscapes = ( Val != 0 );
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = TRUE;
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= REPLACETEXT_HEADER;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/R", 2 ) || !_wcsnicmp( arg[i], L"/X", 2 ) ) {

				wchar_t *Delim			= arg[i] + 2;
				wchar_t *ch				= Delim + 1;
				unsigned int NewFlags	= 0;
				size_t TextLen			= 0;

				OldText[0]		= '\0';
				NewText[0]		= '\0';

				if ( towupper( arg[i][1] ) == 'R' )
					NewFlags |= REPLACE_ENTRY_FIXCASE;

				if ( WordsOnly )
					NewFlags |= REPLACE_ENTRY_WORDSONLY;


				if ( ( *Delim == '\0' ) || !iswpunct( *Delim ) || ( *ch == '\0' ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				for ( unsigned int j = 0; ( *ch != *Delim ) && *ch; ch++ ) {
					if ( j < ( MAX_ARG_LEN - 1 ) ) {
						OldText[j++]	= *ch;
						OldText[j]		= '\0';
					}
				};

				if ( OldText[0] == '\0' ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"From text cannot be empty: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( *ch != *Delim ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				ch++;

				for ( unsigned int j = 0; *ch; ch++ ) {
					if ( j < ( MAX_ARG_LEN - 1 ) ) {
						NewText[j++]	= *ch;
						NewText[j]		= '\0';
					}
				};

				if ( CharEscapes ) {
					ExpandCharEscapes( OldText );
					ExpandCharEscapes( NewText );
				}

				TextLen = wcslen( OldText ) + wcslen( NewText ) + 1;

				ReplaceEntry *NewEntry = NULL;
				NewEntry = (ReplaceEntry *) malloc( sizeof( ReplaceEntry ) + ( sizeof( wchar_t ) * TextLen ) );

				if ( !NewEntry ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Error allocating memory for replacement entry\r\n" );
					goto ErrorExit;
				}

				NewEntry->Next  = NULL;
				NewEntry->Flags = NewFlags;
				wcscpy_s( &(NewEntry->FromString), TextLen + 1, OldText );
				wcscpy_s( &(NewEntry->FromString) + wcslen( OldText ) + 1, wcslen( NewText ) + 1, NewText );


				if ( LastEntry )
					LastEntry->Next = NewEntry;

				LastEntry = NewEntry;

				if ( !FirstEntry )
					FirstEntry = NewEntry;
			}
			else if ( !_wcsicmp( arg[i], L"/W" ) )
				WordsOnly = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				unsigned int Val = 0;

				if ( ParseInt( arg[i] + j, &Val, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				WordsOnly = ( Val != 0 );
			}
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}
		}
	}

	if ( !FirstEntry ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No replacements specified\r\n" );
		goto ErrorExit;
	}


	if ( Pause )
		CL_SetPauseLines( Pause );
	 
	CL_PickHighlightColor( NoHighlight );


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	Args.FirstEntry = FirstEntry;


	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = ReplaceWords( (LPTSTR) StdinFilename, Flags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, ReplaceWords, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( (void**) &FirstEntry );
	FH_FreeLinkedList( &FirstRange );
	return 0;

ErrorExit:

	FH_FreeLinkedList( (void**) &FirstEntry );
	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort :

	FH_FreeLinkedList( (void**) &FirstEntry );
	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	REPLACETEXT_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				ReplaceWords() filename is defined as a constant:  fix for mmfiles 1.4.4.2

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
