
//			SAVEARRAY and LOADARRAY commands:


#ifndef SAVEARRAY_COMMAND_VER
#define SAVEARRAY_COMMAND_VER		1.0.0.0


#define MAKESAFE_ESCAPE_1			0xa4
#define MAKESAFE_ESCAPE_2			0xac
#define SAVEARRAY_LINE_LEN			MAX_ARG_LEN + 64



const wchar_t LoadArrayHelpText[] = L"Load data from a file into an array.\n\
\n\
LOADARRAY filename [arrayname]\r\n\
\n\
The filename is required; the arrayname is optional.  The array \n\
will be created to the correct size.\n";



const wchar_t SaveArrayHelpText[] = L"Save data from an array to a file.\n\
\n\
SAVEARRAY /O /P /X:m,n /Y:m,n /Z:m,n /W:m,n arrayname filename \r\n\
\n\
\t/O   Overwrite an existing file\n\
\t/P   save Partial array as if whole\r\n\
\t/X:  save only X index m through n\n\
\t/Y:  save only Y index m through n\n\
\t/Z:  save only Z index m through n\n\
\t/W:  save only W index m through n\n\
\n\
You must specify both the arrayname and the filename.\n";


const wchar_t SaveArrayHeaderString[] = L"SaveArray";


wchar_t svSetDos[32] = L"";
wchar_t fxSetDos[32] = L"";



int PrepSetDos()
{
	wcscpy_s(svSetDos, 32, L"%_expansion");

	ExpandVariables(svSetDos, 1);
	if ( svSetDos[0] == '\0' )
		wcscpy_s(svSetDos, 32, L"0");

	CallInternal( L"setdos", L" /x-1478+3");
	return 0;
}


int ResetSetDos()
{
	wcscpy_s(fxSetDos, 32, L"/X");
	wcscat_s(fxSetDos, 32, svSetDos);
	CallInternal( L"setdos", fxSetDos );
	return 0;
}


int GetArrayInfo( LPTSTR ArrayName, int Info, unsigned int *outValue )
{
	if ( ( ArrayName == NULL ) || ( outValue == NULL ) )
		return -1;

	if ( ArrayName[0] == 0x00 )
		return -1;

	wchar_t temp[SHORT_BUF_LEN] = L"%@arrayinfo[";
	wchar_t InfoString[SHORT_BUF_LEN] = L"";
	int Value = 0;
	wchar_t *ch = NULL;

	wcscat_s( temp, SHORT_BUF_LEN, ArrayName );
	wsprintf( InfoString, L"%d", Info );
	wcscat_s( temp, SHORT_BUF_LEN, L"," );
	wcscat_s( temp, SHORT_BUF_LEN, InfoString );
	wcscat_s( temp, SHORT_BUF_LEN, L"]" );

	ExpandVariables( temp, 1 );

	if ( iswdigit( temp[0] ) == 0 )
		return -1;

	_set_errno( 0 );
	Value = wcstol( temp, &ch, 10 );

	if ( errno || ( *ch != 0x00 ) )
		return -1;

	*outValue = Value;
	return 0;
}


int MakeSafe( LPTSTR String )
{
	if ( String == NULL )
		return -1;

	wchar_t temp[MAX_ARG_LEN] = L"";
	wchar_t ch = 0x00;

	int i = 0, j = 0;

	for ( i = 0; String[i]; i++ ) {
		ch = String[i];
		
		if ( ( ch == MAKESAFE_ESCAPE_1 ) || ( ch == MAKESAFE_ESCAPE_2 ) ) {
			if ( j >= ( MAX_ARG_LEN - 2 ) )
				return -1;

			temp[j++] = ch;
			temp[j++] = ch;
			temp[j] = 0x00;
		}
		else if ( ( ch < SPACE ) || ( ( ch >= 0x7f ) && ( ch <= 0x9f ) ) ) {
			if ( ( j + 4 ) >= MAX_ARG_LEN )
				return -1;

			wsprintf( temp + j, L"%c%02x", MAKESAFE_ESCAPE_1, ch );
			j += 3;
		}
		else {
			if ( j >= ( MAX_ARG_LEN - 1 ) )
				return -1;

			temp[j++] = ch;
			temp[j] = 0;
		}
	}

	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;
}


int HexDigitValue( wchar_t ch )
{
	ch = towupper( ch );

	if ( iswdigit( ch ) )
		return ch - '0';
	else if ( iswalpha( ch ) )
		return ch - 'A' + 10;
	else
		return -1;
}


int MakeUnsafe( LPTSTR String )
{
	if ( String == NULL )
		return -1;

	wchar_t temp[MAX_ARG_LEN] = L"";
	wchar_t ch = 0x00;

	int i = 0, j = 0;

	for ( i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( ch == MAKESAFE_ESCAPE_1 ) {
			if ( String[i + 1] == ch )
				i++;
			else if ( iswxdigit( String[i + 1] ) ) {
				i++;
				ch = HexDigitValue( String[i] );

				if ( iswxdigit( String[i + 1] ) ) {
					i++;
					ch = ( ch << 4 ) | HexDigitValue( String[i] );
				}
			}
		}
		else if ( ch == MAKESAFE_ESCAPE_2 ) {
			if ( String[i + 1] == ch )
				i++;
			else if ( iswxdigit( String[i + 1] ) ) {
				ch = HexDigitValue( String[i] );

				if ( iswxdigit( String[i + 1] ) ) {
					i++;
					ch = ( ch << 4 ) | HexDigitValue( String[i] );

					if ( iswxdigit( String[i + 1] ) ) {
						i++;
						ch = ( ch << 4 ) | HexDigitValue( String[i] );

						if ( iswxdigit( String[i + 1] ) ) {
							i++;
							ch = ( ch << 4 ) | HexDigitValue( String[i] );
						}
					}
				}
			}
		}

		if ( j < ( MAX_ARG_LEN - 1 ) ) {
			temp[j++] = ch;
			temp[j] = 0x00;
		}
	}


	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;
}


int WriteLineAsUTF8( HANDLE FileHandle, BOOL StdOut, LPTSTR Text )
{
	if ( Text == NULL )
		return -1;

	if ( StdOut ) {
		if ( Text[0] == BOM )
			Text += 1;

		QPuts( Text );
		QPuts( L"\r\n" );
		return 0;
	}

	DWORD BufferSize = ( MAX_ARG_LEN + 3 ) * 3;
	DWORD BytesWritten = 0;

	BYTE Buffer[ ( MAX_ARG_LEN + 3 ) * 3 ];
	
	int cb = UTF16toUTF8( Text, Buffer, BufferSize );
	if ( cb < 0 ) 
		return -1;

	if ( ( cb + 1 ) >= (int) BufferSize )
		return -1;

	Buffer[cb++] = '\n';
	Buffer[cb] = 0x00;

	WriteFile( FileHandle, Buffer, cb, &BytesWritten, NULL );
	return 0;
}


int SaveArrayItem( HANDLE FileHandle, BOOL StdOut, LPTSTR ArrayName, unsigned int ArrayDimensions, unsigned int X, unsigned int Y, unsigned int Z, unsigned int W, unsigned int XBase, unsigned int YBase, unsigned int ZBase, unsigned int WBase )
{
	if ( ArrayName == NULL )
		return -1;
	else if ( ArrayName[0] == 0x00 )
		return -1;


	wchar_t LineBuffer[SAVEARRAY_LINE_LEN] = L"";
	wchar_t Value[MAX_ARG_LEN] = L"";


	switch ( ArrayDimensions ) {
		case 1 :
			wsprintf( LineBuffer, L"%u=", X - XBase );
			wsprintf( Value, L"%%%s[%u]", ArrayName, X );
			break;

		case 2 :
			wsprintf( LineBuffer, L"%u,%u=", X - XBase, Y - YBase );
			wsprintf( Value, L"%%%s[%u,%u]", ArrayName, X, Y );
			break;

		case 3 :
			wsprintf( LineBuffer, L"%u,%u,%u=", X - XBase, Y - YBase, Z - ZBase );
			wsprintf( Value, L"%%%s[%u,%u,%u]", ArrayName, X, Y, Z );
			break;

		default :
			wsprintf( LineBuffer, L"%u,%u,%u,%u=", X - XBase, Y - YBase, Z - ZBase, W - WBase );
			wsprintf( Value, L"%%%s[%u,%u,%u,%u]", ArrayName, X, Y, Z, W );
	}

	ExpandVariables( Value, 1 );
	if ( Value[0] == 0x00 )
		return 0;

	MakeSafe( Value );
	wcscat_s( LineBuffer, SAVEARRAY_LINE_LEN, Value );

	WriteLineAsUTF8( FileHandle, StdOut, LineBuffer );
	return 0;
}



DLLExports int WINAPI savearray( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t ArrayName[SHORT_BUF_LEN]	= L"";
	wchar_t Filename[MAX_FILENAME_LEN]	= L"";
	wchar_t FileHeader[SHORT_BUF_LEN]	= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";

	BOOL Overwrite						= FALSE;
	BOOL StdOut							= FALSE;
	BOOL Quiet							= FALSE;

	unsigned int ArrayDimensions		= 0;
	unsigned int ArraySize[4]			= { 1, 1, 1, 1 };
	BOOL Partial						= FALSE;
	BOOL SavePartialAsWhole				= FALSE;
	BOOL NoOverlap						= FALSE;
	BOOL PartialX						= FALSE;
	BOOL PartialY						= FALSE;
	BOOL PartialZ						= FALSE;
	BOOL PartialW						= FALSE;
	unsigned int XStart					= 0;
	unsigned int YStart					= 0;
	unsigned int ZStart					= 0;
	unsigned int WStart					= 0;
	unsigned int XEnd					= UINT_MAX;
	unsigned int YEnd					= UINT_MAX;
	unsigned int ZEnd					= UINT_MAX;
	unsigned int WEnd					= UINT_MAX;
	unsigned int XBase					= 0;
	unsigned int YBase					= 0;
	unsigned int ZBase					= 0;
	unsigned int WBase					= 0;

	HANDLE FileHandle					= NULL;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );

	
	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( SaveArrayHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( SAVEARRAY_COMMAND_VER ) );
			else if ( !_wcsicmp( arg[i], L"/O" ) )
				Overwrite = TRUE;
			else if ( !_wcsicmp( arg[i], L"/P" ) )
				SavePartialAsWhole = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/X", 2 ) ) {
				if ( ParseRange( arg[i], &XStart, &XEnd, 0, UINT_MAX, L"XStart", L"XEnd" ) )
					return 1;

				XEnd++;
				Partial = PartialX = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/Y", 2 ) ) {
				if ( ParseRange( arg[i], &YStart, &YEnd, 0, UINT_MAX, L"YStart", L"YEnd" ) )
					return 1;

				YEnd++;
				Partial = PartialY = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/Z", 2 ) ) {
				if ( ParseRange( arg[i], &ZStart, &ZEnd, 0, UINT_MAX, L"ZStart", L"ZEnd" ) )
					return 1;

				ZEnd++;
				Partial = PartialZ = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				if ( ParseRange( arg[i], &WStart, &WEnd, 0, UINT_MAX, L"WStart", L"WEnd" ) )
					return 1;

				WEnd++;
				Partial = PartialW = TRUE;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( ArrayName[0] == '\0' ) {
				wcscpy_s( ArrayName, SHORT_BUF_LEN, arg[i] );
				_wcsupr_s( ArrayName, SHORT_BUF_LEN );
			}
			else if ( Filename[0] == '\0' ) {
				if ( !wcscmp( arg[i], StdinFilename ) ) {
					wcscpy_s( Filename, MAX_FILENAME_LEN, arg[i] );
					StdOut = TRUE;
				}
				else
					FH_Truename( arg[i], Filename );
			}
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( ArrayName[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing array name\r\n" );
		return 1;
	}

	if ( Filename[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return 1;
	}

	if ( !GoodArrayName( ArrayName ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad array name: \"%s\"\r\n", ArrayName );
		return 1;
	}



	if ( !StdOut && !Overwrite ) {
		if ( QueryIsFile( Filename ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File exists: \"%s\"\r\n", Filename );
			return 2;
		}
	}

	if ( GetArrayInfo( ArrayName, 0, &ArrayDimensions ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Array does not exist: \"%s\"\r\n", ArrayName );
		return 2;
	}

	for ( unsigned int i = 0; i < ArrayDimensions; i++ ) {
		if ( GetArrayInfo( ArrayName, i + 1, &ArraySize[i] ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error getting dimension %u of array \"%s\"\r\n", i + 1, ArrayName );
			return 2;
		}
	}


	if ( PartialX ) {
		if ( XStart >= ArraySize[0] )
			NoOverlap = TRUE;

		if ( XEnd > ArraySize[0] )
			XEnd = ArraySize[0];
	}
	else
		XEnd = ArraySize[0];

	if ( PartialY ) {
		if ( YStart >= ArraySize[1] )
			NoOverlap = TRUE;

		if ( YEnd > ArraySize[1] )
			YEnd = ArraySize[1];
	}
	else
		YEnd = ArraySize[1];

	if ( PartialZ ) {
		if ( ZStart >= ArraySize[2] )
			NoOverlap = TRUE;

		if ( ZEnd > ArraySize[2] )
			ZEnd = ArraySize[2];
	}
	else
		ZEnd = ArraySize[2];

	if ( PartialW ) {
		if ( WStart >= ArraySize[3] )
			NoOverlap = TRUE;

		if ( WEnd > ArraySize[3] )
			WEnd = ArraySize[3];
	}
	else
		WEnd = ArraySize[3];

	if ( !Partial )
		SavePartialAsWhole = FALSE;

	if ( NoOverlap ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Specified partial array does not overlap existing array\r\n" );
		return 1;
	}

	if ( ( PartialY && ( ArrayDimensions < 2 ) ) || ( PartialZ && ( ArrayDimensions < 3 ) ) || ( PartialW && ( ArrayDimensions < 4 ) ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Partial dimensions specified not in existing array\r\n" );
		return 1;
	}

	if ( SavePartialAsWhole ) {
		XBase = XStart;
		YBase = YStart;
		ZBase = ZStart;
		WBase = WStart;
	}


	if ( !StdOut ) {
		FileHandle = CreateFile( Filename, GENERIC_WRITE, FILE_SHARE_READ, NULL, Overwrite ? CREATE_ALWAYS : CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL );

		if ( FileHandle == INVALID_HANDLE_VALUE ) {
			GetErrorText( GetLastError(), ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening \"%s\": %s\r\n", Filename, ErrorText );
			return 2;
		}
	}

	switch ( ArrayDimensions ) {
		case 1 :
			if ( SavePartialAsWhole )
				wsprintf( FileHeader, L"%c%s %s[%u]", BOM, SaveArrayHeaderString, ArrayName, XEnd - XStart );
			else
				wsprintf( FileHeader, L"%c%s %s[%u]", BOM, SaveArrayHeaderString, ArrayName, ArraySize[0] );

			break;

		case 2 :
			if ( SavePartialAsWhole )
				wsprintf( FileHeader, L"%c%s %s[%u,%u]", BOM, SaveArrayHeaderString, ArrayName, XEnd - XStart, YEnd - YStart );
			else
				wsprintf( FileHeader, L"%c%s %s[%u,%u]", BOM, SaveArrayHeaderString, ArrayName, ArraySize[0], ArraySize[1] );

			break;

		case 3 :
			if ( SavePartialAsWhole )
				wsprintf( FileHeader, L"%c%s %s[%u,%u,%u]", BOM, SaveArrayHeaderString, ArrayName, XEnd - XStart, YEnd - YStart, ZEnd - ZStart );
			else
				wsprintf( FileHeader, L"%c%s %s[%u,%u,%u]", BOM, SaveArrayHeaderString, ArrayName, ArraySize[0], ArraySize[1], ArraySize[2] );

			break;

		default :
			if ( SavePartialAsWhole )
				wsprintf( FileHeader, L"%c%s %s[%u,%u,%u,%u]", BOM, SaveArrayHeaderString, ArrayName, XEnd - XStart, YEnd - YStart, ZEnd - ZStart, WEnd - WStart );
			else
				wsprintf( FileHeader, L"%c%s %s[%u,%u,%u,%u]", BOM, SaveArrayHeaderString, ArrayName, ArraySize[0], ArraySize[1], ArraySize[2], ArraySize[3] );
	}

	WriteLineAsUTF8( FileHandle, StdOut, FileHeader );

	PrepSetDos();

	for ( unsigned int X = XStart; X < XEnd; X++ )
		for ( unsigned int Y = YStart; Y < YEnd; Y++ )
			for ( unsigned int Z = ZStart; Z < ZEnd; Z++ )
				for ( unsigned int W = WStart; W < WEnd; W++ )
					SaveArrayItem( FileHandle, StdOut, ArrayName, ArrayDimensions, X, Y, Z, W, XBase, YBase, ZBase, WBase );

	if ( !StdOut )
		CloseHandle( FileHandle );

	if ( !Quiet )
		Printf( L"Saved array %s to \"%s\".\r\n", ArrayName, Filename );

	ResetSetDos();
	return 0;
}



DLLExports int WINAPI loadarray( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t ArrayName[SHORT_BUF_LEN]			= L"";
	wchar_t Filename[MAX_FILENAME_LEN]			= L"";
	wchar_t Line[SAVEARRAY_LINE_LEN]			= L"";
	wchar_t ArrayNameFromFile[SHORT_BUF_LEN]	= L"";
	wchar_t Value[MAX_ARG_LEN]					= L"";
	wchar_t CmdBuf[SAVEARRAY_LINE_LEN]			= L"";
	wchar_t *ch									= NULL;
	const wchar_t BadFile[]						= L"Not a SAVEARRAY file: \"%s\"\r\n";

	_MMFile MMFile;

	BOOL Quiet = FALSE;

	unsigned int ArrayDimensions				= 0;
	unsigned int ArraySize[4]					= { 1, 1, 1, 1 };
	unsigned int ArrayIndex[4]					= { 0, 0, 0, 0 };

	unsigned int CodePage						= 0;
	unsigned int EbcdicCodePage					= 0;



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( LoadArrayHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( SAVEARRAY_COMMAND_VER ) );
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &CodePage, &EbcdicCodePage ) )
					return 1;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( Filename[0] == '\0' ) {
				if ( wcscmp( arg[i], StdinFilename ) )
					FH_Truename( arg[i], Filename );
				else
					wcscpy_s( Filename, arg[i] );
			}
			else if ( ArrayName[0] == '\0' ) {
				wcscpy_s( ArrayName, SHORT_BUF_LEN, arg[i] );
				_wcsupr_s( ArrayName, SHORT_BUF_LEN );
			}
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( Filename[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return 1;
	}

	if ( ArrayName[0] ) {
		if ( !GoodArrayName( ArrayName ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad array name: \"%s\"\r\n", ArrayName );
			return 1;
		}
	}



	int rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES, CodePage, EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, BadFile, Filename );
		return 2;
	}
	else if ( rv )
		return 2;


	if ( MMGetLine( &MMFile, Line, SAVEARRAY_LINE_LEN, 0 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, BadFile, Filename );
		goto ErrorExit;
	}

	if ( wcsncmp( Line, SaveArrayHeaderString, wcslen( SaveArrayHeaderString ) ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, BadFile, Filename );
		goto ErrorExit;
	}

	int i = (int) wcslen( SaveArrayHeaderString ), j = 0;

	if ( Line[i] != SPACE ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, BadFile, Filename );
		goto ErrorExit;
	}

	while ( Line[i] == SPACE )
		i++;

	if ( !iswalpha( Line[i] ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, BadFile, Filename );
		goto ErrorExit;
	}

	while ( iswalnum( Line[i] ) || ( Line[i] == '_' ) || ( Line[i] == '$' ) ) {
		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			ArrayNameFromFile[j++] = towupper( Line[i] );
			ArrayNameFromFile[j] = '\0';
		}
		i++;
	}

	if ( Line[i++] != '[' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, BadFile, Filename );
		goto ErrorExit;
	}

	ch = Line + i;
	ArrayDimensions = 0;

	do {
		if ( !iswdigit( *ch ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, BadFile, Filename );
			goto ErrorExit;
		}

		if ( ParseInt( ch, &ArraySize[ArrayDimensions], &ch ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, BadFile, Filename );
			goto ErrorExit;
		}

		ArrayDimensions++;
		if ( ArrayDimensions > 4 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, BadFile, Filename );
			goto ErrorExit;
		}

		if ( !wcscmp( ch, L"]" ) )
			break;
		else if ( *ch != ',' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, BadFile, Filename );
			goto ErrorExit;
		}
		
		ch++;

	} while ( TRUE );

	if ( ArrayName[0] == '\0' ) {
		if ( !GoodArrayName( ArrayNameFromFile ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad array name: \"%s\"\r\n", ArrayNameFromFile );
			goto ErrorExit;
		}

		wcscpy_s( ArrayName, SHORT_BUF_LEN, ArrayNameFromFile );
	}


	wcscpy_s( CmdBuf, SAVEARRAY_LINE_LEN, L" /Q " );			//  delete any existing array:
	wcscat_s( CmdBuf, SAVEARRAY_LINE_LEN, ArrayName );
	CallInternal( L"unsetarray", CmdBuf );

	switch ( ArrayDimensions ) {
		case 1 :
			wsprintf( CmdBuf, L" %s[%u]", ArrayName, ArraySize[0] );
			break;

		case 2 :
			wsprintf( CmdBuf, L" %s[%u,%u]", ArrayName, ArraySize[0], ArraySize[1] );
			break;

		case 3 :
			wsprintf( CmdBuf, L" %s[%u,%u,%u]", ArrayName, ArraySize[0], ArraySize[1], ArraySize[2] );
			break;

		default :
			wsprintf( CmdBuf, L" %s[%u,%u,%u,%u]", ArrayName, ArraySize[0], ArraySize[1], ArraySize[2], ArraySize[3] );
	}

	if ( CallInternal( L"setarray", CmdBuf ) ) {			//  create the array
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error creating array %s\r\n", ArrayName );
		goto ErrorExit;
	}


	if ( !Quiet )
		Printf( L"Created array variable %s.\r\n", CmdBuf + 1 );


	BOOL Valid = TRUE;

	do {
		if ( MMGetLine( &MMFile, Line, SAVEARRAY_LINE_LEN, MMGL_FLAG_SKIP_BLANKS ) )
			break;

		ch = Line;
		Valid = TRUE;

		for ( unsigned int i = 0; i < ArrayDimensions; i++ ) {
			if ( ParseInt( ch, &ArrayIndex[i], &ch ) )
				Valid = FALSE;
			else if ( *ch != ( ( i == ( ArrayDimensions - 1 ) ) ? '=' : ',' ) )
				Valid = FALSE;
			
			ch++;
		}

		if ( *ch == '\0' )
			Valid = FALSE;

		if ( Valid ) {

			wcscpy_s( Value, MAX_ARG_LEN, ch );
			if ( MakeUnsafe( Value ) == 0 ) {
				
				switch ( ArrayDimensions ) {
					case 1 :
						wsprintf( CmdBuf, L" %s[%u]=", ArrayName, ArrayIndex[0] );
						break;

					case 2 :
						wsprintf( CmdBuf, L" %s[%u,%u]=", ArrayName, ArrayIndex[0], ArrayIndex[1] );
						break;

					case 3 :
						wsprintf( CmdBuf, L" %s[%u,%u,%u]=", ArrayName, ArrayIndex[0], ArrayIndex[1], ArrayIndex[2] );
						break;

					default :
						wsprintf( CmdBuf, L" %s[%u,%u,%u,%u]=", ArrayName, ArrayIndex[0], ArrayIndex[1], ArrayIndex[2], ArrayIndex[3] );
				}

				wcscat_s( CmdBuf, SAVEARRAY_LINE_LEN, Value );
				CallInternal( L"set", CmdBuf );
			}
		}


	} while ( TRUE );


	CloseMMFile( &MMFile );
	return 0;

ErrorExit:

	CloseMMFile( &MMFile );
	return 2;
}


#endif	//	SAVEARRAY_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Adds a version number to this source file.  Help text tweaks.

*/
