
//		TextUtils.cpp - Text file utilities for TCC


#define UNICODE						1
#define _UNICODE					1
#define _CRT_RAND_S

#include <stdio.h>
#include <tchar.h>
#include <windows.h>
#include <shlobj.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					0
#define VER_MINOR					85
#define VER_BUILD					0
#define VER_PATCH					3

#define PLUGIN_NAME					L"TextUtils"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif


#define MACRO_EXP( arg )			#arg
#define MACRO_STR( arg )			MACRO_EXP( arg )


#define SHORT_BUF_LEN				1024
#define MAX_FILENAME_LEN			32768
#define BIG_LINE_LEN				65536
#define PLUGIN_BUFFER_MAX			32768


#define FMT_UNKNOWN					0
#define FMT_UNFORMATTED				1
#define FMT_PREWRAPPED				2
#define FMT_UNFORMATTED_BLANKS		3
#define FMT_MAX_VALUE				3

#define MIN_COLUMNS					40
#define MAX_COLUMNS					512
#define DEFAULT_COLUMNS				100

#define MAX_TAB_SIZE				20


#define LF							0x000a
#define CR							0x000d
#define TAB							0x0009
#define ESC							0x001b
#define SPACE						0x0020
#define NBSP						0x00a0

#define HYPHEN						0x002d
#define SHYPHEN						0x00ad
#define HYPHEN2						0x2010

#define QUOTE						0x0022
#define APOSTROPHE					0x0027
#define GRAVE						0x0060
#define MIDDOT						0x00b7
#define EMDASH						0x2014
#define LSQUOTE						0x2018
#define RSQUOTE						0x2019
#define LDQUOTE						0x201c
#define RDQUOTE						0x201d
#define ELLIPSIS					0x2026
#define INTERROBANG					0x203d
#define FWQUOTE						0xff02
#define FWAPOSTROPHE				0xff07
#define FWPERIOD					0xff0e

#define BOM							0xfeff
#define NON_CHARACTER				0xfffe

#define MAX_INDENT					20

#define UPEND_EXPAND				0x0001
#define UPEND_NO_BLANKS				0x0002
#define UPEND_INVERT				0x0004
#define UPEND_HEADER				0x0008
#define UPEND_REPLACE_CONTROLS		0x0010
#define UPEND_TRIM					0x0020

#define SHUFFLE_LINE_NUMBERS		0x0001
#define SHUFFLE_NO_BLANKS			0x0002
#define SHUFFLE_ORIG_LINE_NUMBERS	0x0004
#define SHUFFLE_HEADER				0x0008

#define OINK_HEADER					1
#define OINK_FIX_QUOTES				4

#define ROT13_HEADER				1
#define ROT13_FIX_QUOTES			4

#define RECASE_HEADER				1
#define RECASE_MODE_LOWER			1
#define RECASE_MODE_UPPER			2
#define RECASE_MODE_CAPS			3
#define RECASE_MODE_RANDOM			4


#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )


HINSTANCE MyDLLHandle				= NULL;

int ShellVerMajor = 0, ShellVerMinor = 0, ShellVerBuild = 0, ShellTccLe = 0;


//			Variables used by FixQuotes()

wchar_t OpenQuote = 0x201c,		CloseQuote = 0x201d;
wchar_t OpenSQuote = 0x2018,	CloseSQuote = 0x2019;
wchar_t Prime = APOSTROPHE,		DoublePrime = QUOTE;
wchar_t Apostrophe = 0x2019,	Okina = 0x2018;
wchar_t EmDash = 0x2014;
BOOL ReplaceDoubles		= TRUE;
BOOL ReplaceGraves		= TRUE;

//	--------------------------------------------------



wchar_t PluginFeatures[] = L"charencoding,clip2text,context,copychars,countchars,dedup,degas,\
dehtml,ffields,filterfiles,loadarray,oink,parseargs,password,recase,replacetext,rot13,savearray,\
shuffle,text2clip,textutilshelp,unicodify,upend,utype,words,wrap,xfilter,\
@b85tobin,@between,@bintob85,@clarify,@inivalue,@lineends,@metaphone,@mkentities,@obscure,@oink,\
@rot13,@roughlysimilar,@stripaccents,@textencoding,@textformat,@uchar,@ucode,@ucodex,@ulen,\
@uquotes,@vowels,\
_characters,_charactersall,_getacp,_inivaluerc,_lines,_linesall,_longestline,_longestlineall,\
_nonblanklines,_nonblanklinesall,_paragraphs,_paragraphsall,_password,_propernouns,\
_propernounsall,_sentences,_sentencesall,_sentencesd,_sentencesdall,_sentencese,\
_sentenceseall,_sentencesq,_sentencesqall,_sentencewords,_sentencewordsall,_titles,_titlesall,\
_uniquewords,_uniquewordsall,_wc,_wcall,_wordfiles,_words,_wordsall";



const wchar_t OinkHelpText[] = L"Translate a text file to Pig Latin.\n\
\n\
OINK /CP:n /H /P /S filename....\r\n\
\n\
\t/CP:  interpret OEM text as code page n\n\
\t/H    display filenames\n\
\t/P    page output\n\
\t/S    search in subdirectories for matching files\n\
\n\
Input filenames may be specified on the command line, or text may \n\
be piped or redirected into OINK.\n";


const wchar_t RecaseHelpText[] = L"Change the case of text.\n\
\n\
RECASE /CP:n /H /L /P /S /T /U filename....\r\n\
\n\
\t/CP:  interpret OEM text as code page n\n\
\t/H    display filenames\n\
\t/L    make text lowercase\n\
\t/P    page output\n\
\t/S    search in subdirectories for matching files\n\
\t/T    capitalize first letter of each word\n\
\t/U    make text uppercase\n\
\n\
Input filenames may be specified on the command line, or text may \n\
be piped or redirected into RECASE.\n";


const wchar_t Rot13HelpText[] = L"Encode or decode text with ROT13.\n\
\n\
ROT13 /CP:n /H /P /S filename....\r\n\
\n\
\t/CP:  interpret OEM text as code page n\n\
\t/H    display filenames\n\
\t/P    page output\n\
\t/S    search in subdirectories for matching files\n\
\n\
Input filenames may be specified on the command line, or text may \n\
be piped or redirected into ROT13.\n";


const wchar_t ShuffleHelpText[] = L"Dump randomized lines from a text file.\n\
\n\
SHUFFLE /B /CP:n /H /J /L /M:n /P /S filename....\r\n\
\n\
\t/B    skip blank lines\n\
\t/CP:  interpret OEM text as code page n\n\
\t/H    display filenames\n\
\t/J    show original line numbers\n\
\t/L    show line numbers\n\
\t/M:   maximum lines to display\n\
\t/P    page output\n\
\t/S    search in subdirectories for matching files\n\
\n\
Input filenames may be specified on the command line, or text may \n\
be piped or redirected into SHUFFLE.\n";


const wchar_t TextUtilsHelpHelpText[] = L"Open the TextUtils plugin help file.\n\
\n\
TEXTUTILSHELP /C /F /S /S:text /V topic\r\n\
\n\
\t/C\tselect the Contents tab\n\
\t/F\tselect the Favorites tab\n\
\t/S\tselect the Search tab\n\
\t/S:\tselect the Search tab and search for text\n\
\t/V\tshow detailed plugin version info\n\
\n\
Opens this plugin\u2019s help file to the specified topic.\n\
If no topic is given, opens the Overview page by default. \n";


const wchar_t UpendHelpText[] = L"Display lines from a file in reverse order.\n\
\n\
UPEND /B /C /CP:n /E /H /L:string /P /R:string /S /T /V /W:n filename.... \r\n\
\n\
\t/B    discard blank lines\n\
\t/C    replace control characters\n\
\t/CP:  interpret OEM text as code page n\n\
\t/E    expand variables in /L: and /R: strings\n\
\t/H    display filenames\n\
\t/L:   insert string to the left of each line\n\
\t/P    page output\n\
\t/R:   insert string to the right of each line\n\
\t/S    search in subdirectories for matching files\n\
\t/T    trim leading and trailing whitespace\n\
\t/V    also reverse each line\n\
\t/W:   truncate lines to n characters\n\
\n\
Input filenames may be specified on the command line, or text may\n\
be piped or redirected into UPEND.\n";



const wchar_t Unaccented[65] = L"AAAAAA\u00c6CEEEEIIII\u00d0NOOOOO\u00d7\u00d8UUUUY\u00de\u00dfaaaaaa\u00e6ceeeeiiii\u00f0nooooo\u00f7\u00f8uuuuy\u00fey";

const wchar_t UnaccentedC0[64][3]		= { L"A", L"A", L"A", L"A", L"A", L"A", L"AE", L"C", L"E", L"E", L"E", L"E", L"I", L"I", L"I", L"I",
											L"\u00d0", L"N", L"O", L"O", L"O", L"O", L"O", L"\u00d7", L"\u00d8", L"U", L"U", L"U", L"U", L"Y", L"\u00de", L"\u00df",
											L"a", L"a", L"a", L"a", L"a", L"a", L"ae", L"c", L"e", L"e", L"e", L"e", L"i", L"i", L"i", L"i",
											L"\u00f0", L"n", L"o", L"o", L"o", L"o", L"o", L"\u00f7", L"\u00f8", L"u", L"u", L"u", L"u", L"y", L"\u00fe", L"y" };



const wchar_t StringNotAvailable[] = L"N/A";

const wchar_t StdinFilename[] = L"-";
const wchar_t ClipFilename[]  = L"clip:";



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_FILENAME_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_FILENAME_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int ConsoleHandle( HANDLE inHandle )
{
	if ( !inHandle )
		return 0;

	if ( ( inHandle == (HANDLE) STD_INPUT_HANDLE ) || ( inHandle == (HANDLE) STD_OUTPUT_HANDLE ) || ( inHandle == (HANDLE) STD_ERROR_HANDLE ) )
		inHandle = GetStdHandle( (DWORD) inHandle );

	DWORD Mode = UINT_MAX;
	int rv = GetConsoleMode( inHandle, &Mode );

	if ( !rv || ( Mode == UINT_MAX ) )
		return 0;
	else
		return 1;
}



int ConsoleHandle( DWORD inHandle )
{
	return ConsoleHandle( (HANDLE) inHandle );
}



int StripQuotes( LPTSTR String )
{
	if ( !String )
		return -1;


	size_t i	= 0;
	size_t j	= 0;
	wchar_t ch	= '\0';
	int k		= 0;


	do {
		ch = String[i++];
		if ( ch == QUOTE ) {
			k++;
			continue;
		}

		String[j++] = ch;
	}
	while ( ch );
	
	return k;
}



LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( Filename == NULL )
		return NULL;
	else if ( Filename[0] == 0x00 )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion, LPCSTR SourceFile, LPCSTR SourceFileVer )
{
	wchar_t File[MAX_FILENAME_LEN]		= L"";
	wchar_t Ver[SHORT_BUF_LEN]			= L"";


	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion ) {
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

		if ( SourceFile && SourceFileVer ) {
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFile, -1, File, MAX_FILENAME_LEN );
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFileVer, -1, Ver, SHORT_BUF_LEN );
			StripQuotes( File );
			Printf( L"\t\t\t\t%s  %s\r\n", _FilenamePart( File, NULL), Ver );
		}
	}

	return 0;
}



int SeedRandomGenerator()
{
	DWORD q = 0, a = 0, b = 0;
	SYSTEMTIME time;

	GetSystemTime( &time );
	a = ( ( time.wHour << 16 ) | time.wMinute ) * 1997;
	b = ( ( time.wSecond << 16 ) | time.wMilliseconds ) * 449;
	q = ( ( time.wMonth << 16 ) | time.wDay ) * 17;
	q = q ^ a ^ b;

	srand( q );
	return 0;
}



unsigned int GetRand( unsigned int Range )
{
	unsigned int RandVal = 0;
	double real = 0.0;

	if ( !rand_s( &RandVal ))
		real = (double) RandVal / ( UINT_MAX + 1.0 );
	else
		real = (double) ( rand() | ( rand() << 5 ) ) / ( ( RAND_MAX + 1.0 ) * 32.0 );

	real = real * (double) Range;
	return (unsigned int) real;
}



int DisableOneFeature( LPTSTR FeaturesList, LPTSTR FeatureToDisable )
{
	if ( !FeaturesList || !FeatureToDisable )
		return -1;

	if ( ( FeaturesList[0] == '\0' ) || ( FeatureToDisable[0] == '\0' ) )
		return -1;

	size_t i = 0, j = 0, base = 0;
	wchar_t ch = '\0', ch2 = '\0';
	BOOL done = FALSE, match = FALSE;

	do {
		// Compare word in 'FeatureToDisable' against word at 'base' :

		i = 0;
		match = TRUE;

		do {
			ch = towupper( FeaturesList[base + i] );
			ch2 = ch;
			if ( ( ch2 == ',' ) || ( ch2 == ';' ) )
				ch2 = '\0';

			match &= ( ch2 == towupper( FeatureToDisable[i] ) );

			i++;
		} while ( ch2 != '\0' );

		if ( match ) {
			// disable the feature:

			if ( ch == '\0' )
				FeaturesList[base] = ch;
			else {
				j = 0;

				do {
					ch = FeaturesList[base + i + j];
					FeaturesList[base + j] = ch;
					j++;
				} while ( ch != '\0' );
			}
		}
		else
			base += i;

	} while ( ch != '\0' );

	i = wcslen( FeaturesList );
	if ( i > 0 )
		if ( ( FeaturesList[i - 1] == ',' ) || ( FeaturesList[i - 1] == ';' ) )
			FeaturesList[i - 1] = '\0';

	return 0;
}



BOOL ArraysAvailable( )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"";
	int ShellMajor = -1, TccLe = 0;
	wchar_t *sep1 = NULL;

	wcscpy_s( shortbuf, SHORT_BUF_LEN, L"%_4ver" );
	ExpandVariables( shortbuf, 1 );
	ShellMajor = wcstol( shortbuf, &sep1, 10 );

	wcscpy_s( shortbuf, SHORT_BUF_LEN, L"%_cmdproc" );
	ExpandVariables( shortbuf, 1 );
	if ( !_wcsicmp( shortbuf, L"TCCLE" ) )
		TccLe = 1;

	if ( ( ShellMajor < 10 ) || ( ShellMajor > 255 ) )
		return FALSE;

	if ( TccLe )
		return FALSE;

	return TRUE;
}



BOOL HasBinaryBuffers()
{
	wchar_t TempBuffer[SHORT_BUF_LEN] = L"%_4ver";
	wchar_t *sep1 = NULL;
	long int ShellMajor = -1, ShellMinor = -1;

	ExpandVariables( TempBuffer, 1 );
	ShellMajor = wcstol( TempBuffer, &sep1, 10 );

	if ( *sep1 ) {
		sep1++;
		if ( *sep1 )
			ShellMinor = wcstol( sep1, NULL, 10 );
	}

	wcscpy_s( TempBuffer, SHORT_BUF_LEN, L"%_cmdproc" );
	ExpandVariables( TempBuffer, 1 );

	if ( ( ShellMajor < 7 ) || ( ShellMajor > 99 ) || ( ShellMinor < 0 ) || ( ShellMinor > 99 ) )
		return FALSE;				//  version number fails sanity checks

	if ( ShellMajor < 10 )
		return FALSE;

	if ( !_wcsicmp( TempBuffer, L"TCCLE" ) )
		return FALSE;

	return TRUE;
}



int GetShellVersion()
{
	wchar_t temp[SHORT_BUF_LEN] = L"%_cmdproc";
	wchar_t *sep1 = NULL;


	ExpandVariables( temp, 1 );
	if ( !_wcsicmp( temp, L"TCCLE" ) )
		ShellTccLe = 1;
	else
		ShellTccLe = 0;


	wcscpy_s( temp, SHORT_BUF_LEN, L"%_4ver" );
	ExpandVariables( temp, 1 );
	ShellVerMajor = wcstol( temp, &sep1, 10 );

	if ( *sep1 ) {
		sep1++;
		if ( *sep1 )
			ShellVerMinor = wcstol( sep1, NULL, 10 );
	}

	wcscpy_s( temp, SHORT_BUF_LEN, L"%_build" );
	ExpandVariables( temp, 1 );
	ShellVerBuild = wcstol( temp, &sep1, 10 );

	if ( ( ShellVerMajor < 7 ) || ( ShellVerMajor > 99 ) || ( ShellVerMinor < 0 ) || ( ShellVerMinor > 99 ) || ( ShellVerBuild < 0 ) || ( ShellVerBuild > 999 ) ) {
		ShellVerMajor = ShellVerMinor = ShellVerBuild = -1;
		return -1;
	}

	return 0;
}



#include "ParseArgs.cpp"



int CallInternal( LPTSTR Cmd, LPTSTR Args )
{
	if ( !Cmd )
		return -1;
	else if ( Cmd[0] == '\0' )
		return -1;


	wchar_t CmdBuf[MAX_ARG_LEN]		= L"";
	wchar_t ExpVar[SHORT_BUF_LEN]	= L"%_expansion";
	BOOL AddStar					= TRUE;


	ExpandVariables( ExpVar, 1 );
	if ( wcschr( ExpVar, '1' ) )
		AddStar = FALSE;


	if ( ( Cmd[0] != '*' ) && AddStar )
		wcscpy_s( CmdBuf, MAX_ARG_LEN, L"*" );

	wcscat_s( CmdBuf, MAX_ARG_LEN, Cmd );

	if ( Args ) {
		if ( Args[0] ) {
			if ( !iswspace( Args[0] ) )
				wcscat_s( CmdBuf, MAX_ARG_LEN, L" " );

			wcscat_s( CmdBuf, MAX_ARG_LEN, Args );
		}
	}

	return Command( CmdBuf, 0 );
}



int ConsoleWidth( )
{
	HANDLE ConsoleHandle = CreateFile( L"CONOUT$", ( GENERIC_READ | GENERIC_WRITE ), ( FILE_SHARE_READ | FILE_SHARE_WRITE ), 0, OPEN_EXISTING, NULL, NULL );
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	int Width = 0, rv = 0;

	if ( ConsoleHandle != INVALID_HANDLE_VALUE ) {
		rv = GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
		CloseHandle( ConsoleHandle );

		if ( rv )
			Width = ConsoleInfo.dwSize.X;
	}

	if ( ( Width < MIN_COLUMNS ) || ( Width > MAX_COLUMNS ) )
		Width = 0;

	return Width;
}



int ConsoleHeight( )
{
	HANDLE ConsoleHandle = CreateFile( L"CONOUT$", ( GENERIC_READ | GENERIC_WRITE ), ( FILE_SHARE_READ | FILE_SHARE_WRITE ), 0, OPEN_EXISTING, NULL, NULL );
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	int Height = 0, rv = 0;

	if ( ConsoleHandle != INVALID_HANDLE_VALUE ) {
		rv = GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo );
		CloseHandle( ConsoleHandle );

		if ( rv )
			Height = ConsoleInfo.srWindow.Bottom - ConsoleInfo.srWindow.Top + 1;
	}

	if ( ( Height < 10 ) || ( Height > 100 ) )
		Height = 0;

	return Height;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



BOOL IsSpaceOrTab( wchar_t ch )
{
	return ( ( ch == SPACE ) || ( ch == TAB ) );
}



BOOL IsRangeSep( wchar_t ch )
{
	return ( ( ch == ',' ) || ( ch == ';' ) || ( ch == '-' ) );
}



BOOL IsVowel( wchar_t ch )
{
	wchar_t ch2 = towupper( ch );
	if ( ( ch2 == 'A' ) || ( ch2 == 'E' ) || ( ch2 == 'I' ) || ( ch2 == 'O' ) || ( ch2 == 'U' ) || ( ch2 == 'Y' ) )
		return TRUE;

	if ( ( ch >= 0xc0 ) && ( ch <= 0xff ) )
		if ( ( ch != 0xc7 ) && ( ch != 0xd0 ) && ( ch != 0xd1 ) && ( ch != 0xd7 ) && ( ch != 0xde ) && ( ch != 0xdf ) && ( ch != 0xe7 ) && ( ch != 0xf0 ) && ( ch != 0xf1 ) && ( ch != 0xf7 ) && ( ch != 0xfe ) )
			return TRUE;

	return FALSE;
}



BOOL IsHyphen( wchar_t ch )
{
	return ( ( ch == HYPHEN ) || ( ch == HYPHEN2 ) );
}



BOOL IsTerminalPunct( wchar_t ch )
{
	return ( ( ch == '!' ) || ( ch == '?' ) || ( ch == INTERROBANG ) );
}



BOOL LCStringContains( LPTSTR MainString, LPTSTR SearchString )
{
	if ( !MainString || !SearchString )
		return FALSE;

	wchar_t MainStringLC[BIG_LINE_LEN]	= L"";

	wcscpy_s( MainStringLC, BIG_LINE_LEN, MainString );
	_wcslwr_s( MainStringLC, BIG_LINE_LEN );

	if ( wcsstr( MainStringLC, SearchString ) )
		return TRUE;
	else
		return FALSE;
}



BOOL StuffNotToBreakBefore( wchar_t ch1, wchar_t ch2 )
{
	if ( ( ch1 == '&' ) || ( ch1 == '/' ) || ( ch1 == HYPHEN2 ) || ( ch1 == 0x2026 ) )
		return TRUE;

	if ( ( ch1 == HYPHEN ) && ( ch2 == HYPHEN ) )
		return TRUE;

	if ( ( ch1 == '.' ) && ( ch2 == '.' ) )
		return TRUE;

	return FALSE;
}



BOOL HasDigits( LPTSTR inString )
{
	if ( !inString )
		return FALSE;

	for ( size_t i = 0; inString[i]; i++ )
		if ( iswdigit( inString[i] ) )
			return TRUE;

	return FALSE;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value = 0;
	wchar_t *sep1 = NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int EnvironmentNumericValue( LPTSTR VarName, int Min, int Max )
{
	if ( !VarName )
		return -1;
	else if ( VarName[0] == '\0' )
		return -1;

	wchar_t Value[SHORT_BUF_LEN] = L"";
	unsigned int Numeric = 0;
	int i = 0;
	wchar_t *sep1 = NULL;

	if ( GetEnvironmentVariable( VarName, Value, SHORT_BUF_LEN ) < 1 )
		return -1;

	if ( Value[0] == '\0' )
		return -1;

	while ( iswspace( Value[i] ) )
		i++;

	if ( !iswdigit( Value[i] ) )
		return -1;

	if ( ParseInt( Value + i, &Numeric, &sep1 ) )
		return -1;

	while ( iswspace( *sep1 ) )
		sep1++;

	if ( *sep1 )
		return -1;

	if ( ( (int) Numeric < Min ) || ( (int) Numeric > Max ) )
		return -1;

	return Numeric;
}



int ParseRange( LPTSTR inString, unsigned int *outValue1, unsigned int *outValue2, unsigned int Minimum, unsigned int Maximum, LPTSTR Value1Name, LPTSTR Value2Name )
{
	if ( !inString || !outValue1 || !outValue2 )
		return -1;

	wchar_t Value1DefaultName[] = L"Minimum value";
	wchar_t Value2DefaultName[] = L"Maximum value";

	if ( !Value1Name )
		Value1Name = Value1DefaultName;

	if ( !Value2Name )
		Value2Name = Value2DefaultName;


	wchar_t *ch = inString;

	if ( ( *ch == '/' ) && ( iswalpha( *(ch + 1) ) ) ) {
		ch++;
		while ( iswalpha( *ch ) )
			ch++;
	}

	if ( IsColonOrEquals( *ch ) )
		ch++;

	if ( ( *ch == '\0' ) || ( IsRangeSep( *ch ) && ( *(ch + 1) == '\0' ) ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", inString );
		return 1;
	}


	if ( iswdigit( *ch ) ) {
		if ( ParseInt( ch, outValue1, &ch ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad %s \"%s\"\r\n", Value1Name, inString );
			return 1;
		}

		if ( *ch == '\0' )
			*outValue2 = *outValue1;
		else if ( !IsRangeSep( *ch ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error \"%s\"\r\n", inString );
			return 1;
		}

		if ( ( *outValue1 < Minimum ) || ( *outValue1 > Maximum ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"%s must be %u - %u\r\n", Value1Name, Minimum, Maximum );
			return 1;
		}
	}
	else
		*outValue1 = Minimum;

	if ( IsRangeSep( *ch ) ) {
		ch++;

		if ( *ch == '\0' )
			*outValue2 = Maximum;
		else if ( ParseInt( ch, outValue2, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad %s \"%s\"\r\n", Value2Name, ch );
			return 1;
		}

		if ( ( *outValue2 < *outValue1 ) || ( *outValue2 > Maximum ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"%s must be %u - %u\r\n", Value2Name, *outValue1, Maximum );
			return 1;
		}
	}
	else if ( *ch ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error \"%s\"\r\n", inString );
		return 1;
	}

	return 0;
}



int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace > 0 ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int TrimQuotes( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	size_t l = wcslen( String );
	if ( l < 2 )
		return 0;
	
	if ( String[0] != QUOTE )
		return 0;
	else if ( String[l - 1] != QUOTE )
		return 0;

	for ( size_t i = 0; i < ( l - 2 ); i++ )
		String[i] = String[i + 1];

	String[l - 2] = '\0';
	return 0;
}



#define MM_INTERNET_FEATURES
#define MM_ALLOW_CESU8
#include "mmfiles.cpp"
#include "metaphone.cpp"
#include "FileHandler.cpp"
#include "conlist.cpp"



int Oink( LPTSTR String, size_t BufLen )
{
	if ( !String )
		return 1;

	wchar_t *temp			= NULL;
	size_t j				= 0;
	wchar_t ch				= '\0';
	wchar_t ch2				= '\0';

	BOOL InWord				= FALSE;
	BOOL WasInWord			= FALSE;
	BOOL InitCap			= FALSE;
	BOOL AllCaps			= FALSE;

	int LettersToMove		= 0;
	int WordStart			= 0;

	const wchar_t SuffixString[] = L"way";
	LPTSTR Suffix = (LPTSTR) SuffixString;


	temp = (wchar_t*) malloc( BufLen * sizeof( wchar_t ) );
	if ( !temp )
		return -2;


	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];
		WasInWord = InWord;
		InWord = iswalpha( ch );

		if ( WasInWord && ( ( ch == APOSTROPHE ) || ( ch == RSQUOTE ) ) && iswalpha( String[i + 1] ) )
			InWord = TRUE;

		if ( InWord && !WasInWord ) {				//  start of a word
			LettersToMove = 0;
			WordStart = i;
			AllCaps = InitCap = iswupper( ch );

			if ( !iswalpha( String[i + 1] ) )
				AllCaps = FALSE;

			if ( towupper( ch ) == 'Q' ) {
				LettersToMove++;
				if ( towupper( String[i + 1] ) == 'U' )
					LettersToMove++;
			}
			else if ( towupper( ch ) == 'Y' ) {
				if ( IsVowel( String[i + 1] ) || !iswalpha( String[i + 1] ) )
					LettersToMove++;
			}
			else if ( !IsVowel( ch ) ) {
				while ( iswalpha( String[i + LettersToMove] ) && !IsVowel( String[i + LettersToMove] ) )
					LettersToMove++;
			}

			if ( LettersToMove )
				for ( int k = 0; k < LettersToMove; k++ )
					if ( iswlower( String[i + k] ) )
						AllCaps = FALSE;

			i += LettersToMove - 1;
		}
		else if ( InWord ) {						//  continue copying word
			if ( iswlower( ch ) )
				AllCaps = FALSE;

			if ( InitCap ) {
				ch = towupper( ch );
				InitCap = FALSE;
			}

			if ( j < ( BufLen - 1 ) ) {
				temp[j++] = ch;
				temp[j] = '\0';
			}
		}
		else if ( WasInWord ) {						//  end of word

			if ( LettersToMove ) {
				for ( int k = 0; k < LettersToMove; k++ ) {
					ch2 = String[WordStart + k];

					if ( AllCaps || InitCap )
						ch2 = towupper( ch2 );
					else
						ch2 = towlower( ch2 );

					if ( j < ( BufLen - 1 ) ) {
						temp[j++] = ch2;
						temp[j] = '\0';
					}
				}

				Suffix = (LPTSTR) SuffixString + 1;
			}
			else
				Suffix = (LPTSTR) SuffixString;

			for ( int k = 0; Suffix[k]; k++ ) {
				ch2 = Suffix[k];

				if ( AllCaps )
					ch2 = towupper( ch2 );
				else
					ch2 = towlower( ch2 );

				if ( j < ( BufLen - 1 ) ) {
					temp[j++] = ch2;
					temp[j] = '\0';
				}
			}

			if ( j < ( BufLen - 1 ) ) {
				temp[j++] = ch;
				temp[j] = '\0';
			}

			AllCaps = FALSE;
		}
		else {										//  not in a word; probably spaces or punctuation
			if ( j < ( BufLen - 1 ) ) {
				temp[j++] = ch;
				temp[j] = '\0';
			}
		}
	}

	if ( InWord ) {									//  partial word remaining -- word terminated by a null character:
		if ( LettersToMove ) {
			for ( int k = 0; k < LettersToMove; k++ ) {
				ch2 = String[WordStart + k];

				if ( AllCaps || InitCap )
					ch2 = towupper( ch2 );
				else
					ch2 = towlower( ch2 );

				if ( j < ( BufLen - 1 ) ) {
					temp[j++] = ch2;
					temp[j] = '\0';
				}
			}

			Suffix = (LPTSTR) SuffixString + 1;
		}
		else
			Suffix = (LPTSTR) SuffixString;

		for ( int k = 0; Suffix[k]; k++ ) {
			ch2 = Suffix[k];

			if ( AllCaps )
				ch2 = towupper( ch2 );
			else
				ch2 = towlower( ch2 );

			if ( j < ( BufLen - 1 ) ) {
				temp[j++] = ch2;
				temp[j] = '\0';
			}
		}
	}

	wcscpy_s( String, BufLen, temp );
	free( temp );
	return 0;
}



DLLExports int WINAPI f_oink( LPTSTR String )
{
	return Oink( String, MAX_ARG_LEN );
}



DLLExports int WINAPI f_rot13( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t *ch = String;

	while ( *ch ) {
		if ( ( *ch >= 'A' ) && ( *ch <= 'M' ) )
			*ch += 13;
		else if ( ( *ch >= 'N' ) && ( *ch <= 'Z' ) )
			*ch -= 13;
		else if ( ( *ch >= 'a' ) && ( *ch <= 'm' ) )
			*ch += 13;
		else if ( ( *ch >= 'n' ) && ( *ch <= 'z' ) )
			*ch -= 13;

		ch++;
	}

	return 0;
}



int StripAccents( LPTSTR inString, LPTSTR outString, size_t buflen )
{
	if ( !inString || !outString || ( buflen < 16 ) )
		return -1;


	wchar_t *temp		= (wchar_t*) malloc( ( buflen + 16 ) * sizeof( wchar_t ) );

	size_t j		= 0;
	int k			= 0;


	if ( !temp )
		return ERROR_NOT_ENOUGH_MEMORY;

	temp[0] = '\0';


	for ( wchar_t *ch = inString; *ch; ch++ ) {
		if ( ( *ch >= 0x00c0 ) && ( *ch <= 0x00ff ) ) {
			wcscat_s( temp, buflen + 16, UnaccentedC0[*ch - 0x00c0] );
			j += wcslen( UnaccentedC0[*ch - 0x00c0] );
		}
		else if ( *ch == 0x0152 ) {
			wcscat_s( temp, buflen + 16, L"OE" );
			j += 2;
		}
		else if ( *ch == 0x0153 ) {
			wcscat_s( temp, buflen + 16, L"oe" );
			j += 2;
		}
		else
			temp[j++] = *ch;

		temp[j] = '\0';

		if ( j >= buflen ) {
			free( temp );
			return ERROR_NOT_ENOUGH_MEMORY;
		}
	}

	wcscpy_s( outString, buflen, temp );
	free( temp );
	return 0;
}



DLLExports int WINAPI f_stripaccents( LPTSTR String )
{
	if ( !String )
		return 1;


	int rv = StripAccents( String, String, MAX_ARG_LEN );


	if ( rv ) {
		GetErrorText( rv, String );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s\r\n", String );
		return -1;
	}

	return 0;
}



DLLExports int WINAPI f_textencoding( LPTSTR String )
{
	if ( !String )
		return 1;

	ParseArgs( String, PARSE_BREAK_COMMAS );


	_MMFile MMFile;
	unsigned int Flags	= 0;


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Flags, NULL ) ) {
			DisplayErrorHeader();
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad flags \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	int rv = OpenMMFile( &MMFile, arg[0], MM_FLAG_ALLOW_EBCDIC );

	if ( rv == ERROR_EMPTY_FILE ) {
		wcscpy_s( String, MAX_ARG_LEN, L"Empty" );
		return 0;
	}
	else if ( rv ) {
		wcscpy_s( String, MAX_ARG_LEN, L"ERROR" );
		return 0;
	}

	if ( ( MMFile.FileEncoding == FILE_OEM ) && ( Flags & 2 ) )
		if ( LooksLikeAscii( &MMFile ) )
			MMFile.FileEncoding = FILE_ASCII;

	MM_GetEncodingString( &MMFile, ( Flags & 1 ), FALSE, String, SHORT_BUF_LEN );
	CloseMMFile( &MMFile );
	return 0;
}



DLLExports int WINAPI f_textformat( LPTSTR String )
{
	if ( !String )
		return 1;

	ParseArgs( String, PARSE_BREAK_COMMAS );


	_MMFile MMFile;
	unsigned int Flags = 0;


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Flags, NULL ) ) {
			DisplayErrorHeader();
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad flags \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	int rv = OpenMMFile( &MMFile, arg[0], MM_FLAG_ALLOW_EBCDIC );

	if ( rv == ERROR_EMPTY_FILE ) {
		wcscpy_s( String, MAX_ARG_LEN, L"Empty" );
		return 0;
	}
	else if ( rv ) {
		wcscpy_s( String, MAX_ARG_LEN, L"ERROR" );
		return 0;
	}

	if ( MMFile.FileSize == MMFile.BOMSize ) {
		wcscpy_s( String, MAX_ARG_LEN, L"Empty" );
		CloseMMFile( &MMFile );
		return 0;
	}

	GuessFileFormat( &MMFile );

	switch ( MMFile.FileFormat ) {
		case FMT_PREWRAPPED :
			wcscpy_s( String, MAX_ARG_LEN, L"Prewrapped" );
			break;

		case FMT_UNFORMATTED :
			wcscpy_s( String, MAX_ARG_LEN, L"Unformatted" );
			break;

		case FMT_UNFORMATTED_BLANKS :
			wcscpy_s( String, MAX_ARG_LEN, L"Unformatted" );

			if ( Flags & 1 )
				wcscat_s( String, MAX_ARG_LEN, L" with blanks" );

			break;

		default:
			wcscpy_s( String, MAX_ARG_LEN, L"Unknown" );
	}

	CloseMMFile( &MMFile );
	return 0;
}



DLLExports int WINAPI f_lineends( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";
	wchar_t NotStdin[]					= L".\\-";

	_MMFile MMFile;

	unsigned int Which			= 0;
	unsigned __int64 nCr		= 0;
	unsigned __int64 nLf		= 0;
	unsigned __int64 nCrLf		= 0;
	unsigned __int64 nLfCr		= 0;
	unsigned __int64 nNel		= 0;
	unsigned __int64 nTotal		= 0;
	int ch						= 0;
	int rv						= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Which, NULL ) ) {
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}

	if ( !wcscmp( arg[0], StdinFilename ) )
		arg[0] = NotStdin;


	rv = OpenMMFile( &MMFile, arg[0], MM_FLAG_ALLOW_EBCDIC );

	if ( rv == ERROR_EMPTY_FILE ) {
		wcscpy_s( String, MAX_ARG_LEN, L"Empty" );
		return 0;
	}
	else if ( rv ) {
		GetErrorText( MMFile.ErrorCode, ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
		Qprintf( ERRHANDLE, L"   \"%s\"\r\n", arg[0] );
		return -1;
	}


	while ( MMFile.Index < MMFile.FileSize ) {

		ch = MMGetChar( &MMFile );

		if ( ch < 0 )
			goto ErrorExit;

		if ( ch == CR ) {
			if ( MMPeekAhead( &MMFile, 0 ) == LF ) {
				nCrLf++;
				MMGetChar( &MMFile );
			}
			else
				nCr++;

			nTotal++;
		}
		else if ( ch == LF ) {
			if ( MMPeekAhead( &MMFile, 0 ) == CR ) {
				nLfCr++;
				MMGetChar( &MMFile );
			}
			else
				nLf++;

			nTotal++;
		}
		else if ( ch == NELCHR ) {
			nNel++;
			nTotal++;
		}
	}

	CloseMMFile( &MMFile );


	switch ( Which ) {

	case 1 :
		wsprintf( String, L"%I64u", nCrLf );
		break;

	case 2 :
		wsprintf( String, L"%I64u", nLfCr );
		break;

	case 3 :
		wsprintf( String, L"%I64u", nCr );
		break;

	case 4 :
		wsprintf( String, L"%I64u", nLf );
		break;

	case 5 :
		wsprintf( String, L"%I64u", nNel );
		break;

	case 6 :
	case 7 :
	case 8 :
	case 9 :
		String[0] = '\0';
		break;

	case 10 :
		wsprintf( String, L"%I64u", nTotal );
		break;

	case 128 :
		wsprintf( String, L"nCrLf=%I64u  nLfCr=%I64u  nCr=%I64u  nLf=%I64u  nNel=%I64u  nTotal=%I64u", nCrLf, nLfCr, nCr, nLf, nNel, nTotal );
		break;

	default :

		if ( nTotal == 0 )
			wcscpy_s( String, MAX_ARG_LEN, L"None" );
		else if ( nCrLf == nTotal )
			wcscpy_s( String, MAX_ARG_LEN, L"CR/LF" );
		else if ( nLfCr == nTotal )
			wcscpy_s( String, MAX_ARG_LEN, L"LF/CR" );
		else if ( nCr == nTotal )
			wcscpy_s( String, MAX_ARG_LEN, L"CR" );
		else if ( nLf == nTotal )
			wcscpy_s( String, MAX_ARG_LEN, L"LF" );
		else if ( nNel == nTotal )
			wcscpy_s( String, MAX_ARG_LEN, L"NEL" );
		else
			wcscpy_s( String, MAX_ARG_LEN, L"Mixed" );
	}

	return 0;


ErrorExit:

	CloseMMFile( &MMFile );
	wcscpy_s( String, MAX_ARG_LEN, L"ERROR" );
	return 0;
}



DLLExports int WINAPI f_between( LPTSTR String )
{
	if ( !String )
		return 1;

	if ( wcslen( String ) < 3 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing argument\r\n" );
		return -1;
	}

	if ( String[2] != ',' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing comma\r\n" );
		return -1;
	}


	wchar_t Start = String[0], End = String[1];
	int i = 3, j = 0;

	String[j] = '\0';


	while ( String[i] && ( String[i] != Start ) )
		i++;

	if ( String[i] == '\0' )
		return 0;

	i++;


	while ( String[i] && ( String[i] != End ) ) {
		if ( j < ( MAX_ARG_LEN - 1 ) ) {
			String[j++] = String[i];
			String[j] = '\0';
		}

		i++;
	}

	return 0;
}



int StringToUnicode( LPTSTR String, int Format )
{
	if ( !String )
		return 1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	wchar_t val[SHORT_BUF_LEN]			= L"";
	size_t len							= 0;
	wchar_t *ch							= String;
	unsigned int CharVal				= 0;


	EscapeLine( String );

	while ( *ch ) {
		CharVal = *ch;

		if ( ( CharVal & 0xfc00 ) == 0xd800 )
			if ( ( ch[1] & 0xfc00 ) == 0xdc00 )
				CharVal = ( ( ( ch[0] & 0x3ff ) << 10 ) | ( ch[1] & 0x3ff ) ) + 0x10000, ch++;

		switch ( Format ) {

		case 1 :
			wsprintf( val, L"0x%x ", CharVal );
			break;

		case 2 :
			wsprintf( val, L"U+%04X ", CharVal );
			break;

		case 3 :
			wsprintf( val, L"%u ", CharVal );
			break;

		default :
			wsprintf( val, L"%x ", CharVal );
		}

		if ( ( len + wcslen( val ) ) >= MAX_FILENAME_LEN ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
			return 2;
		}

		wcscat_s( temp, MAX_FILENAME_LEN, val );
		len += wcslen( val );
		ch++;
	}

	if ( len )
		temp[--len] = '\0';

	wcscpy_s( String, MAX_FILENAME_LEN, temp );
	return 0;
}



DLLExports int WINAPI f_ucode( LPTSTR String )
{
	return StringToUnicode( String, 0 );
}



DLLExports int WINAPI f_ucodex( LPTSTR String )
{
	return StringToUnicode( String, 1 );
}



DLLExports int WINAPI f_ucodeu( LPTSTR String )
{
	return StringToUnicode( String, 2 );
}



DLLExports int WINAPI f_ulen( LPTSTR String )
{
	if ( !String )
		return 1;


	size_t nChars		= MM_CodePointsIn( String );


	wsprintf( String, L"%u", (unsigned int) nChars );
	return 0;
}



DLLExports int WINAPI f_uchar( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	unsigned int j						= 0;
	wchar_t *ch							= String;
	wchar_t *sep1						= NULL;
	unsigned int value					= 0;
	unsigned int k						= 0;


	while ( *ch ) {

		while ( iswspace( *ch ) )
			ch++;

		if ( *ch == '\0' )
			break;

		if ( ( ch[0] == '0' ) && ( towupper( ch[1] ) == 'X' ) && iswxdigit( ch[2] ) )
			value = wcstoul( ch + 2, &sep1, 16 );
		else if ( ( towupper( ch[0] ) == 'U' ) && ( ch[1] == '+' ) && iswxdigit( ch[2] ) )
			value = wcstoul( ch + 2, &sep1, 16 );
		else if ( ( ch[0] == '#' ) && iswdigit( ch[1] ) )
			value = wcstoul( ch + 1, &sep1, 10 );
		else if ( iswxdigit( ch[0] ) )
			value = wcstoul( ch, &sep1, 16 );
		else
			goto Fail;

		if ( *sep1 && !iswspace( *sep1 ) )
			goto Fail;

		if ( !value || ( value > 0x10ffff ) )
			goto Fail2;

		ch = sep1;

		k = ( value > 0xffff ) ? 2 : 1;

		if ( ( j + k ) >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
			return -1;
		}

		if ( k == 2 ) {
			MM_CreateSurrogatePair( value, temp + j );
			j += 2;
			temp[j]		= '\0';
		}
		else {
			temp[j++]	= value;
			temp[j]		= '\0';
		}
	}


	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;


Fail :

	sep1 = ch;
	while ( *sep1 && !iswspace( *sep1 ) )
		sep1++;

	if ( *sep1 )
		*sep1 = '\0';

	_wcsupr_s( ch, MAX_ARG_LEN );
	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Bad character value: \"%s\"\r\n", ch );
	return -1;

Fail2 :

	DisplayErrorHeader();
	Qprintf( ERRHANDLE, L"Bad character value; should be 0x0001 - 0x10FFFF: 0x%04X\r\n", value );
	return -1;
}



DLLExports int WINAPI f_metaphone( LPTSTR String )
{
	if ( !String )
		return 1;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	unsigned int Length = 8, Flags = 0;
	BOOL Compatible = FALSE, Done = FALSE;
	wchar_t WordBuf[SHORT_BUF_LEN] = L"";
	int j = 0;
	wchar_t ch = '\0';

	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing word\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Length, NULL ) ) {
			DisplayErrorHeader();
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad length \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( ( Length < METAPHONE_MIN ) || ( Length > METAPHONE_MAX ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Length must be %u - %u\r\n", METAPHONE_MIN, METAPHONE_MAX );
			return -1;
		}
	}

	if ( arglen[2] ) {
		if ( ParseInt( arg[2], &Flags, NULL ) ) {
			DisplayErrorHeader();
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad flags \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	String[0] = '\0';
	for ( int i = 0; !Done; i++ ) {
		ch = arg[0][i];

		if ( iswspace( ch ) && ( WordBuf[0] == '\0' ) )
			continue;

		if ( iswalnum( ch ) || ( ( ( ch == APOSTROPHE ) || ( ch == RSQUOTE ) || ( ch == '-' ) ) && j && iswalpha( arg[0][i + 1] ) ) ) {
			if ( j < ( SHORT_BUF_LEN - 1 ) ) {
				WordBuf[j++] = ch;
				WordBuf[j] = '\0';
			}
		}
		else {
			if ( j ) {
				Metaphone( WordBuf, WordBuf, Length, ( Flags & 1 ) > 0 );

				if ( String[0] )
					wcscat_s( String, MAX_ARG_LEN, L" " );

				wcscat_s( String, MAX_ARG_LEN, WordBuf );
			}

			j = 0;
			WordBuf[j] = '\0';

			if ( ch == '\0' )
				Done = TRUE;
		}
	}

	return 0;
}



DLLExports int WINAPI f_vowels( LPTSTR String )
{
	if ( !String )
		return 1;

	int Count = 0;

	for ( int i = 0; String[i]; i++ )
		if ( IsVowel( String[i] ) )
			Count++;

	wsprintf( String, L"%u", Count );
	return 0;
}



int GetEnvironmentChar( LPTSTR Name, wchar_t *outChar )
{
	if ( !Name )
		return -1;
	else if ( Name[0] == '\0' )
		return 0;

	wchar_t Temp[SHORT_BUF_LEN] = L"";
	unsigned int NumericChar = 0;
	int i = 0;
	wchar_t *sep1 = NULL;

	if ( GetEnvironmentVariable( Name, Temp, SHORT_BUF_LEN ) < 1 )
		return 0;
	else if ( Temp[0] == '\0' )
		return 0;

	if ( ( Temp[0] > 31 ) && ( Temp[1] == '\0' ) )
		if ( !iswdigit( Temp[0] ) ) {
			*outChar = Temp[0];
			return 0;
		}

	while ( iswspace( Temp[i] ) )
		i++;

	if ( !iswdigit( Temp[i] ) )
		return -1;

	if ( ParseInt( Temp + i, &NumericChar, &sep1 ) )
		return -1;

	while ( iswspace( *sep1 ) )
		sep1++;

	if ( *sep1 )
		return -1;

	if ( ( NumericChar < 32 ) || ( NumericChar > 0xfffd ) )
		return -1;

	*outChar = NumericChar;
	return 0;
}



BOOL IsFraction( wchar_t ch )
{
	if ( ( ch >= 0xbc ) && ( ch <= 0xbe ) )
		return TRUE;
	else if ( ( ch >= 0x2150 ) && ( ch <= 0x215e ) )
		return TRUE;
	else
		return FALSE;
}



BOOL IsVowel2( wchar_t ch )
{
	ch = towlower( ch );
	if ( ( ch == 'a' ) || ( ch == 'e' ) || ( ch == 'i' ) || ( ch == 'o' ) || ( ch == 'u' ) )
		return TRUE;

	if ( ( ch == 0x0100 ) || ( ch == 0x0101 ) || ( ch == 0x0112 ) || ( ch == 0x0113 ) || ( ch == 0x012a ) || ( ch == 0x012b ) || ( ch == 0x014c ) || ( ch == 0x014d ) || ( ch == 0x016a ) || ( ch == 0x016b ) )
		return TRUE;		//	vowels with macrons -- used in hawai'ian

	return FALSE;
}



BOOL IsWordBeginningWithApostrophe( LPTSTR inString )
{
	if ( !inString )
		return FALSE;
	else if ( inString[0] == '\0' )
		return FALSE;

	wchar_t Temp[SHORT_BUF_LEN] = L"";
	wchar_t ch = '\0';
	int i = 0;

	const int NumWords = 21;
	const wchar_t Word[NumWords][12] = { L"'cos", L"'cus", L"'em", L"'mid", L"'mongst",
		L"'n", L"'n'", L"'neath", L"'tain't", L"'til", L"'tis", L"'tisn't", L"'twas", L"'twasn't",
		L"'tween", L"'twere", L"'tweren't", L"'twixt", L"'twould", L"'twouldn't", L"'un" };

	do {
		ch = towlower( inString[i] );
		if ( ( ch == 0x2018 ) || ( ch == 0x2019 ) )
			ch = 0x27;

		if ( !iswalpha( ch ) && ( ch != 0x27 ) )
			ch = '\0';

		if ( i < ( SHORT_BUF_LEN - 1 ) )
			Temp[i++] = ch;

		Temp[i] = '\0';

	} while ( ch && ( i < ( SHORT_BUF_LEN - 1 ) ) );

	for ( i = 0; i < NumWords; i++ )
		if ( !wcscmp( Temp, Word[i] ) )
			return TRUE;

	return FALSE;
}



BOOL AnyKindOfQuote( wchar_t ch )
{
	if ( ( ch == QUOTE ) || ( ch == APOSTROPHE ) || ( ch == GRAVE ) )
		return TRUE;

	if ( ( ch == LSQUOTE ) || ( ch == RSQUOTE ) )
		return TRUE;

	if ( ( ch == LDQUOTE ) || ( ch == RDQUOTE ) )
		return TRUE;

	if ( ( ch == FWQUOTE ) || ( ch == FWAPOSTROPHE ) )
		return TRUE;

	return FALSE;
}



int FixQuotes( LPTSTR String, size_t BufLen )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;


	wchar_t *Temp					= NULL;
	wchar_t ch						= SPACE;
	wchar_t lastch					= '\0';
	wchar_t nextch					= '\0';
	size_t j						= 0;


	Temp = (wchar_t*) malloc( BufLen * sizeof( wchar_t ) );
	if ( !Temp )
		return -2;


	for ( int i = 0; String[i]; i++  ) {
		lastch = ch;
		ch = String[i];
		nextch = String[i + 1];

		if ( ReplaceDoubles && ( ch == APOSTROPHE ) && ( nextch == APOSTROPHE ) ) {
			if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( String[i + 2] ) ) {
				ch = QUOTE;
				i++;
				nextch = String[i + 1];
			}
		}

		if ( ReplaceDoubles && ReplaceGraves && ( ch == GRAVE ) && ( nextch == GRAVE ) ) {
			if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( String[i + 2] ) ) {
				ch = LDQUOTE;
				i++;
				nextch = String[i + 1];
			}
		}

		if ( ReplaceGraves && ( ch == GRAVE ) )
			if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( nextch ) )
				ch = LSQUOTE;

		if ( ( ch == QUOTE ) || ( ch == FWQUOTE ) ) {
			if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswspace( nextch ) || _MMIsEOL( nextch ) || ( nextch == '\0' ) ) )
				ch = DoublePrime;
			else if ( iswalpha( lastch ) && !iswalpha( nextch ) )
				ch = CloseQuote;
			else if ( iswalpha( nextch ) && !iswalpha( lastch ) )
				ch = OpenQuote;
			else if ( iswgraph( lastch ) && ( lastch != OpenSQuote ) )
				ch = CloseQuote;
			else {
				ch = OpenQuote;

				if ( iswspace( lastch ) )
					if ( _MMIsEOL( nextch ) || ( nextch == '\0' ) )
						ch = CloseQuote;
			}
		}
		else if ( ( ch == APOSTROPHE ) || ( ch == FWAPOSTROPHE ) ) {
			if ( iswgraph( lastch ) && ( lastch != OpenQuote ) ) {
				ch = CloseSQuote;

				if ( iswalnum( lastch ) && iswalpha( nextch ) )
					ch = Apostrophe;
			}
			else {
				ch = OpenSQuote;

				if ( iswspace( lastch ) )
					if ( _MMIsEOL( nextch ) || ( nextch == '\0' ) )
						ch = CloseSQuote;
			}

			if ( !iswalpha( lastch ) && iswalpha( nextch ) ) {
				if ( IsWordBeginningWithApostrophe( String + i ) )
					ch = Apostrophe;
			}
			else if ( IsVowel2( lastch ) && IsVowel2( nextch ) )
				ch = Okina;

			if ( !iswalnum( lastch ) && iswdigit( nextch ) )
				if ( iswdigit( String[i + 2] ) )
					if ( !iswalnum( String[i + 3] ) )
						ch = Apostrophe;

			if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswdigit( nextch ) || IsFraction( nextch ) || iswspace( nextch ) || _MMIsEOL( nextch ) || ( nextch == '\0' ) ) )
				ch = Prime;
		}
		else if ( ( ch == '-' ) || ( ch == 0xff0d ) ) {
			if ( ( ( lastch != '-' ) && ( lastch != 0xff0d ) ) && ( nextch == ch ) )
				if ( ( String[i + 2] != '-' ) && ( String[i + 2] != 0xff0d ) ) {
					ch = EmDash;
					i++;
				}
		}

		if ( j < ( BufLen - 1 ) ) {
			Temp[j++] = ch;
			Temp[j] = '\0';
		}
	}

	wcscpy_s( String, BufLen, Temp );
	free( Temp );
	return 0;
}



int InitializeFixQuotesVars()
{
	OpenQuote		= 0x201c;
	CloseQuote		= 0x201d;
	OpenSQuote		= 0x2018;
	CloseSQuote		= 0x2019;
	Prime			= APOSTROPHE;
	DoublePrime		= QUOTE;
	Apostrophe		= 0x2019;
	Okina			= 0x2018;
	EmDash			= 0x2014;
	ReplaceDoubles	= TRUE;
	ReplaceGraves	= TRUE;

	GetEnvironmentChar( L"OPENQUOTE",   &OpenQuote );
	GetEnvironmentChar( L"CLOSEQUOTE",  &CloseQuote );
	GetEnvironmentChar( L"OPENSQUOTE",  &OpenSQuote );
	GetEnvironmentChar( L"CLOSESQUOTE", &CloseSQuote );
	GetEnvironmentChar( L"APOSTROPHE",  &Apostrophe );
	GetEnvironmentChar( L"PRIME",		&Prime );
	GetEnvironmentChar( L"DOUBLEPRIME",	&DoublePrime );
	GetEnvironmentChar( L"'OKINA",		&Okina );
	GetEnvironmentChar( L"EMDASH",		&EmDash );

	if ( EnvironmentNumericValue( L"UQUOTES_DOUBLES", 0, INT_MAX ) == 0 )
		ReplaceDoubles	= FALSE;

	if ( EnvironmentNumericValue( L"UQUOTES_GRAVES", 0, INT_MAX ) == 0 )
		ReplaceGraves	= FALSE;

	return 0;
}



DLLExports int WINAPI f_uquotes( LPTSTR String )
{
	if ( !String )
		return 1;
	else if ( String[0] == '\0' )
		return 0;

	InitializeFixQuotesVars();
	return FixQuotes( String, MAX_ARG_LEN );
}



DLLExports int WINAPI _getacp( LPTSTR outString )
{
	if ( !outString )
		return 1;

	unsigned int CodePage = GetACP();
	wsprintf( outString, L"%u", CodePage );
	return 0;
}



int OutputIsUnicode()
{
	if ( QueryUnicodeOutput() )
		return 16;

	if ( ShellVerMajor < 13 )
		return 0;
	else if ( ( ShellVerMajor == 13 ) && ( ShellVerMinor < 4 ) )
		return 0;
	else if ( ( ShellVerMajor == 13 ) && ( ShellVerMinor == 4 ) && ( ShellVerBuild < 46 ) )
		return 0;

	wchar_t temp[SHORT_BUF_LEN] = L"%@option[utf8output]";
	ExpandVariables( temp, 1 );

	if ( !_wcsicmp( temp, L"yes" ) )
		return 8;
	else
		return 0;
}



int SetOutputEncoding( int Encoding )
{
	BOOL Utf8Supported			= FALSE;


	if ( ShellVerMajor > 13 )
		Utf8Supported	= TRUE;
	else if ( ( ShellVerMajor == 13 ) && ( ShellVerMinor > 4 ) )
		Utf8Supported	= TRUE;
	else if ( ( ShellVerMajor == 13 ) && ( ShellVerMinor == 4 ) && ( ShellVerBuild >= 46 ) )
		Utf8Supported	= TRUE;


	if ( ( Encoding == 8 ) && !Utf8Supported )
		return ERROR_NOT_SUPPORTED;


	if ( ( Encoding == 0 ) || ( Encoding == 1 ) ) {
		CallInternal( L"option", L"//unicodeoutput=no" );

		if ( Utf8Supported )
			CallInternal( L"option", L"//utf8output=no" );

		return 0;
	}
	else if ( Encoding == 8 ) {
		CallInternal( L"option", L"//unicodeoutput=no" );

		if ( Utf8Supported )
			CallInternal( L"option", L"//utf8output=yes" );

		return 0;
	}
	else if ( Encoding == 16 ) {
		CallInternal( L"option", L"//unicodeoutput=yes" );

		if ( Utf8Supported )
			CallInternal( L"option", L"//utf8output=no" );

		return 0;
	}
	else
		return 1;
}



int InsertCommas( LPTSTR Buffer, unsigned int BufferLen )
{
	if ( !Buffer )
		return -1;

	if ( wcslen( Buffer ) > ( MAX_ARG_LEN - 10 ) )
		return 1;

	wchar_t temp[MAX_ARG_LEN]	= L"%@comma[";
	wcscat_s( temp, MAX_ARG_LEN, Buffer );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	if ( ExpandVariables( temp, 1 ) )
		return 2;

	if ( wcslen( temp ) >= BufferLen )
		return 3;

	wcscpy_s( Buffer, BufferLen, temp );
	return 0;
}



LPTSTR Commafy( unsigned __int64 Value )
{
	static wchar_t Buffer[64];


	Buffer[0] = '\0';
	wsprintf( Buffer, L"%%@comma[%I64u]", Value );
	ExpandVariables( Buffer, 1 );
	return Buffer;
}



int ShowFilenameHeader( _MMFile *MMFile, BOOL Verbose )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";


	if ( CL_EndLine() == USER_ABORT )
		return USER_ABORT;

	if ( !_wcsicmp( MMFile->Filename, ClipFilename ) ) {
		if ( HLputs( L"<clipboard>", 2 ) == USER_ABORT )
			return USER_ABORT;
	}
	else if ( !wcscmp( MMFile->Filename, StdinFilename ) ) {
		if ( HLputs( L"<stdin>", 2 ) == USER_ABORT )
			return USER_ABORT;
	}
	else {
		if ( HLprintf( L"\x12\rFile \"%s\" :\n", MMFile->Filename ) == USER_ABORT )
			return USER_ABORT;
	}

	if ( Verbose ) {

		Printf( L"  \u2022  %s byte%s", Commafy( MMFile->FileSize ), ( MMFile->FileSize == 1 ) ? L"" : L"s" );

		if ( MMFile->FileSize >= 1073741824 ) {
			swprintf( temp, MAX_ARG_LEN, L"%.1d", (double) MMFile->FileSize / 1073741824.0 );
			InsertCommas( temp, MAX_ARG_LEN );
			Printf( L"  (%s GB)", temp );
		}
		else if ( MMFile->FileSize >= 1048576 ) {
			swprintf( temp, MAX_ARG_LEN, L"%.1d", (double) MMFile->FileSize / 1048576.0 );
			InsertCommas( temp, MAX_ARG_LEN );
			Printf( L"  (%s MB)", temp );
		}
		else if ( MMFile->FileSize >= 1024 ) {
			swprintf( temp, MAX_ARG_LEN, L"%.1f", (double) MMFile->FileSize / 1024.0 );
			InsertCommas( temp, MAX_ARG_LEN );
			Printf( L"  (%s kB)", temp );
		}

		temp[0] = '\0';
		MM_GetEncodingString( MMFile, TRUE, TRUE, temp, MAX_ARG_LEN );
		if ( temp[0] )
			Printf( L"    %s", temp );

		if ( CL_EndLine( 2 ) == USER_ABORT )
			return USER_ABORT;
	}
	else {
		if ( CL_EndLine( 1 ) == USER_ABORT )
			return USER_ABORT;
	}

	return 0;
}



int ShowFilenameHeader( _MMFile *MMFile )
{
	return ShowFilenameHeader( MMFile, FALSE );
}



int Detab( LPTSTR String, unsigned int TabWidth, wchar_t SpaceChar )
{
	if ( !String )
		return -1;

	if ( !wcschr( String, TAB ) )
		return 0;

	if ( TabWidth == 0 )
		TabWidth = 1;
	else if ( TabWidth > MAX_TAB_SIZE )
		TabWidth = 4;

	wchar_t temp[BIG_LINE_LEN] = L"";
	int j = 0;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( ch == TAB ) {
			int n = TabWidth - ( j % TabWidth );

			if ( SpaceChar )
				for ( int k = 0; k < n; k++ )
					if ( j < ( BIG_LINE_LEN - 1 ) )
						temp[j++] = SpaceChar;

			temp[j] = '\0';
		}
		else {
			if ( j < ( BIG_LINE_LEN - 1 ) )
				temp[j++] = ch;

			temp[j] = '\0';
		}
	}

	wcscpy_s( String, BIG_LINE_LEN, temp );
	return 0;
}



int ReplaceChar( LPTSTR String, wchar_t FromChar, wchar_t ToChar )
{
	if ( !String )
		return -1;

	int k = 0;

	for ( int i = 0; String[i]; i++ ) {
		if ( String[i] == FromChar ) {
			String[i] = ToChar;
			k++;
		}
	}

	return k;
}



wchar_t StripAccent( wchar_t ch )
{
	if ( ( ch < 0x00c0 ) || ( ch > 0x00ff ) )
		return ch;

	return Unaccented[ch - 0x00c0];
}



int TrimRight( LPTSTR inString )
{
	if ( !inString )
		return -1;
	else if ( inString[0] == '\0' )
		return 0;

	int i = (int) wcslen( inString ) - 1;

	while ( iswspace( inString[i] ) ) {
		inString[i--] = '\0';
		if ( i < 0 )
			break;
	}
	
	return 0;
}



LPTSTR Stristr( LPTSTR Buffer, LPTSTR SearchText )
{
	if ( !Buffer || !SearchText )
		return NULL;

	if ( SearchText[0] == '\0' )
		return NULL;

	LPTSTR Index = Buffer;
	BOOL Found = FALSE;

	while ( Index[0] ) {
		Found = TRUE;

		for ( size_t i = 0; SearchText[i]; i++ ) {
			if ( towupper( Index[i] ) != towupper( SearchText[i] ) ) {
				Found = FALSE;
				break;
			}
		}

		if ( Found )
			return Index;

		Index++;
	}
	
	return NULL;
}



LPTSTR GetClipboardText( size_t *Characters )
{
	HANDLE ClipHandle		= NULL;
	wchar_t *ClipText		= NULL;
	wchar_t *TextCopy		= NULL;
	size_t ClipTextLen		= 0;



	if ( Characters )
		*Characters = 0;

	if ( !OpenClipboard( GetConsoleWindow() ) )
		return NULL;


	ClipHandle = GetClipboardData( CF_UNICODETEXT );

	if ( ClipHandle ) {
		ClipText = (wchar_t *) GlobalLock( ClipHandle );

		if ( ClipText ) {
			ClipTextLen = wcslen( ClipText );

			if ( ClipTextLen ) {
				TextCopy = (wchar_t *) malloc( ( ClipTextLen + 1 ) * sizeof( wchar_t ) );

				if ( TextCopy )
					wcscpy_s( TextCopy, ClipTextLen + 1, ClipText );
			}

			GlobalUnlock( ClipHandle );
		}
	}

	CloseClipboard();

	if ( TextCopy && Characters )
		*Characters = ClipTextLen;

	return TextCopy;
}



int GetYorN()
{
	unsigned int ch = 0;

	EatKeystrokes();

	do {
		if ( QueryKeyWaiting() ) {

			ch = GetKeystroke( 0x0030 );

			ch = towupper( ch );

			if ( ch == 'Y' ) {
				_putwch( ch );
				_putwch( LF );
				return 1;
			}
			else if ( ( ch == 'N' ) || ( ch == ESC ) ) {
				_putwch( 'N' );
				_putwch( LF );
				return 0;
			}
			else
				honk();
		}
		else
			tty_yield( 0 );

	} while ( TRUE );

	return 0;
}



int PromptUser2( LPTSTR Prompt )
{
	/************************************************************

	A more complex prompt-user routine.  This one can return:

		0			for No
		1			for Yes
		2			for All
		USER_ABORT	on Ctrl-C or Ctrl-Break or Q

	************************************************************/


	unsigned int ch		= '\0';
	wchar_t Erase[]		= L"\r                                                            \r";


	try {
		if ( Prompt )
			_cputws( Prompt );

		EatKeystrokes();

		do {
			if ( QueryKeyWaiting() ) {

				ch = GetKeystroke( 0x0230 );

				if ( ch == 'Y' ) {
					_cputws( Erase );
					return 1;
				}
				else if ( ch == 'A' ) {
					_cputws( Erase );
					return 2;
				}
				else if ( ( ch == 'N' ) || ( ch == ESC ) ) {
					_cputws( Erase );
					return 0;
				}
				else if ( ch == 'Q' ) {
					_cputws( Erase );
					return USER_ABORT;
				}
				else
					honk();
			}
			else
				tty_yield( 0 );

		}
		while ( TRUE );

	}
	catch (...) {
		return USER_ABORT;
	}

	return 0;
}



int GoodArrayName( LPTSTR ArrayName )
{
	if ( !ArrayName )
		return 0;
	else if ( ArrayName[0] == '\0' )
		return 0;

	if ( wcslen( ArrayName ) > 31 )
		return 0;

	if ( !iswalpha( ArrayName[0] ) )
		return 0;

	for ( int i = 0; ArrayName[i]; i++ )
		if ( !iswalnum( ArrayName[i] ) )
			if ( ( ArrayName[i] != '_' ) && ( ArrayName[i] != '$' ) )
				return 0;

	return 1;
}



int SafeChars( LPTSTR String )
{
	if ( !String )
		return -1;

	int k = 0;

	for ( int i = 0; String[i]; i++ ) {
		if ( ( String[i] == QUOTE ) || ( String[i] == '%' ) || ( String[i] == '[' ) || ( String[i] == ']' ) || ( String[i] == '(' ) || ( String[i] == ')' ) || ( String[i] == '^' ) || ( String[i] == '`' ) || ( String[i] == ',' ) ) {
			String[i] += 0xfee0;
			k++;
		}
	}

	return k;
}



int UnsafeChars( LPTSTR String )
{
	if ( !String )
		return -1;

	int k = 0;

	for ( int i = 0; String[i]; i++ ) {
		if ( ( String[i] >= 0xff01 ) && ( String[i] <= 0xff5e ) ) {
			String[i] -= 0xfee0;
			k++;
		}
	}

	return k;
}



int ParseNOption2( LPTSTR inString, int *Recurse, BOOL *NoColor, BOOL *WriteBOM )
{
	if ( !inString )
		return -1;
	else if ( inString[0] == '\0' )
		return -1;

	int j = 0;
	wchar_t optchar	= 'N';


	if ( ( inString[j] == '/' ) || ( inString[j] == '-' ) )
		j++;

	if ( iswalpha( inString[j] ) ) {
		optchar = towupper( inString[j] );
		j++;
	}

	if ( FH_IsColonOrEquals( inString[j] ) )
		j++;

	if ( inString[j] == '\0' ) {
		DisplayErrorHeader();
		_wcsupr_s( inString, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", inString );
		return 1;
	}


	while ( inString[j] ) {
		switch ( towupper( inString[j] ) ) {

		case 'B' :
			if ( WriteBOM ) {
				*WriteBOM = FALSE;
				break;
			}
			else
				goto Default;

		case 'C' :
			if ( NoColor ) {
				*NoColor = TRUE;
				break;
			}
			else
				goto Default;

		case 'D' :
			*Recurse |= FH_FLAG_NO_HIDDEN_DIRS;
			break;

		case 'F' :
			*Recurse |= FH_FLAG_SUPPRESS_FNF_ERROR;
			break;

		case 'J' :
			*Recurse |= FH_FLAG_NO_JUNCTIONS;
			break;

		case 'Z' :
			*Recurse |= FH_FLAG_NO_SYSTEM_DIRS;
			break;

		default :
		Default:
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown suboption: /%c:%c\r\n", optchar, towupper( inString[j] ) );
			return 1;
		}

		j++;
	}

	return 0;
}




#include "Context.cpp"
#include "UType.cpp"
#include "Unicodify.cpp"
#include "Wrap.cpp"
#include "Words.cpp"
#include "DeHtml.cpp"
#include "Obscure.cpp"
#include "SaveArray.cpp"
#include "ReplaceText.cpp"
#include "ParseArgsCmd.cpp"
#include "PasswordCmd.cpp"
#include "DedupCmd.cpp"
#include "XFilter.cpp"
#include "FFieldsCmd.cpp"
#include "Text2Clip.cpp"
#include "CopyCharsCmd.cpp"
#include "CharEncodingCmd.cpp"
#include "DegasCmd.cpp"
#include "FilterFilesCmd.cpp"
#include "CountCharsCmd.cpp"
#include "IniValue.cpp"
#include "RoughlyFunc.cpp"



struct UpendFileArgs {
	int MaxLength;
	LPTSTR LeftString;
	LPTSTR RightString;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};



int UpendFile( LPCTSTR Filename, int Flags, void *Args )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;
	else if ( !Args )
		return -1;


	_MMFile MMFile;
	size_t TextBegin						= 0;
	size_t Index							= 0;
	size_t NewIndex							= 0;
	size_t SaveIndex						= 0;
	wchar_t LineBuffer[MAX_ARG_LEN]			= L"";
	wchar_t LineNumberText[SHORT_BUF_LEN]	= L"";
	wchar_t OldLineVar[MAX_ARG_LEN]			= L"";
	wchar_t temp[SHORT_BUF_LEN]				= L"";
	wchar_t VarName[]						= L"_LINE";
	int ch									= 0;
	int ch2									= 0;
	int j									= 0;
	UpendFileArgs *MoreArgs					= (UpendFileArgs*) Args;	
	int MaxLength							= MoreArgs->MaxLength;
	LPTSTR LeftString						= MoreArgs->LeftString;
	LPTSTR RightString						= MoreArgs->RightString;
	unsigned int LineNumber					= 0;
	BOOL BlankLine							= TRUE;

	int rv									= 0;


	rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, MoreArgs->CodePage, MoreArgs->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		Printf( L"\r\nFile \"%s\" is empty.\r\n", Filename );
		return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();

	TextBegin = MMFile.BOMSize;
	Index = MMFile.FileSize;


	if ( Flags & UPEND_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	if ( MaxLength > ( MAX_ARG_LEN - 1 ) )
		MaxLength = MAX_ARG_LEN - 1;

	GetEnvironmentVariable( VarName, OldLineVar, MAX_ARG_LEN );


	while ( Index > TextBegin ) {
		ch = MMPrevChar( &MMFile, Index, &Index );

		if ( _MMIsEOL( ch ) ) {
			ch2 = MMPrevChar( &MMFile, Index, &Index );

			if ( ch == _MM2PartEOL( ch2 ) )
				MMPrevChar( &MMFile, Index, &Index );
		}

		while ( Index > TextBegin ) {
			if ( _MMIsEOL( MMPrevChar( &MMFile, Index, &NewIndex ) ) )
				break;
			else
				Index = NewIndex;
		}

		MMFile.Index	= Index;
		SaveIndex		= Index;

		LineBuffer[0] = '\0';
		j = 0;

		BlankLine = TRUE;

		for ( int i = 0; j < MAX_ARG_LEN; i++ ) {
			ch = MMGetChar( &MMFile );
			if ( _MMIsEOL( ch ) )
				break;

			LineBuffer[j++] = ch;
			LineBuffer[j] = '\0';

			if ( !iswspace( ch ) )
				BlankLine = FALSE;
		}

		//	wprintf( L"   * Point A:  LineBuffer = '%s'\n", LineBuffer );
		if ( BlankLine && ( Flags & UPEND_NO_BLANKS ) )
			continue;

		if ( Flags & UPEND_EXPAND ) {
			wsprintf( LineNumberText, L"%u", LineNumber );
			SetEnvironmentVariable( VarName, LineNumberText );
		}
		
		if ( Flags & UPEND_TRIM )
			TrimLine( LineBuffer );

		if ( Flags & UPEND_REPLACE_CONTROLS )
			ReplaceControlChars( LineBuffer );

		if ( Flags & UPEND_INVERT )
			_wcsrev( LineBuffer );

		if ( MaxLength )
			LineBuffer[MaxLength] = '\0';

		if ( LeftString ) {
			if ( LeftString[0] ) {
				if ( Flags & UPEND_EXPAND ) {
					wcscpy_s( temp, SHORT_BUF_LEN, LeftString );
					ExpandVariables( temp, 1 );
					QPuts( temp );
				}
				else
					QPuts( LeftString );
			}
		}

		QPuts( LineBuffer );

		if ( RightString ) {
			if ( RightString[0] ) {
				if ( Flags & UPEND_EXPAND ) {
					wcscpy_s( temp, SHORT_BUF_LEN, RightString );
					ExpandVariables( temp, 1 );
					QPuts( temp );
				}
				else
					QPuts( RightString );
			}
		}

		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;

		LineNumber++;
	}

	CloseMMFile( &MMFile );


	if ( Flags & UPEND_EXPAND ) {
		if ( OldLineVar[0] == '\0' )
			SetEnvironmentVariable( VarName, NULL );
		else
			SetEnvironmentVariable( VarName, OldLineVar );
	}

	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI upend( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t LeftString[SHORT_BUF_LEN]		= L"";
	wchar_t RightString[SHORT_BUF_LEN]		= L"";

	int Flags					= 0;
	int Recurse					= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth		= 0;
	unsigned int MaxDepth		= UINT_MAX;
	DWORD AttribMask			= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues			= 0;
	void *FirstRange			= 0;
	BOOL NoHighlight			= 0;
	unsigned int Pause			= 0;
	unsigned int MaxLength		= 0;
	BOOL FilenameSpecified		= FALSE;
	BOOL StdinSpecified			= FALSE;
	BOOL WriteBOM				= TRUE;
	UpendFileArgs PassArgs		= { 0, NULL, NULL, 0, 0 };
	int rv						= 0;
	int fhrv					= 0;


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( UpendHelpText, FALSE );
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( UpendHelpText, TRUE );
			}
			else if ( !_wcsicmp( arg[i], L"/B" ) )
				Flags |= UPEND_NO_BLANKS;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags |= UPEND_REPLACE_CONTROLS;
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(PassArgs.CodePage), &(PassArgs.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/E" ) )
				Flags |= UPEND_EXPAND;
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= UPEND_HEADER;
			else if ( !_wcsnicmp( arg[i], L"/L", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				LeftString[0] = '\0';
				wcscpy_s( LeftString, SHORT_BUF_LEN, arg[i] + j );
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/R", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				RightString[0] = '\0';
				wcscpy_s( RightString, SHORT_BUF_LEN, arg[i] + j );
			}
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Flags |= UPEND_TRIM;
			else if ( !_wcsicmp( arg[i], L"/V" ) )
				Flags |= UPEND_INVERT;
			else if ( !_wcsicmp( arg[i], L"/W" ) )
				MaxLength = UINT_MAX;
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &MaxLength, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad length: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				if ( MaxLength && ( ( MaxLength < 16 ) || ( MaxLength > ( MAX_ARG_LEN - 1 ) ) ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Length must be 0 or %u - %u\r\n", 16, MAX_ARG_LEN - 1 );
					goto ErrorExit;
				}
			}
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}
		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( MaxLength == UINT_MAX ) {
		MaxLength = ConsoleWidth();
		if ( MaxLength )
			MaxLength--;
	}


	PassArgs.MaxLength		= MaxLength;
	PassArgs.LeftString		= LeftString;
	PassArgs.RightString	= RightString;


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( ( ConsoleHandle( STD_INPUT_HANDLE ) == 0 ) || !FilenameSpecified ) {
			fhrv = UpendFile( (LPTSTR) StdinFilename, Flags, &PassArgs );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &PassArgs, UpendFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 1;
}



struct ShuffleFileArgs {
	unsigned int MaxLines;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};


unsigned int UnpickedLine( size_t *LineIndex, unsigned int n )
{
	unsigned int LineNumber		= 0;
	unsigned int Count			= 0;


	do {
		while ( LineIndex[LineNumber] == -1 )
			LineNumber++;

		if ( Count == n )
			break;

		Count++;
		LineNumber++;

	} while ( TRUE );

	return LineNumber;
}


int ShuffleFile( LPCTSTR Filename, int Flags, void *ShuffleArgs )
{
	if ( !Filename || !ShuffleArgs )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t LineBuffer[BIG_LINE_LEN]	= L"";
	_MMFile MMFile;
	size_t SaveIndex					= 0;
	unsigned int NumLines				= 0;
	unsigned int Count					= 0;
	size_t *LineIndex					= NULL;
	ShuffleFileArgs *MoreArgs = (ShuffleFileArgs*) ShuffleArgs;
	BOOL SkipBlanks						= ( Flags & SHUFFLE_NO_BLANKS ) > 0;
	int rv								= 0;


	rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, MoreArgs->CodePage, MoreArgs->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();


	if ( Flags & SHUFFLE_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	//		first step:  count the lines in the file

	MMFile.Index = MMFile.BOMSize;

	while ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, SkipBlanks ) == 0 )
		NumLines++;

	if ( NumLines ) {

		//		second step:  allocate buffer and get index to each line

		LineIndex = (size_t*) malloc( NumLines * sizeof( size_t ) );

		if ( LineIndex ) {

			MMFile.Index = MMFile.BOMSize;

			for ( unsigned int n = 0; n < NumLines; n++ )
				LineIndex[n] = -1;

			for ( Count = 0; Count < NumLines; ) {
				SaveIndex = MMFile.Index;
				LineBuffer[0] = '\0';

				if ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, FALSE ) )
					break;

				if ( SkipBlanks && ( LineBuffer[0] == '\0' ) )
					continue;

				LineIndex[Count++] = SaveIndex;
			}

			if ( Count ) {

				//		third step:  pick out random lines and display them until done

				for ( unsigned int n = 0; n < Count; n++ ) {

 					if ( MoreArgs->MaxLines )
						if ( n == MoreArgs->MaxLines )
							break;

					unsigned int LineNumber = UnpickedLine( LineIndex, GetRand( Count - n ) );
					MMFile.Index = LineIndex[LineNumber];

					if ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, FALSE ) == 0 ) {

						LineIndex[LineNumber] = -1;

						if ( Flags & SHUFFLE_LINE_NUMBERS ) {
							if ( Flags & SHUFFLE_ORIG_LINE_NUMBERS )
								Printf( L"%4u : ", LineNumber + 1 );
							else
								Printf( L"%4u : ", n + 1 );
						}

						QPuts( LineBuffer );
						if ( CL_EndLine() == USER_ABORT )
							goto UserAbort;

					}
				}
			}
		}
	}


	CloseMMFile( &MMFile );
	free( LineIndex );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	free( LineIndex );
	CL_ReleaseBreak();
	return USER_ABORT;
}


DLLExports int WINAPI shuffle( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	unsigned int MaxLines	= 0;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	ShuffleFileArgs PassArgs	= { 0, 0, 0 };


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( ShuffleHelpText, FALSE );
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( ShuffleHelpText, TRUE );
			}
			else if ( !_wcsicmp( arg[i], L"/B" ) )
				Flags |= SHUFFLE_NO_BLANKS;
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(PassArgs.CodePage), &(PassArgs.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = TRUE;
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= SHUFFLE_HEADER;
			else if ( !_wcsicmp( arg[i], L"/J" ) )
				Flags |= ( SHUFFLE_LINE_NUMBERS | SHUFFLE_ORIG_LINE_NUMBERS );
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				Flags = Flags & ~SHUFFLE_ORIG_LINE_NUMBERS | SHUFFLE_LINE_NUMBERS;
			else if ( !_wcsnicmp( arg[i], L"/M", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &MaxLines, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad MaxLines value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	PassArgs.MaxLines	= MaxLines;


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( ( ConsoleHandle( STD_INPUT_HANDLE ) == 0 ) || !FilenameSpecified ) {
			fhrv = ShuffleFile( (LPTSTR) StdinFilename, Flags, &PassArgs );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &PassArgs, ShuffleFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 1;
}



struct _OinkFileArgs {
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};


int OinkFile( LPCTSTR inFilename, int Flags, void *Args )
{
	if ( !inFilename )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;


	_MMFile MMFile;
	wchar_t LineBuffer[BIG_LINE_LEN] = L"";

	_OinkFileArgs *args		= (_OinkFileArgs*) Args;


	//  Open file:

	int rv = OpenMMFile( &MMFile, inFilename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, args->CodePage, args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", inFilename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();


	if ( Flags & OINK_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	while ( MMFile.Index <= MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, MMGL_FLAG_SPLIT_LONG_LINES ) )
			break;

		if ( LineBuffer[0] ) {
			Oink( LineBuffer, BIG_LINE_LEN );

			if ( Flags & OINK_FIX_QUOTES )
				FixQuotes( LineBuffer, BIG_LINE_LEN );
		}

		QPuts( LineBuffer );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}
	
	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI oink( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	_OinkFileArgs Args		= { 0, 0 };



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	CL_NoPausing();


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( OinkHelpText, FALSE );
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( OinkHelpText, TRUE );
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = TRUE;
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= OINK_HEADER;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Flags |= OINK_FIX_QUOTES;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( Flags & OINK_FIX_QUOTES )
		InitializeFixQuotesVars();

	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = OinkFile( (LPTSTR) StdinFilename, Flags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, OinkFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort :

	FH_FreeLinkedList( &FirstRange );
	return 3;
}



struct _Rot13FileArgs {
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};


int Rot13File( LPCTSTR inFilename, int Flags, void *Args )
{
	if ( !inFilename )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;


	_MMFile MMFile;
	wchar_t LineBuffer[BIG_LINE_LEN]	= L"";

	_Rot13FileArgs *args				= (_Rot13FileArgs*) Args;


	//  Open file:

	int rv = OpenMMFile( &MMFile, inFilename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, args->CodePage, args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", inFilename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();


	if ( Flags & ROT13_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	while ( MMFile.Index <= MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, MMGL_FLAG_SPLIT_LONG_LINES ) )
			break;

		if ( LineBuffer[0] )
			f_rot13( LineBuffer );

		if ( Flags & ROT13_FIX_QUOTES )
			FixQuotes( LineBuffer, BIG_LINE_LEN );

		QPuts( LineBuffer );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}
	
	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI rot13( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	_Rot13FileArgs Args		= { 0, 0 };


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( Rot13HelpText, FALSE );
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( Rot13HelpText, TRUE );
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = TRUE;
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= ROT13_HEADER;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Flags |= ROT13_FIX_QUOTES;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = Rot13File( (LPTSTR) StdinFilename, Flags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, Rot13File, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 3;
}



struct _RecaseFileArgs {
	unsigned int Mode;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};



BOOL FurrinLetter( wchar_t ch )
{
	if ( ( ch >= 0x0410 ) && ( ch <= 0x042f ) )
		return TRUE;

	if ( ( ch >= 0x0391 ) && ( ch <= 0x03a9 ) )
		return TRUE;

	if ( ( ch >= 0x03b1 ) && ( ch <= 0x03c9 ) )
		return TRUE;

	return FALSE;
}



int RecaseLineCaps( LPTSTR Line )
{
	if ( !Line )
		return -1;


	BOOL FirstLetter	= TRUE;
	int k				= 0;


	for ( wchar_t *ch = Line; *ch; ch++ ) {

		if ( iswalnum( *ch ) || FurrinLetter( *ch ) ) {
			if ( FirstLetter )
				*ch = (wchar_t) CharUpper( (LPWSTR) *ch ), k++;
				//	*ch = towupper( *ch ), k++;

			FirstLetter = FALSE;
		}
		else if ( ( *ch == APOSTROPHE ) || ( *ch == '-' ) )
			;
		else
			FirstLetter = TRUE;
	}

	return k;
}



int RecaseLineRandom( LPTSTR Line )
{
	if ( !Line )
		return -1;


	for ( wchar_t *ch = Line; *ch; ch++ ) {

		if ( GetRand( 10 ) >= 6 )
			*ch = (wchar_t) CharUpper( (LPWSTR) *ch );			//	*ch = towupper( *ch );
		else
			*ch = (wchar_t) CharLower( (LPWSTR) *ch );			//	*ch = towlower( *ch );
	}

	return 0;
}



int RecaseFile( LPCTSTR Filename, int Flags, void *Args )
{
	if ( !Filename || !Args )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	_MMFile MMFile;
	wchar_t LineBuffer[BIG_LINE_LEN]	= L"";

	_RecaseFileArgs *args		= (_RecaseFileArgs*) Args;
	int rv						= 0;


	//	Open file :

	rv = OpenMMFile( &MMFile, Filename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, args->CodePage, args->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", Filename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();


	if ( Flags & RECASE_HEADER )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	while ( MMFile.Index < MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, LineBuffer, BIG_LINE_LEN, MMGL_FLAG_SPLIT_LONG_LINES ) )
			break;

		if ( LineBuffer[0] ) {

			switch ( args->Mode ) {

			case RECASE_MODE_LOWER :

				CharLower( LineBuffer );
				//	_wcslwr_s( LineBuffer, BIG_LINE_LEN );
				break;

			case RECASE_MODE_UPPER :

				CharUpper( LineBuffer );
				//	_wcsupr_s( LineBuffer, BIG_LINE_LEN );
				break;

			case RECASE_MODE_CAPS :

				RecaseLineCaps( LineBuffer );
				break;

			case RECASE_MODE_RANDOM :

				RecaseLineRandom( LineBuffer );
				break;
			}

		}

		QPuts( LineBuffer );
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}


	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return 0;

UserAbort :

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI recase( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= 0;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	BOOL WriteBOM			= TRUE;
	int rv					= 0;
	int fhrv				= 0;

	_RecaseFileArgs Args	= { 0, 0, 0 };



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	CL_NoPausing();


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( RecaseHelpText, FALSE );
			}
			else if ( !wcscmp( arg[i], L"/??" ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( RecaseHelpText, TRUE );
			}
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(Args.CodePage), &(Args.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoHighlight = TRUE;
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= RECASE_HEADER;
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				Args.Mode = RECASE_MODE_LOWER;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption2( arg[i], &Recurse, &NoHighlight, &WriteBOM ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Args.Mode = RECASE_MODE_RANDOM;
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				Args.Mode = RECASE_MODE_CAPS;
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				Args.Mode = RECASE_MODE_UPPER;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( wcscmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );

	if ( Args.Mode == 0 )
		Args.Mode = RECASE_MODE_UPPER;


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( !ConsoleHandle( STD_INPUT_HANDLE ) || !FilenameSpecified ) {
			fhrv = RecaseFile( (LPTSTR) StdinFilename, Flags, &Args );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &Args, RecaseFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit :

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort :

	FH_FreeLinkedList( &FirstRange );
	return 3;
}



/*
DLLExports int WINAPI _cciso2( LPTSTR outString )
{
	if ( !outString )
		return 1;


	GEOID GeoId			= NULL;


	outString[0] = '\0';

	GeoId = GetUserGeoID( GEOCLASS_NATION );

	if ( GeoId != GEOID_NOT_AVAILABLE )
		GetGeoInfo( GeoId, GEO_ISO2, outString, SHORT_BUF_LEN, 0 );

	return 0;
}
*/


#include "NewHelp.cpp"


const int NumHelpTexts		= 27;

const wchar_t HelpFeature[NumHelpTexts][14]	= {
	L"charencoding", L"clip2text", L"context", L"copychars", L"countchars", L"dedup", L"degas", L"dehtml",
	L"ffields", L"filterfiles", L"loadarray", L"oink", L"parseargs", L"password", L"recase", L"replacetext",
	L"rot13", L"savearray", L"shuffle", L"text2clip", L"textutilshelp", L"unicodify", L"upend",
	L"utype", L"words", L"wrap", L"xfilter" };

const wchar_t *HelpText[NumHelpTexts]		= { 
	CharEncodingHelpText, Clip2TextHelpText, ContextHelpText, CopyCharsHelpText, CountCharsHelpText, DedupHelpText, DegasHelpText, DeHTMLHelpText,
	FFieldsHelpText, FilterFilesHelpText, LoadArrayHelpText, OinkHelpText, ParseArgsHelpText, PasswordHelpText, RecaseHelpText, ReplaceTextHelpText,
	Rot13HelpText, SaveArrayHelpText, ShuffleHelpText, Text2ClipHelpText, TextUtilsHelpHelpText, UnicodifyHelpText, UpendHelpText,
	UtypeHelpText, WordsHelpText, WrapHelpText, XFilterHelpText };



DLLExports int WINAPI textutilshelp( LPTSTR inString )
{
	return PluginHelpCmd( inString, TextUtilsHelpHelpText );
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	if ( !HasBinaryBuffers() ) {
		DisableOneFeature( PluginFeatures, L"@bintob85" );
		DisableOneFeature( PluginFeatures, L"@b85tobin" );
	}

	if ( !ArraysAvailable() ) {
		DisableOneFeature( PluginFeatures, L"loadarray" );
		DisableOneFeature( PluginFeatures, L"savearray" );
		DisableOneFeature( PluginFeatures, L"parseargs" );
	}

	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/textutils.html";
	piInfo.pszDescription	= (LPTSTR) L"Sundry text-related features.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	SeedRandomGenerator();
	GetShellVersion();
	CL_Initialize();
	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	CL_Shutdown();
	CloseHelpWindow();
	return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}
