
//			WORDS command, and associated variables:


#ifndef WORDS_COMMAND_VER
#define WORDS_COMMAND_VER				1.0.1.1


#define CW_FLAG_REMOVE_HYPHENS			0x0010
#define CW_FLAG_REJECT_NUMBERS			0x0020
#define CW_FLAG_REJECT_INITIAL_DIGITS	0x0040
#define CW_FLAG_ALLOW_UNDERSCORES		0x0080
#define CW_FLAG_ALLOW_DOLLAR_SIGNS		0x0100
#define CW_FLAG_CODE_MODE				0x1000
#define CW_FLAG_WC_INCLUDES_BOM			0x4000
#define CW_FLAG_DUMP_WORDS_LIST			0x8000

#define UNIQUEMODE_EACH_FILE			1
#define UNIQUEMODE_ALL_FILES			2

#define CASE_ALL_CAPS					0x20000
#define CASE_ALL_LOWER					0x10000
#define CASE_FIRST_FEW					0x0ffff



const wchar_t WordsHelpText[] = L"Count words, sentences, and paragraphs in English text.\n\
\n\
WORDS /CP:n /D /F:n /K /M:n /N /P /S /U:n /X filename.... \r\n\
\n\
\t/CP:  interpret OEM text as code page n\n\
\t/D    dump word lists\n\
\t/F:   input text format:\n\
\t       0: best guess (default)\n\
\t       1: unformatted\n\
\t       2: prewrapped\n\
\t/K    keep hyphens when reassembling split words\n\
\t/M:   minimum number of letters in a word\n\
\t/N    no words containing digits\n\
\t/P    page output\n\
\t/S    search in subdirectories for matching files\n\
\t/U:   count unique words for:\n\
\t       0: none (fastest)\n\
\t       1: each file individually (default)\n\
\t       2: all files together (slower)\n\
\t       3: each file and all files together (slowest)\n\
\t/X    no words beginning with a digit\n\
\n\
Input filenames may be specified on the command line, or text may \n\
be piped or redirected into WORDS.\n";


wchar_t CW_PluralStr[2] = L"s";
wchar_t CW_PrewrappedStr[2][12] = { L"unformatted", L"prewrapped" };



struct WordEntry {
	WordEntry *Next;
	unsigned long int Count;
	unsigned int Flags;
	wchar_t Word;
};

struct CountWordsArgs {
	unsigned int UniqueMode;
	unsigned int UserFormat;
	unsigned int MinLetters;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};


WordEntry *FirstWordAll = NULL;

int WordCountsValid = 0;


unsigned long int TotalWords = 0, UniqueWords = 0, ProperNouns = 0, CountRuns = 0;
unsigned long int TotalSentenceWords = 0, CountSentencesD = 0, CountSentencesE = 0, CountSentencesQ = 0;
unsigned long int CountGrafs = 0, CountTitles = 0, GrafSentences = 0;
unsigned long int CountLines = 0, NonEmptyLines = 0, LongestLine = 0;
unsigned long int CountChars = 0;

int TotalCountsValid = 0;

unsigned long int UniqueWordsAll = 0, FilesProcessed = 0;
size_t GrandTotalBytes = 0;
unsigned long int GrandTotalChars = 0;
unsigned long int GrandTotalWords = 0, GrandTotalProperNouns = 0, GrandTotalRuns = 0;
unsigned long int GrandTotalSentencesD = 0, GrandTotalSentencesE = 0, GrandTotalSentencesQ = 0;
unsigned long int GrandTotalSentenceWords = 0;
unsigned long int GrandTotalLines = 0, GrandTotalNonEmptyLines = 0, GrandTotalLongestLine = 0;
unsigned long int GrandTotalGrafs = 0, GrandTotalTitles = 0;





wchar_t CW_MakeLower( wchar_t ch )
{
	if ( ( ch >= 0xc0 ) && ( ch <= 0xde ) && ( ch != 0xd7 ) )
		return ch + 0x20;

	return towlower( ch );
}


wchar_t CW_MakeUpper( wchar_t ch )
{
	if ( ( ch >= 0xe0 ) && ( ch <= 0xfe ) && ( ch != 0xf7 ) )
		return ch - 0x20;

	return towupper( ch );
}


int CW_Indent( int n )
{
	for ( int i = 0; i < n; i++ )
		QPuts( L" " );

	return 0;
}


int CW_DumpWordsByFreq( WordEntry **FirstWord )
{
	if ( *FirstWord == NULL )
		return -1;


	wchar_t Recased[SHORT_BUF_LEN]	= L"";
	WordEntry *Word			= *FirstWord;
	unsigned long int Freq	= 0;
	unsigned int mask		= 0;
	int col					= 0;
	int indent				= 0;
	int len					= 0;
	

	int Width = ConsoleWidth();
	if ( !Width )
		Width = DEFAULT_COLUMNS;


	do {

		if ( Word->Count != Freq ) {
			Freq = Word->Count;
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;

			wsprintf( Recased, L"%lu: ", Freq );
			col = (int) wcslen( Recased );
			QPuts( Recased );
			indent = col + 1;
		}

		len = (int) wcslen( &(Word->Word) );
		wcscpy_s( Recased, SHORT_BUF_LEN, &(Word->Word) );
		mask = 1;

		for ( int i = 0; Recased[i]; i++ ) {
			if ( iswalpha( Recased[i] ) ) {
				if ( (Word->Flags) & ( mask | CASE_ALL_CAPS ) )
					Recased[i] = CW_MakeUpper( Recased[i] );

				mask = ( mask << 1 ) & CASE_FIRST_FEW;
			}
		}

		if ( ( col + len + 1 ) < Width ) {
			Printf( L" %s", Recased );
			col += len + 1;
		}
		else {
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;
			
			CW_Indent( indent );
			Printf( Recased );
			col = len + indent;
		}
		Word = Word->Next;

	} while ( Word != NULL );

	return CL_EndLine();
}


unsigned int CW_CaseCheck( LPTSTR inString )
{
	if ( !inString )
		return 0;
	else if ( inString[0] == '\0' )
		return 0;

	wchar_t ch = '\0';
	unsigned int mask = 1, rv = 0;
	BOOL AllCaps = TRUE, AllLower = TRUE;

	for ( int i = 0; inString[i]; i++ ) {
		ch = inString[i];
		if ( iswalpha( ch ) ) {
			if ( iswlower( ch ) )
				AllCaps = FALSE;
			else {
				AllLower = FALSE;
				rv |= mask;
			}

			mask = ( mask << 1 ) & CASE_FIRST_FEW;
		}
	}

	if ( AllCaps )
		rv |= CASE_ALL_CAPS;
	else if ( AllLower )
		rv |= CASE_ALL_LOWER;

	return rv;
}


int CW_CopyAsLower( LPTSTR inString, LPTSTR outString )
{
	size_t i = 0;
	wchar_t ch = '\0';

	do {
		ch = CW_MakeLower( inString[i] );
		outString[i++] = ch;
	} while ( ch );

	return 0;
}


WordEntry *CW_CreateWordEntry( LPTSTR Word, unsigned int Flags )
{
	if ( !Word )
		return NULL;
	else if ( Word[0] == '\0' )
		return NULL;


	int len = (int) wcslen( Word );

	WordEntry *NewWord = (WordEntry*) malloc( sizeof( WordEntry ) + ( len * sizeof( wchar_t ) ) );
	if ( !NewWord ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error allocating memory\n" );
		return NULL;
	}

	wcscpy_s( &(NewWord->Word), len + 1, Word );
	NewWord->Next  = NULL;
	NewWord->Count = 1;
	NewWord->Flags = Flags;
	return NewWord;
}


int CW_InsertWord( LPTSTR inWord, WordEntry **FirstWord, unsigned long int *Counter, unsigned long int *ProperCount )
{
	if ( !inWord )
		return -1;
	else if ( inWord[0] == '\0' )
		return -1;


	wchar_t WordBuf[SHORT_BUF_LEN] = L"";


	WordEntry *NewWord = NULL, *SearchWord = NULL;
	WordEntry *PrevWord = NULL, *NextWord = NULL;
	WordEntry *InsertAfter = NULL;
	unsigned int Case = 0;
	int cmp = 0;

	CW_CopyAsLower( inWord, WordBuf );
	Case = CW_CaseCheck( inWord );

	//  wprintf( L"   * Insert word \"%s\"  %u\n", inWord, Case );

	if ( *FirstWord == NULL ) {
		*FirstWord = CW_CreateWordEntry( WordBuf, Case );
		if ( *FirstWord == NULL )
			return -2;
		
		(*Counter)++;

		if ( Case != CASE_ALL_LOWER )
			(*ProperCount)++;

		return 0;
	}

	SearchWord = *FirstWord;

	do {

		cmp = wcscmp( WordBuf, &(SearchWord->Word) );

		if ( cmp == 0 ) {
			(SearchWord->Count)++;

			if ( ( Case == CASE_ALL_LOWER ) && ( SearchWord->Flags != CASE_ALL_LOWER ) ) {
				SearchWord->Flags = CASE_ALL_LOWER;
				(*ProperCount)--;
			}
			else if ( !( Case & CASE_ALL_CAPS ) && ( SearchWord->Flags & CASE_ALL_CAPS ) )
				SearchWord->Flags = Case;

			//  move word forward in list if needed:

			if ( SearchWord == *FirstWord )
				return 0;
			else if ( PrevWord->Count > SearchWord->Count )
				return 0;

			if ( SearchWord->Count == PrevWord->Count )
				if ( wcscmp( WordBuf, &(PrevWord->Word) ) > 0 )
					return 0;

			if ( ( SearchWord->Count > (*FirstWord)->Count ) || ( ( SearchWord->Count == (*FirstWord)->Count ) && ( wcscmp( WordBuf, &((*FirstWord)->Word) ) < 0 ) ) ) {
				PrevWord->Next = SearchWord->Next;
				SearchWord->Next = *FirstWord;
				*FirstWord = SearchWord;
				return 0;
			}

			InsertAfter = *FirstWord;

			do {
				NextWord = InsertAfter->Next;
				if ( NextWord == NULL )
					break;

				if ( NextWord->Count < SearchWord->Count )
					break;

				if ( NextWord->Count == SearchWord->Count )
					if ( wcscmp( WordBuf, &(NextWord->Word) ) < 0 )
						break;

				InsertAfter = NextWord;
			} while ( TRUE );

			PrevWord->Next = SearchWord->Next;
			SearchWord->Next = InsertAfter->Next;
			InsertAfter->Next = SearchWord;
			return 0;
		}
		else if ( ( cmp < 0 ) && ( SearchWord->Count == 1 ) ) {
			NewWord = CW_CreateWordEntry( WordBuf, Case );
			if ( NewWord == NULL )
				return -2;

			(*Counter)++;

			if ( Case != CASE_ALL_LOWER )
				(*ProperCount)++;

			NewWord->Next = SearchWord;
			if ( PrevWord == NULL )
				*FirstWord = NewWord;
			else
				PrevWord->Next = NewWord;
			return 0;
		}
		else if ( SearchWord->Next == NULL ) {
			NewWord = CW_CreateWordEntry( WordBuf, Case );
			if ( NewWord == NULL )
				return -2;

			(*Counter)++;

			if ( Case != CASE_ALL_LOWER )
				(*ProperCount)++;

			SearchWord->Next = NewWord;
			return 0;
		}

		PrevWord = SearchWord;
		SearchWord = SearchWord->Next;

	} while ( TRUE );
	
	return -1;
}


int CW_FreeList( WordEntry **FirstWord )
{
	WordEntry *NextWord;

	while ( *FirstWord != NULL ) {
		NextWord = (*FirstWord)->Next;
		free( *FirstWord );
		*FirstWord = NextWord;
	}

	return 0;
}


LPTSTR CW_Plural( unsigned long int Value )
{
	if ( Value == 1 )
		return CW_PluralStr + 1;
	else
		return CW_PluralStr;
}


LPTSTR CW_Plural( size_t Value )
{
	if ( Value == 1 )
		return CW_PluralStr + 1;
	else
		return CW_PluralStr;
}


LPTSTR CW_Plural( double Value )
{
	if ( ( Value > 0.0 ) && ( Value <= 1.0 ) )
		return CW_PluralStr + 1;
	else
		return CW_PluralStr;
}


LPTSTR CW_PrewrappedString( BOOL Value )
{
	return CW_PrewrappedStr[Value];
}



int CW_ShowReport( LPCTSTR Filename, size_t CountBytes, unsigned long int CountChars, unsigned long int TotalWords, unsigned long int UniqueWords, unsigned long int ProperNouns, unsigned long int CountRuns, unsigned long int SentencesD, unsigned long int SentencesE, unsigned long int SentencesQ, unsigned long int TotalSentenceWords, unsigned long int Lines, unsigned long int NonEmptyLines, unsigned long int LongestLine, unsigned long int Grafs, unsigned long int Titles, int FileEncoding, int BOMSize, BOOL Prewrapped, BOOL CodeMode )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	unsigned long int Sentences = SentencesD + SentencesE + SentencesQ;
	double AverageSentence = 0.0, AverageGraf = 0.0;
	int rv			= 0;


	if ( Sentences )
		AverageSentence = (double) TotalSentenceWords / Sentences;

	if ( Grafs )
		AverageGraf = (double) Sentences / Grafs;


	if ( CL_EndLine() == USER_ABORT )
		return USER_ABORT;

	if ( !Filename )
		rv = HLputs( L"GRAND TOTAL:", 1 );
	else if ( Filename[0] == '\0' )
		rv = HLputs( L"GRAND TOTAL:", 1 );
	else if ( !_wcsicmp( Filename, ClipFilename ) )
		rv = HLputs( L"<clipboard>", 1 );
	else if ( !_wcsicmp( Filename, StdinFilename ) )
		rv = HLputs( L"<stdin>", 1 );
	else
		rv = HLprintf( L"\x12\rFile \"%s\" :\n", Filename );

	if ( rv == USER_ABORT )
		return USER_ABORT;

	Printf( L"  %lu word%s total", TotalWords, CW_Plural( TotalWords ) );

	if ( UniqueWords != ULONG_MAX )
		Printf( L", %lu unique, %lu proper", UniqueWords, ProperNouns );

	Printf( L".  %lu run%s of non-blanks.", CountRuns, CW_Plural( CountRuns ) );
	if ( CL_EndLine() == USER_ABORT )
		return USER_ABORT;

	if ( !CodeMode ) {
		Printf( L"  %lu sentence%s total:  %lu.  %lu!  %lu?   Average sentence %.1f word%s.", Sentences, CW_Plural( Sentences ), SentencesD, SentencesE, SentencesQ, AverageSentence, CW_Plural( AverageSentence ) );
		if ( CL_EndLine() == USER_ABORT )
			return USER_ABORT;

		Printf( L"  %lu paragraph%s, %lu title%s.  Average paragraph %.1f sentence%s.", Grafs, CW_Plural( Grafs ), Titles, CW_Plural( Titles ), AverageGraf, CW_Plural( AverageGraf ) );
		if ( CL_EndLine() == USER_ABORT )
			return USER_ABORT;
	}

	Printf( L"  %lu line%s total, %lu not blank; the longest had %lu character%s.", Lines, CW_Plural( Lines ), NonEmptyLines, LongestLine, CW_Plural( LongestLine ) );
	if ( CL_EndLine() == USER_ABORT )
		return USER_ABORT;

	Printf( L"  %lu character%s in %lu byte%s", CountChars, CW_Plural( CountChars ), CountBytes, CW_Plural( CountBytes ) );

	if ( FileEncoding >= 0 ) {
		MM_GetEncodingString( FileEncoding, BOMSize, temp, SHORT_BUF_LEN );
		Printf( L" (%s, %s).", temp, CW_PrewrappedString( Prewrapped ) );
	}
	else
		QPuts( L"." );

	if ( CL_EndLine() == USER_ABORT )
		return USER_ABORT;

	return 0;
}


int CW_EndOfSentence( _MMFile *MMFile, size_t Pointer, wchar_t Char )
{
	int ch = '\0';
	unsigned int MaxCRs = 0, MaxLFs = 0;
	int Type = 1;

	if ( Char == '!' )
		Type = 2;
	else if ( ( Char == '?' ) || ( Char == INTERROBANG ) )
		Type = 3;

	if ( MMFile->FileFormat == FMT_PREWRAPPED )
		MaxCRs = MaxLFs = 1;

	for ( unsigned int i = 0; i < 30; i++ ) {
		ch = MMPeekAhead( MMFile, i );

		if ( ch < 0 )
			return Type;
		else if ( iswalpha( ch ) ) {
			if ( iswlower( ch ) )
				return 0;
			else
				return Type;
		}
		else if ( ( Type == 1 ) && ( ch == '!' ) )
			Type = 2;
		else if ( ( Type == 1 ) && ( ( ch == '?' ) || ( ch == INTERROBANG ) ) )
			Type = 3;
		else if ( ch == CR ) {
			if ( MaxCRs-- == 0 )
				return Type;
		}
		else if ( ch == LF ) {
			if ( MaxLFs-- == 0 )
				return Type;
		}
	}

	return FALSE;
}



int CW_CountWords( LPCTSTR inFilename, int Flags, void *CountArgs )
{
	if ( !inFilename )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;


	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";
	wchar_t WordBuf[SHORT_BUF_LEN]		= L"";
	wchar_t AposWord[SHORT_BUF_LEN]		= L"";

	CountWordsArgs *MoreArgs = (CountWordsArgs*) CountArgs;

	_MMFile MMFile;
	int ch						= 0;
	int lastch					= 0;
	int WordLen					= 0;
	int WordPeriods				= 0;
	int WordLetters				= 0;
	int WordVowels				= 0;
	BOOL InSingleQuotes			= FALSE;
	int SaveEncoding			= 0;
	size_t SaveFileSize			= 0;
	size_t SaveBOMSize = 0;

	unsigned int UniqueMode		= MoreArgs->UniqueMode;
	unsigned int UserFormat		= MoreArgs->UserFormat;
	BOOL RemoveHyphens			= Flags & CW_FLAG_REMOVE_HYPHENS;
	BOOL RejectNumbers			= Flags & CW_FLAG_REJECT_NUMBERS;
	BOOL RejectInitialDigits	= Flags & CW_FLAG_REJECT_INITIAL_DIGITS;
	BOOL AllowUnderscores		= Flags & CW_FLAG_ALLOW_UNDERSCORES;
	BOOL AllowDollarSigns		= Flags & CW_FLAG_ALLOW_DOLLAR_SIGNS;
	BOOL CodeMode				= Flags & CW_FLAG_CODE_MODE;
	BOOL Dump					= Flags & CW_FLAG_DUMP_WORDS_LIST;
	BOOL WCIncludesBOM			= Flags & CW_FLAG_WC_INCLUDES_BOM;
	unsigned int LineChars		= 0;
	BOOL BlankLine				= TRUE;

	WordEntry *FirstWord		= NULL;

	TotalWords = 0, UniqueWords = 0, ProperNouns = 0, CountRuns = 0;
	CountGrafs = 0, CountTitles = 0, GrafSentences = 0;
	CountLines = 0, NonEmptyLines = 0, LongestLine = 0;

	unsigned long int SentenceWords = 0;
	TotalSentenceWords = 0;
	CountSentencesD = CountSentencesE = CountSentencesQ = 0;

	BOOL GrafWords = FALSE;

	BOOL FileIsPrewrapped = FALSE;

	int ch2 = 0;
	WordCountsValid = 0;
	BOOL ShowProgress = FALSE;


	//  Open file:

	int rv = OpenMMFile( &MMFile, inFilename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC | MM_FLAG_HIGH_CHARS, MoreArgs->CodePage, MoreArgs->EbcdicCodePage );
	CountChars = 0;

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", inFilename ) == USER_ABORT )
			return USER_ABORT;
		return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();

	SaveEncoding	= MMFile.FileEncoding;
	SaveFileSize	= MMFile.FileSize;
	SaveBOMSize		= MMFile.BOMSize;


	if ( UserFormat )
		MMFile.FileFormat = UserFormat;
	else
		GuessFileFormat( &MMFile );

	FileIsPrewrapped = ( MMFile.FileFormat == FMT_PREWRAPPED );

	if ( MMFile.BOMSize && WCIncludesBOM ) {
		ch = MMPeekAhead( &MMFile, 0 );
		if ( ( ch <= 0 ) || iswspace( ch ) )
			CountRuns++;
	}


	if ( ConsoleHandle( STD_OUTPUT_HANDLE ) )
		ShowProgress = TRUE;


	//  Loop through the contents of the file:

	ch = SPACE;

	while ( MMFile.Index <= MMFile.FileSize ) {
		lastch = ch;

		ch = MMGetChar( &MMFile );

		if ( ch > 0xffff )
			ch = MM_INVALID_CHAR;

		if ( ch >= 0 )
			CountChars++;

		if ( ShowProgress && ( MMFile.Index >= 0x40000 ) )
			if ( ( MMFile.Index & 0x01fff ) == 0 )
				_cwprintf( L"\r%5.1f%% ", 100.0 * MMFile.Index / MMFile.FileSize );

		if ( LineChars && ( ch < 0 ) )
			ch = CR;

		if ( ch < 0 )
			break;						//  abort on error or eof
		else if ( ch == '\0' )
			continue;					//  ignore any nulls in input text

		if ( !iswspace( ch ) && iswspace( lastch ) )
			CountRuns++;

		if ( iswgraph( ch ) )
			BlankLine = FALSE;

		if ( ( ch == SHYPHEN ) && ( iswalnum( lastch ) ) )
			if ( iswalnum( MMPeekAhead( &MMFile, 0 ) ) )
				continue;

		if ( ( ch == HYPHEN2 ) || ( ch == SHYPHEN ) )
			ch = HYPHEN;

		if ( ( ch == LSQUOTE ) || ( ch == RSQUOTE ) || ( ch == APOSTROPHE ) ) {
			ch = APOSTROPHE;
			wchar_t nextch = MMPeekAhead( &MMFile, 0 );

			if ( iswgraph( lastch ) && ( lastch != LDQUOTE ) && ( lastch != QUOTE ) ) {
				ch = RSQUOTE;

				if ( iswalnum( lastch ) && ( iswalpha( nextch ) ) )
					ch = APOSTROPHE;

				if ( !InSingleQuotes && ( iswalpha( lastch ) ) && ( ch == RSQUOTE ) )
					ch = APOSTROPHE;

				if ( ch == RSQUOTE )
					InSingleQuotes = FALSE;

			}
			else {
				ch = LSQUOTE;

				if ( _MMIsEOL( nextch ) )
					ch = RSQUOTE;
			}

			if ( ( iswalnum( lastch ) == 0 ) && ( iswalpha( nextch ) ) ) {
				AposWord[0] = APOSTROPHE;
				int peek = 0;
				for ( int j = 0; j < 15; j++ ) {
					peek = MMPeekAhead( &MMFile, j );

					if ( peek < SPACE )
						peek = SPACE;
					else if ( peek > 0xffff )
						peek = MM_INVALID_CHAR;

					AposWord[j + 1] = peek;
				}

				AposWord[16] = '\0';
				if ( IsWordBeginningWithApostrophe( AposWord ) )
					ch = APOSTROPHE;
			}

			if ( ch == LSQUOTE )
				InSingleQuotes = TRUE;
		}

		if ( !InSingleQuotes && ( ch == '`' ) && ( iswgraph( lastch ) == 0 ) ) {
			if ( iswalnum( MMPeekAhead( &MMFile, 0 ) ) )
				InSingleQuotes = TRUE;
			else if ( MMPeekAhead( &MMFile, 0 ) == '`' )
				if ( iswalnum( MMPeekAhead( &MMFile, 1 ) ) )
					InSingleQuotes = TRUE;
		}
		
		if ( ch == '.' ) {
			if ( WordLen ) {
				if ( ( WordVowels > 0 ) && ( WordPeriods == 0 ) )
					if ( iswalpha( MMPeekAhead( &MMFile, 0 ) ) == 0 )
						ch = FWPERIOD;

				if ( ( ch == FWPERIOD ) && !iswalpha( MMPeekAhead( &MMFile, 0 ) ) ) {
					if ( _wcsicmp( WordBuf, L"etc" ) == 0 )
						ch = '.';
					else if ( _wcsicmp( WordBuf, L"viz" ) == 0 )
						ch = '.';
				}

				if ( ch == '.' )
					WordPeriods++;
			}
			else
				ch = FWPERIOD;
		}

		if ( WordLen && ( ch == HYPHEN ) )
			if ( iswpunct( MMPeekAhead( &MMFile, 0 ) ) )
				ch = HYPHEN2;

		if ( _MMIsEOL( ch ) && WordLen && FileIsPrewrapped && ( lastch == HYPHEN ) ) {
			int i = 0;

			if ( MMPeekAhead( &MMFile, i ) == _MM2PartEOL( ch ) )
				i++;

			while ( ( i < 20 ) && IsSpaceOrTab( MMPeekAhead( &MMFile, i ) ) )
				i++;

			if ( iswalnum( MMPeekAhead( &MMFile, i ) ) ) {

				for ( int j = 0; j < i; j++ ) {
					MMGetChar( &MMFile );
					CountChars++;
				}

				if ( LineChars > LongestLine )
					LongestLine = LineChars;

				CountLines++;
				NonEmptyLines++;
				ch = HYPHEN;
				LineChars = i;
				CountRuns++;

				if ( RemoveHyphens && ( WordLen > 1 ) ) {
					WordLen--;
					WordBuf[WordLen] = '\0';
					ch = WordBuf[WordLen - 1];
				}

				continue;
			}
		}

		if ( iswalnum( ch ) || ( ch == APOSTROPHE ) || ( AllowUnderscores && ( ch == '_' ) ) || ( ( WordLen > 0 ) && ( ( ch == '.' ) || ( ch == HYPHEN ) || ( AllowDollarSigns && ( ch == '$' ) ) ) ) ) {
			if ( WordLen < ( SHORT_BUF_LEN - 1 ) ) {
				WordBuf[WordLen++] = ch;
				WordBuf[WordLen] = '\0';

				if ( iswalpha( ch ) )
					WordLetters++;

				if ( IsVowel( ch ) )
					WordVowels++;
			}

		}
		else if ( WordLen ) {
			if ( RejectInitialDigits )
				if ( iswdigit( WordBuf[0] ) )
					WordLetters = 0;

			if ( RejectNumbers )
				if ( HasDigits( WordBuf ) )
					WordLetters = 0;

			if ( WordLetters ) {
				if ( WordLen > 1 )
					if ( WordBuf[WordLen - 1] == HYPHEN ) {
						WordLen--;
						WordBuf[WordLen] = '\0';
					}

				if ( (unsigned int) WordLetters >= MoreArgs->MinLetters ) {
					TotalWords++;
					SentenceWords++;
					GrafWords = TRUE;

					if ( UniqueMode & UNIQUEMODE_EACH_FILE )
						CW_InsertWord( WordBuf, &FirstWord, &UniqueWords, &ProperNouns );

					if ( UniqueMode & UNIQUEMODE_ALL_FILES )
						CW_InsertWord( WordBuf, &FirstWordAll, &UniqueWordsAll, &GrandTotalProperNouns );
				}
			}

			WordLen = 0;
			WordBuf[0] = '\0';
			WordLetters = WordVowels = WordPeriods = 0;
		}

		if ( SentenceWords && ( WordLen == 0 ) && ( IsTerminalPunct( ch ) || ( ch == FWPERIOD ) || ( ch == '.' ) ) ) {
			int k = CW_EndOfSentence( &MMFile, MMFile.Index, ch );

			if ( k == 1 )
				CountSentencesD++;
			else if ( k == 2 )
				CountSentencesE++;
			else if ( k == 3 )
				CountSentencesQ++;

			if ( k ) {
				GrafSentences++;
				TotalSentenceWords += SentenceWords;
				SentenceWords = 0;
			}
		}

		if ( _MMIsEOL( ch ) ) {
			if ( MMPeekAhead( &MMFile, 0 ) == _MM2PartEOL( ch ) ) {
				MMGetChar( &MMFile );
				CountChars++;
			}

			CountLines++;

			if ( !FileIsPrewrapped || BlankLine )
				if ( GrafWords ) {
					if ( GrafSentences )
						CountGrafs++;
					else
						CountTitles++;

					GrafSentences = 0;
					GrafWords = FALSE;
				}

			if ( BlankLine ) {
				InSingleQuotes = FALSE;
				SentenceWords = 0;
			}
			else
				NonEmptyLines++;

			if ( LineChars > LongestLine )
				LongestLine = LineChars;

			LineChars = 0;
			BlankLine = TRUE;
		}
		else
			LineChars++;

	}

	if ( GrafWords ) {
		if ( GrafSentences )
			CountGrafs++;
		else
			CountTitles++;

		GrafSentences = 0;
		GrafWords = FALSE;
	}


	//  Close file handle:

	CloseMMFile( &MMFile );


	//  Report:

	if ( ShowProgress )
		_cputws( L"\r          \r" );

	if ( CW_ShowReport( inFilename, SaveFileSize - SaveBOMSize, CountChars, TotalWords, ( UniqueMode & UNIQUEMODE_EACH_FILE ) ? UniqueWords : ULONG_MAX, ProperNouns, CountRuns, CountSentencesD, CountSentencesE, CountSentencesQ, TotalSentenceWords, CountLines, NonEmptyLines, LongestLine, CountGrafs, CountTitles, SaveEncoding, (int) SaveBOMSize, FileIsPrewrapped, CodeMode ) == USER_ABORT )
		goto UserAbort;

	if ( Dump )
		CW_DumpWordsByFreq( &FirstWord );


	//  Update grand-total variables:

	FilesProcessed++;
	GrandTotalBytes += ( SaveFileSize - SaveBOMSize );
	GrandTotalChars += CountChars;
	GrandTotalWords += TotalWords;
	GrandTotalRuns += CountRuns;
	GrandTotalSentencesD += CountSentencesD;
	GrandTotalSentencesE += CountSentencesE;
	GrandTotalSentencesQ += CountSentencesQ;
	GrandTotalSentenceWords += TotalSentenceWords;
	GrandTotalLines += CountLines;
	GrandTotalNonEmptyLines += NonEmptyLines;
	if ( LongestLine > GrandTotalLongestLine )
		GrandTotalLongestLine = LongestLine;
	GrandTotalGrafs += CountGrafs;
	GrandTotalTitles += CountTitles;


	//  Clean up:

	CW_FreeList( &FirstWord );
	WordCountsValid = ( 1024 | UniqueMode );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CW_FreeList( &FirstWord );
	WordCountsValid = ( 1024 | UniqueMode );
	CL_ReleaseBreak();
	return USER_ABORT;
}



DLLExports int WINAPI words( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags				= CW_FLAG_REMOVE_HYPHENS | CW_FLAG_WC_INCLUDES_BOM;
	int Recurse				= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth	= 0;
	unsigned int MaxDepth	= UINT_MAX;
	DWORD AttribMask		= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues		= 0;
	void *FirstRange		= NULL;
	BOOL NoHighlight		= FALSE;
	unsigned int Pause		= 0;

	unsigned int UniqueMode	= UNIQUEMODE_EACH_FILE;
	unsigned int UserFormat	= FMT_UNKNOWN;
	unsigned int MinLetters	= 0;
	BOOL WriteBOM			= TRUE;
	BOOL FilenameSpecified	= FALSE;
	BOOL StdinSpecified		= FALSE;
	int rv					= 0;
	int fhrv				= 0;

	CountWordsArgs PassArgs;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	UniqueWordsAll = 0, FilesProcessed = 0;
	GrandTotalBytes = 0, GrandTotalChars = 0;
	GrandTotalWords = 0, GrandTotalProperNouns = 0, GrandTotalRuns = 0;
	GrandTotalSentencesD = 0, GrandTotalSentencesE = 0, GrandTotalSentencesQ = 0;
	GrandTotalSentenceWords = 0;
	GrandTotalLines = 0, GrandTotalNonEmptyLines = 0, GrandTotalLongestLine = 0;
	GrandTotalGrafs = 0, GrandTotalTitles = 0;

	TotalCountsValid = 0;


	CL_NoPausing();


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				CW_FreeList( &FirstWordAll );
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( WordsHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( WORDS_COMMAND_VER ) );
			}
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Flags = ( Flags & ~CW_FLAG_REJECT_NUMBERS ) | CW_FLAG_REJECT_INITIAL_DIGITS | CW_FLAG_ALLOW_UNDERSCORES | CW_FLAG_ALLOW_DOLLAR_SIGNS | CW_FLAG_CODE_MODE;
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(PassArgs.CodePage), &(PassArgs.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				Flags |= CW_FLAG_DUMP_WORDS_LIST;
			else if ( !_wcsicmp( arg[i], L"/F" ) )
				UserFormat = FMT_UNKNOWN;
			else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &UserFormat, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad UserFormat value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
				else if ( UserFormat > FMT_MAX_VALUE ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"UserFormat must be 0 - %u\r\n", FMT_MAX_VALUE );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/K" ) )
				Flags &= ~CW_FLAG_REMOVE_HYPHENS;
			else if ( !_wcsnicmp( arg[i], L"/M", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &MinLetters, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad MinLetters value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/N" ) )
				Flags |= CW_FLAG_REJECT_NUMBERS;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				while ( arg[i][j] ) {
					switch ( towupper( arg[i][j] ) ) {
						case 'B' :
							WriteBOM = FALSE;
							break;

						case 'C' :
							NoHighlight = TRUE;
							break;

						case 'D' :
							Recurse |= FH_FLAG_NO_HIDDEN_DIRS;
							break;

						case 'F' :
							Recurse |= FH_FLAG_SUPPRESS_FNF_ERROR;
							break;

						case 'J' :
							Recurse |= FH_FLAG_NO_JUNCTIONS;
							break;

						case 'W' :
							Flags &= ~CW_FLAG_WC_INCLUDES_BOM;
							break;

						case 'Z' :
							Recurse |= FH_FLAG_NO_SYSTEM_DIRS;
							break;

						default :
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Unknown suboption: /N:%c\r\n", towupper( arg[i][j] ) );
							goto ErrorExit;
					}

					j++;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				UniqueMode = 0;
			else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &UniqueMode, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad UniqueMode value: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}
				else if ( UniqueMode > 3 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"UniqueMode must be 0 - 3\r\n" );
					goto ErrorExit;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}


				while ( arg[i][j] ) {

					switch( towupper( arg[i][j] ) ) {

					case 'D' :
					case '$' :
						Flags |= CW_FLAG_ALLOW_DOLLAR_SIGNS;
						break;

					case 'U' :
					case '_' :
						Flags |= CW_FLAG_ALLOW_UNDERSCORES;
						break;

					default :
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unknown suboption: /W:%c\r\n", towupper( arg[i][j] ) );
						goto ErrorExit;
					}

					j++;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				Flags |= CW_FLAG_REJECT_INITIAL_DIGITS;
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( _wcsicmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}
		}
	}


	if ( !UniqueMode )
		Flags &= ~CW_FLAG_DUMP_WORDS_LIST;

	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	PassArgs.UniqueMode		= UniqueMode;
	PassArgs.UserFormat		= UserFormat;
	PassArgs.MinLetters		= MinLetters;


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );


	if ( !StdinSpecified ) {
		if ( ( ConsoleHandle( STD_INPUT_HANDLE ) == 0 ) || ( !FilenameSpecified ) ) {
			fhrv = CW_CountWords( (LPTSTR) StdinFilename, Flags, &PassArgs );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &PassArgs, CW_CountWords, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	//  Report grand totals if needed:

	if ( FilesProcessed > 1 )
		CW_ShowReport( NULL, GrandTotalBytes, GrandTotalChars, GrandTotalWords, ( UniqueMode & UNIQUEMODE_ALL_FILES ) ? UniqueWordsAll : ULONG_MAX, GrandTotalProperNouns, GrandTotalRuns, GrandTotalSentencesD, GrandTotalSentencesE, GrandTotalSentencesQ, GrandTotalSentenceWords, GrandTotalLines, GrandTotalNonEmptyLines, GrandTotalLongestLine, GrandTotalGrafs, GrandTotalTitles, FILE_NONE, 0, FALSE, ( Flags & CW_FLAG_CODE_MODE ) );

	if ( Flags & CW_FLAG_DUMP_WORDS_LIST )
		if ( UniqueMode & UNIQUEMODE_ALL_FILES ) {
			if ( HLprintf( L"\n\x12WORD LIST FOR ALL FILES:" ) == USER_ABORT )
				goto UserAbort;

			CW_DumpWordsByFreq( &FirstWordAll );
		}

	TotalCountsValid = ( 1024 | UniqueMode );

	CW_FreeList( &FirstWordAll );
	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit:

	CW_FreeList( &FirstWordAll );
	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	CW_FreeList( &FirstWordAll );
	FH_FreeLinkedList( &FirstRange );
	return 3;
}






DLLExports int WINAPI _characters( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountChars );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _charactersall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalChars );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _lines( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountLines );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _linesall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalLines );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _longestline( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", LongestLine );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _longestlineall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalLongestLine );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _nonblanklines( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", NonEmptyLines );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _nonblanklinesall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalNonEmptyLines );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _paragraphs( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountGrafs );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _paragraphsall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalGrafs );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _propernouns( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid & UNIQUEMODE_EACH_FILE )
		wsprintf( outString, L"%u", ProperNouns );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _propernounsall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid & UNIQUEMODE_ALL_FILES )
		wsprintf( outString, L"%u", GrandTotalProperNouns );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentences( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountSentencesD + CountSentencesE + CountSentencesQ );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentencesall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalSentencesD + GrandTotalSentencesE + GrandTotalSentencesQ );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentencesd( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountSentencesD );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentencesdall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalSentencesD );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentencese( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountSentencesE );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentenceseall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalSentencesE );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentencesq( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountSentencesQ );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentencesqall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalSentencesQ );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentencewords( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", TotalSentenceWords );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _sentencewordsall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalSentenceWords );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _titles( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountTitles );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _titlesall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalTitles );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _uniquewords( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid & UNIQUEMODE_EACH_FILE )
		wsprintf( outString, L"%u", UniqueWords );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _uniquewordsall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid & UNIQUEMODE_ALL_FILES )
		wsprintf( outString, L"%u", UniqueWordsAll );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _wc( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", CountRuns );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _wcall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalRuns );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _wordfiles( LPTSTR outString )
{
	if ( !outString )
		return 1;

	wsprintf( outString, L"%u", FilesProcessed );
	return 0;
}


DLLExports int WINAPI _words( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( WordCountsValid )
		wsprintf( outString, L"%u", TotalWords );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


DLLExports int WINAPI _wordsall( LPTSTR outString )
{
	if ( !outString )
		return 1;

	if ( TotalCountsValid )
		wsprintf( outString, L"%u", GrandTotalWords );
	else
		wcscpy_s( outString, MAX_ARG_LEN, StringNotAvailable );

	return 0;
}


#endif	//	WORDS_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				CW_CountWords() filename and CW_ShowReport() filename are defined as constants:
				fix for mmfiles 1.4.4.2+
				Moved CASE_ALL_CAPS, CASE_ALL_LOWER, and CASE_FIRST_FEW #defines in here.

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
