
//			WRAP command:


#ifndef WRAP_COMMAND_VER
#define WRAP_COMMAND_VER			1.0.1.1


#define WRAP_MAX_SPACES				10

#define WRAP_DEBUG					0x0040
#define WRAP_SHOW_HEADERS			0x0080
#define WRAP_IGNORE_EOLS			0x0100
#define WRAP_REMOVE_HYPHENS			0x0200
#define WRAP_DISABLE_SOFT_HYPHEN	0x0400
#define WRAP_NO_HYPHEN_AT_FRACTURE	0x0800
#define WRAP_JUSTIFY				0x1000
#define WRAP_FIX_QUOTES				0x2000
#define WRAP_UNDEFINED				0xffffffff



const wchar_t WrapHelpText[] = L"Word-wrap English text to fit a specified number of columns.\n\
\n\
WRAP /CP:n /D /F:n /G:n /H /N:n /P /R /S /T:n /W:width filename....\r\n\
\n\
\t/CP:  interpret OEM text as code page n\n\
\t/D    disable special handling of soft hyphens\n\
\t/F:   input text format:\n\
\t       0: best guess (default)\n\
\t       1: unformatted; honor all line breaks\n\
\t       2: prewrapped; skip single line breaks\n\
\t       3: unformatted; honor line breaks, no blank between grafs\n\
\t/G:   indent all paragraphs n spaces\n\
\t/H    display filenames\n\
\t/N:   minimum characters before and after to split at a hyphen (3)\n\
\t/P    page output\n\
\t/R    remove hyphens from line ends\n\
\t/S    search in subdirectories for matching files\n\
\t/T:   tab stops every n spaces (4)\n\
\t/W:   desired width\n\
\n\
Input filenames may be specified on the command line, or text may be\n\
piped or redirected into WRAP.  The default width is the current console\n\
width if output is not redirected, 100 columns or per environment variable \n\
COLUMNS if it is.\n";



struct WrapFileArgs {
	int Width;
	int TabSize;
	int HyphenMin;
	int HyphenMin2;
	int MaxSpaces;
	int Indent;
	int Indent2;
	int UserFormat;
	unsigned int CommentChar;
	unsigned int ForceEOL;
	unsigned int CodePage;
	unsigned int EbcdicCodePage;
};



int JustifyLine( LPTSTR inString, int Width )
{
	if ( !inString )
		return -1;
	else if ( inString[0] == '\0' )
		return 0;

	if ( (int) wcslen( inString ) >= ( Width - 1 ) )
		return 0;

	int SpaceRun[MAX_COLUMNS];
	short int Adj[MAX_COLUMNS];

	wchar_t Temp[MAX_ARG_LEN] = L"";

	BOOL Begin = FALSE;
	wchar_t ch = 0x00;
	int NumSpaceRuns = 0, NeedToAdd = 0;

	for ( int i = 0; i < MAX_COLUMNS; i++ )
		SpaceRun[i] = Adj[i] = 0;

	for ( int i = 0; inString[i] != 0x00; i++ ) {
		ch = inString[i];
		if ( Begin && iswspace( ch ) )
			SpaceRun[NumSpaceRuns++] = i;

		Begin = ( iswspace( ch ) == 0 );
	}

	if ( NumSpaceRuns == 0 )
		return 0;

	NeedToAdd = Width - 1 - (int) wcslen( inString );

	if ( NeedToAdd >= NumSpaceRuns ) {
		for ( int i = 0; i < NumSpaceRuns; i++ )
			Adj[i] += ( NeedToAdd / NumSpaceRuns );

		NeedToAdd %= NumSpaceRuns;
	}

	for ( int i = 1; i <= NeedToAdd; i++ )
		Adj[ i *  NumSpaceRuns / ( NeedToAdd + 1 ) ]++;


	int j = 0, sr = 0;

	for ( int i = 0; inString[i] != 0x00; i++ ) {

		if ( i == SpaceRun[sr] ) {
			for ( int k = 0; k < Adj[sr]; k++ )
				Temp[j++] = SPACE;

			sr++;
		}

		Temp[j++] = inString[i];
		Temp[j] = 0x00;
	}

	wcscpy_s( inString, MAX_ARG_LEN, Temp );
	return 0;
}



int WrapDump( LPTSTR inString, int Width, BOOL Justify )
{
	if ( !inString )
		return -1;

	wchar_t Temp[MAX_ARG_LEN + 1] = L"";
	wcscpy_s( Temp, MAX_ARG_LEN + 1, inString );

	TrimRight( Temp );

	if ( Justify )
		JustifyLine( Temp, Width );

	QPuts( Temp );
	return CL_EndLine( 1 );
}



int InsertSpaces( LPTSTR inString, int n )
{
	if ( ( inString == NULL ) || ( n < 1 ) )
		return 0;

	int i = 0;

	for ( i = (int) wcslen( inString ); i >= 0; i-- )
		inString[i + n] = inString[i];

	for ( i = 0; i < n; i++ )
		inString[i] = SPACE;

	return 0;
}



int WrapRemoveChars( LPTSTR inString, int n )
{
	if ( ( inString == NULL ) || ( n < 1 ) )
		return 0;

	if ( n > (int) wcslen( inString ) )
		return 0;

	int i = n;
	wchar_t ch = SPACE;

	while ( inString[i] == SPACE )
		i++;

	for ( int j = 0; ch != 0x00; j++ ) {
		ch = inString[i + j];
		inString[j] = ch;
	}

	return i;
}



int WrapFile( LPCTSTR inFilename, int Flags, void *WrapArgs )
{
	if ( !inFilename )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;


	wchar_t OutBuffer[MAX_ARG_LEN]	= L"";
	wchar_t AposWord[SHORT_BUF_LEN]	= L"";

	int OutCount					= 0;

	BOOL Breaks						= FALSE;
	BOOL Preformatted				= FALSE;
	BOOL Headers					= Flags & WRAP_SHOW_HEADERS;
	BOOL RemoveHyphens				= Flags & WRAP_REMOVE_HYPHENS;
	BOOL Justify					= Flags & WRAP_JUSTIFY;
	BOOL IgnoreEOLs					= Flags & WRAP_IGNORE_EOLS;
	BOOL Debug						= Flags & WRAP_DEBUG;

	_MMFile MMFile;

	WrapFileArgs *MoreArgs	= (WrapFileArgs*) WrapArgs;
	int Width					= MoreArgs->Width;
	int TabSize					= MoreArgs->TabSize;
	int HyphenMin				= MoreArgs->HyphenMin;
	int HyphenMin2				= MoreArgs->HyphenMin2;
	int MaxSpaces				= MoreArgs->MaxSpaces;
	int Indent					= MoreArgs->Indent;
	int Indent2					= MoreArgs->Indent2;
	int UserFormat				= MoreArgs->UserFormat;
	unsigned int CommentChar	= MoreArgs->CommentChar;
	unsigned int ForceEOL		= MoreArgs->ForceEOL;

	int ch						= 0;
	int lastch					= 0;


	if ( !HyphenMin2 )
		HyphenMin2 = HyphenMin;


	BOOL FileIsPrewrapped = FALSE;
	size_t BOMSize = 0;

	int sv = 0x00, sv2 = 0x00;
	int LastSpace = -1, LastHyphen = -1;
	wchar_t LastHyphenChar = 0x00;
	int crcount = 0, lfcount = 0;

	BOOL EatSpaces = FALSE, NewGraf = TRUE, DeleteToEOL = FALSE;

	int SpacesCount = 0;



	//  Open file:

	int rv = OpenMMFile( &MMFile, inFilename, MM_FLAG_ERROR_MESSAGES | MM_FLAG_ALLOW_EBCDIC, MoreArgs->CodePage, MoreArgs->EbcdicCodePage );

	if ( rv == ERROR_EMPTY_FILE ) {
		if ( HLprintf( L"\nFile \"%s\" is empty.\n", inFilename ) == USER_ABORT )
			return USER_ABORT;
		else
			return 0;
	}
	else if ( rv )
		return -1;


	CL_HookBreak();

	if ( UserFormat )
		MMFile.FileFormat = UserFormat;
	else
		GuessFileFormat( &MMFile );


	Breaks			= ( MMFile.FileFormat & 1 ) > 0;
	Preformatted	= ( MMFile.FileFormat & 2 ) > 0;

	if ( Debug ) {
		Printf( L"   * FileEncoding = %d   FileFormat = %d\r\n", MMFile.FileEncoding, MMFile.FileFormat );
		Printf( L"   * Breaks = %d  Preformatted = %d\r\n", Breaks, Preformatted );
	}


	//  Display header if needed:

	if ( Headers )
		if ( ShowFilenameHeader( &MMFile ) == USER_ABORT )
			goto UserAbort;


	//  Loop through contents of the file:

	ch = SPACE;

	while ( MMFile.Index <= MMFile.FileSize ) {
		lastch = ch;

		ch = MMGetChar( &MMFile );

		if ( ch < 0 )					//  abort on error or eof
			break;
		else if ( ch == '\0' )			//  ignore any nulls in input text
			continue;


		if ( IgnoreEOLs && _MMIsEOL( ch ) ) {	//  if we're ignoring eols (undoc feature) ...
			if ( MMPeekAhead( &MMFile, 0 ) == _MM2PartEOL( ch ) )
				MMGetChar( &MMFile );

			if ( SpacesCount )					//  if the preceding character was a space,
				continue;						//  then discard this one
			else
				ch = SPACE;						//  otherwise, covert the eol to a space
		}

		if ( ( ch == SPACE ) && ( MaxSpaces > 0 ) && ( SpacesCount >= MaxSpaces ) )
			continue;							//  condense spaces ( /C:n )

		if ( NewGraf && ( ( Indent >= 0 ) && ( Indent <= MAX_INDENT ) ) && ( ( ch == SPACE ) || ( ch == TAB ) ) )
			continue;							//  if /P: specified, ignore spaces and tabs at start of new paragraph

		if ( EatSpaces && ( ch == SPACE ) )
			continue;							//  if we're eating spaces and this is a space, discard it

		DeleteToEOL |= ( ch == CommentChar );	//  if CommentChar found, set comment flag

		if ( DeleteToEOL && !_MMIsEOL( ch ) && ( ch != ForceEOL ) )
			continue;							//  if CommentChar found, delete everything to eol

		if ( ch != SPACE )						//  if this is not a space,
			EatSpaces = FALSE;					//  we're done eating spaces

		
		if ( ( Flags & WRAP_FIX_QUOTES ) && ( lastch != '\\' ) ) {
			wchar_t nextch = MMPeekAhead( &MMFile, 0 );

			if ( ReplaceDoubles && ( ch == APOSTROPHE ) && ( nextch == APOSTROPHE ) ) {
				if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( MMPeekAhead( &MMFile, 1 ) ) ) {
					ch = QUOTE;
					MMGetChar( &MMFile );
					nextch = MMPeekAhead( &MMFile, 0 );
				}
			}

			if ( ReplaceDoubles && ReplaceGraves && ( ch == GRAVE ) && ( nextch == GRAVE ) ) {
				if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( MMPeekAhead( &MMFile, 1 ) ) ) {
					ch = LDQUOTE;
					MMGetChar( &MMFile );
					nextch = MMPeekAhead( &MMFile, 0 );
				}
			}

			if ( ReplaceGraves && ( ch == GRAVE ) )
				if ( !AnyKindOfQuote( lastch ) && !AnyKindOfQuote( nextch ) )
					ch = LSQUOTE;

			if ( ch == QUOTE ) {
				if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswspace( nextch ) || _MMIsEOL( nextch ) ) )
					ch = DoublePrime;
				else if ( iswalpha( lastch ) && !iswalpha( nextch ) )
					ch = CloseQuote;
				else if ( iswalpha( nextch ) && !iswalpha( lastch ) )
					ch = OpenQuote;
				else if ( ( iswgraph( lastch ) ) || ( _MMIsEOL( nextch ) ) || ( nextch == ForceEOL ) )
					ch = CloseQuote;
				else
					ch = OpenQuote;
			}
			else if ( ch == APOSTROPHE ) {
				if ( iswgraph( lastch ) && ( lastch != OpenQuote ) ) {
					ch = CloseSQuote;

					if ( iswalnum( lastch ) && ( iswalpha( nextch ) ) )
						ch = Apostrophe;
				}
				else {
					ch = OpenSQuote;

					if ( _MMIsEOL( nextch ) || ( nextch == ForceEOL ) )
						ch = CloseSQuote;
				}
				
				if ( ( iswalpha( lastch ) == 0 ) && ( iswalpha( nextch ) ) ) {
					AposWord[0] = APOSTROPHE;
					int peek = 0;
					for ( int j = 0; j < 15; j++ ) {
						peek = MMPeekAhead( &MMFile, j );
						peek = ( peek < SPACE ) ? SPACE : peek;
						AposWord[j + 1] = peek;
					}

					AposWord[16] = 0x00;
					if ( IsWordBeginningWithApostrophe( AposWord ) )
						ch = Apostrophe;
				}
				else if ( IsVowel2( lastch ) && IsVowel2( nextch ) )
					ch = Okina;
				
				if ( ( iswalnum( lastch ) == 0 ) && ( iswdigit( nextch ) ) )
					if ( iswdigit( MMPeekAhead( &MMFile, 1 ) ) )
						if ( iswalnum( MMPeekAhead( &MMFile, 2 ) ) == 0 )
							ch = Apostrophe;

				if ( ( iswdigit( lastch ) || IsFraction( lastch ) ) && ( iswdigit( nextch ) || IsFraction( nextch ) || iswspace( nextch ) || _MMIsEOL( nextch ) ) )
					ch = Prime;
			}
			else if ( ( ch == '-' ) && ( nextch == '-' ) && ( lastch != '-' ) ) {
				if ( MMPeekAhead( &MMFile, 1 ) != '-' ) {
					ch = EmDash;
					MMGetChar( &MMFile );
				}
			}
		}


		//  If the character is a soft hyphen, deal with it:

		if ( ( ch == SHYPHEN ) && ( OutCount > 0 ) && ( OutCount < ( Width - 1 ) ) && ( HyphenMin > 0 ) && ( ( Flags & WRAP_DISABLE_SOFT_HYPHEN ) == 0 ) ) {
			if ( iswalpha( OutBuffer[OutCount - 1] ) && iswalnum( MMPeekAhead( &MMFile, 0 ) ) ) {
				BOOL Ok = TRUE;
				int LookAhead = 0;

				for ( int j = 0; j < HyphenMin2; j++ ) {	// count chars to move to next line
					LookAhead = MMPeekAhead( &MMFile, j );
					if ( ( LookAhead <= 0 ) || iswspace( LookAhead ) || _MMIsEOL( LookAhead ) || ( LookAhead == ForceEOL ) || ( LookAhead == CommentChar ) )
						Ok = FALSE;
				}
				
				if ( Ok ) {						// if enough characters would move to next line,
					LastHyphen = OutCount;		// note the soft hyphen's location
					LastHyphenChar = ch;		// and value
				}
			}
		}

		if ( ( ch == SHYPHEN ) && ( ( Flags & WRAP_DISABLE_SOFT_HYPHEN ) == 0 ) )
			continue;							// discard soft hyphens


		if ( !_MMIsEOL( ch ) && ( ch != SPACE ) && ( ch != TAB ) ) {
			if ( NewGraf ) {					// start of new paragraph and /P: specified:
				for ( int i = 0; i < Indent; i++ ) {
					if ( OutCount < ( MAX_ARG_LEN - 1 ) ) {
						OutBuffer[OutCount++] = SPACE;		// indent it
						OutBuffer[OutCount] = '\0';
					}
				}
			}

			NewGraf = FALSE;
		}


		//  Expand tabs:
		if ( ( ch == TAB ) && ( TabSize > 0 ) ) {
			int k = TabSize - ( OutCount % TabSize );
			for ( int i = 0; i < k; i++ ) {
				if ( OutCount < ( MAX_ARG_LEN - 1 ) ) {
					LastSpace = OutCount;
					OutBuffer[OutCount++] = SPACE;
					OutBuffer[OutCount] = '\0';
				}
			}
			SpacesCount = 0;
			continue;
		}


		//  Handle line terminators:
		if ( _MMIsEOL( ch ) ) {
			if ( MMPeekAhead( &MMFile, 0 ) == _MM2PartEOL( ch ) )
				MMGetChar( &MMFile );

			if ( !Breaks && !_MMIsEOL( MMPeekAhead( &MMFile, 0 ) ) && ( OutCount > 0 ) ) {
				ch = SPACE;
				EatSpaces = TRUE;

				if ( OutCount )
					if ( OutBuffer[OutCount - 1] == SPACE )
						continue;

				if ( OutCount > 1 )
					if ( iswalpha( OutBuffer[OutCount - 2] ) && IsHyphen( OutBuffer[OutCount - 1] ) ) {
						int i = 0;
						while ( ( i < MAX_INDENT ) && ( ( MMPeekAhead( &MMFile, i ) == SPACE ) || ( MMPeekAhead( &MMFile, i ) == TAB ) ) )
							i++;

						if ( iswalpha( MMPeekAhead( &MMFile, i ) ) ) {
							if ( RemoveHyphens ) {
								BOOL Vowels = FALSE, Nonletters = FALSE, Done = FALSE;
								int j = OutCount - 2;
								wchar_t c = '\0';

								do {
									c = OutBuffer[j--];

									if ( iswspace( c ) || IsHyphen( c ) || ( c == QUOTE ) || ( c == LDQUOTE ) || ( c == LSQUOTE ) || ( c == '(' ) )
										Done = TRUE;
									else if ( IsVowel( c ) )
										Vowels = TRUE;
									else if ( ( c == APOSTROPHE ) || ( c == RSQUOTE ) )
										;  // apostrophe may be used within a word
									else if ( iswalpha( c ) == 0 )
										Nonletters = TRUE;

									if ( j < 0 )
										Done = TRUE;

								} while ( !Done );

								if ( Vowels && !Nonletters ) {
									OutCount--;
									OutBuffer[OutCount] = '\0';
								}
							}
						}
						continue;
					}
			}
		}


		// Count spaces for use of /C:

		if ( ch == SPACE )
			SpacesCount++;
		else
			SpacesCount = 0;


		if ( _MMIsEOL( ch ) || ( ch == ForceEOL ) ) {		// end of paragraph:
			if ( WrapDump( OutBuffer, Width, FALSE ) == USER_ABORT )
				goto UserAbort;								// display contents of buffer

			if ( Breaks && !Preformatted && ( OutCount > 0 ) && ( ch != ForceEOL ) )
				if ( CL_EndLine() == USER_ABORT )			// add a blank line
					goto UserAbort;

			OutCount = 0;									// empty the output buffer
			OutBuffer[OutCount] = '\0';
			NewGraf = TRUE;									// start a new paragraph
			DeleteToEOL = FALSE;							// no longer ignoring any comment
			continue;
		}

		if ( ( ch != NBSP ) && iswspace( ch ) ) {		// space in input text:
			if ( !StuffNotToBreakBefore( MMPeekAhead( &MMFile, 0 ), MMPeekAhead( &MMFile, 1 ) ) ) {
				LastSpace = OutCount;					// note its location
				LastHyphen = -1;						// and discard any pending hyphen break
				LastHyphenChar = '\0';
			}
		}

		//  If the character is a hyphen, deal with it:

		if ( ( OutCount > 0 ) && ( OutCount < ( Width - 1 ) ) && ( IsHyphen( ch ) ) && ( HyphenMin > 0 ) )
			if ( iswalpha( OutBuffer[OutCount - 1] ) && iswalnum( MMPeekAhead( &MMFile, 0 ) ) ) {
				BOOL Ok = TRUE;
				int LookAhead = 0;

				for ( int j = 0; j < HyphenMin2; j++ ) {	// count chars to move to next line
					LookAhead = MMPeekAhead( &MMFile, j );
					if ( ( LookAhead <= 0 ) || iswspace( LookAhead ) || _MMIsEOL( LookAhead ) || ( LookAhead == ForceEOL ) || ( LookAhead == CommentChar ) )
						Ok = FALSE;
				}
				
				if ( Ok ) {						// if enough characters would move to next line,
					LastHyphen = OutCount;		// note the hyphen's location
					LastHyphenChar = ch;		// and value
				}
			}

		//  Add character to the output buffer:

		if ( OutCount < ( MAX_ARG_LEN - 1 ) ) {
			OutBuffer[OutCount++] = ch;
			OutBuffer[OutCount] = '\0';
		}

		if ( OutCount >= Width ) {		// if the output buffer is full ....

			if ( ( LastHyphen >= ( Width * 2 / 5 ) ) && ( LastHyphen > ( LastSpace + HyphenMin ) ) ) {
				sv = OutBuffer[LastHyphen + 1];			// break at last hyphen
				sv2 = OutBuffer[LastHyphen];

				OutBuffer[LastHyphen + 1] = '\0';
				if ( LastHyphenChar == SHYPHEN )
					OutBuffer[LastHyphen] = HYPHEN;

				if ( WrapDump( OutBuffer, Width, Justify ) == USER_ABORT )
					goto UserAbort;

				OutBuffer[LastHyphen + 1] = sv;
				OutBuffer[LastHyphen] = sv2;

				if ( LastHyphenChar == SHYPHEN )
					OutCount -= WrapRemoveChars( OutBuffer, LastHyphen );
				else
					OutCount -= WrapRemoveChars( OutBuffer, LastHyphen + 1 );
			}
			else if ( LastSpace >= ( Width * 2 / 5 ) ) {
				sv = OutBuffer[LastSpace];				// break at last space
				OutBuffer[LastSpace] = '\0';

				if ( WrapDump( OutBuffer, Width, Justify ) == USER_ABORT )
					goto UserAbort;

				OutBuffer[LastSpace] = sv;
				OutCount -= WrapRemoveChars( OutBuffer, LastSpace + 1 );

				if ( OutCount == 0 )
					EatSpaces = TRUE;
			}
			else if ( Flags & WRAP_NO_HYPHEN_AT_FRACTURE ) {
				sv = OutBuffer[Width - 1];				// break in mid-word without apology
				OutBuffer[Width - 1] = '\0';

				if ( WrapDump( OutBuffer, Width, Justify ) == USER_ABORT )
					goto UserAbort;

				OutBuffer[Width - 1] = sv;
				OutCount -= WrapRemoveChars( OutBuffer, Width - 1 );
			}
			else {
				sv = OutBuffer[Width - 2];				// break in mid-word without apology
				sv2 = OutBuffer[Width - 1];
				OutBuffer[Width - 2] = HYPHEN;
				OutBuffer[Width - 1] = '\0';

				if ( WrapDump( OutBuffer, Width, Justify ) == USER_ABORT )
					goto UserAbort;

				OutBuffer[Width - 2] = sv;
				OutBuffer[Width - 1] = sv2;
				OutCount -= WrapRemoveChars( OutBuffer, Width - 2 );
			}

			LastSpace = LastHyphen = -1;
			LastHyphenChar = '\0';

			if ( ( Indent != UINT_MAX ) && ( Indent2 != 0 ) ) {		// if second-line indent
				InsertSpaces( OutBuffer, Indent2 );					// value was specified,
				OutCount += Indent2;								// indent the next line
			}
		}
	}

	//  Dump anything remaining in the output buffer:

	if ( OutCount )
		if ( WrapDump( OutBuffer, Width, FALSE ) == USER_ABORT )
			goto UserAbort;


	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return 0;

UserAbort:

	CloseMMFile( &MMFile );
	CL_ReleaseBreak();
	return 0;
}



DLLExports int WINAPI wrap( LPTSTR inString )
{
	if ( !inString )
		return 1;


	int Flags					= 0;
	int Recurse					= FH_FLAG_ALLOW_FILE_LISTS | FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_URLS | FH_FLAG_ALLOW_CLIPBOARD | FH_FLAG_MINUS_IS_SPECIAL;
	unsigned int MinDepth		= 0;
	unsigned int MaxDepth		= UINT_MAX;
	DWORD AttribMask			= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues			= 0;
	void *FirstRange			= NULL;
	BOOL NoHighlight			= FALSE;
	unsigned int Pause			= 0;

	unsigned int UserFormat		= 0;
	unsigned int Width			= 0;
	unsigned int MaxSpaces		= 0;
	unsigned int TabSize		= UINT_MAX;
	unsigned int Indent			= UINT_MAX;
	unsigned int Indent2		= 0;
	unsigned int HyphenMin		= 3;
	unsigned int HyphenMin2		= 0;
	unsigned int ForceEOL		= WRAP_UNDEFINED;
	unsigned int CommentChar	= WRAP_UNDEFINED;

	BOOL FilenameSpecified		= FALSE;
	BOOL StdinSpecified			= FALSE;
	BOOL WriteBOM				= TRUE;
	int rv						= 0;
	int fhrv					= 0;

	WrapFileArgs PassArgs;

	const wchar_t MsgSyntaxError[]	= L"Syntax error: \"%s\"\r\n";


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	CL_NoPausing();


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				FH_FreeLinkedList( &FirstRange );
				return ShowCmdHelp( WrapHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( WRAP_COMMAND_VER ) );
			}
			else if ( !wcscmp( arg[i], L"/$" ) )
				Flags |= WRAP_DEBUG;
			else if ( !_wcsnicmp( arg[i], L"/CP", 3 ) ) {
				if ( ParseCodePageOption( arg[i], &(PassArgs.CodePage), &(PassArgs.EbcdicCodePage) ) )
					goto ErrorExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &MaxSpaces, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
					goto ErrorExit;
				}
				else if ( MaxSpaces > WRAP_MAX_SPACES ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Count must be 0 - %u\r\n", WRAP_MAX_SPACES );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				Flags |= WRAP_DISABLE_SOFT_HYPHEN;
			else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &UserFormat, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
					goto ErrorExit;
				}
				else if ( UserFormat > FMT_MAX_VALUE ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Format must be 0 - %u\r\n", FMT_MAX_VALUE );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/G" ) ) {
				Indent = ULONG_MAX;
				Indent2 = 0;
			}
			else if ( !_wcsnicmp( arg[i], L"/G", 2 ) ) {
				int j = 2;
				wchar_t *sep1 = NULL;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Indent, &sep1 ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
					goto ErrorExit;
				}
				else if ( *sep1 && ( *sep1 != ',' ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
					goto ErrorExit;
				}
				else if ( Indent > MAX_INDENT ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Indent must be 0 - %u\r\n", MAX_INDENT );
					goto ErrorExit;
				}

				Indent2 = 0;

				if ( *sep1 == ',' ) {
					if ( ParseInt( sep1 + 1, &Indent2, NULL ) ) {
						DisplayErrorHeader();
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
						goto ErrorExit;
					}
					else if ( Indent2 > MAX_INDENT ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Indent2 must be 0 - %u\r\n", MAX_INDENT );
						goto ErrorExit;
					}
				}
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) )
				Flags |= WRAP_SHOW_HEADERS;
			else if ( !_wcsicmp( arg[i], L"/J" ) )
				Flags |= WRAP_JUSTIFY;
			else if ( !_wcsicmp( arg[i], L"/K" ) )
				Flags &= ~WRAP_REMOVE_HYPHENS;
			else if ( !_wcsicmp( arg[i], L"/N" ) ) {
				HyphenMin = 3;
				HyphenMin2 = 0;
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) && ( iswalpha( arg[i][2] ) || ( IsColonOrEquals( arg[i][2] ) && iswalpha( arg[i][3] ) ) ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == 0x00 ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					goto ErrorExit;
				}

				while ( arg[i][j] ) {
					switch ( towupper( arg[i][j] ) ) {
						case 'B' :
							WriteBOM = FALSE;
							break;

						case 'C' :
							NoHighlight = TRUE;
							break;

						case 'D' :
							Recurse |= FH_FLAG_NO_HIDDEN_DIRS;
							break;

						case 'H' :
							Flags |= WRAP_NO_HYPHEN_AT_FRACTURE;
							break;

						case 'J' :
							Recurse |= FH_FLAG_NO_JUNCTIONS;
							break;

						case 'Z' :
							Recurse |= FH_FLAG_NO_SYSTEM_DIRS;
							break;

						default :
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Unknown suboption: /N:%c\r\n", towupper( arg[i][j] ) );
							goto ErrorExit;
					}

					j++;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				int j = 2;
				wchar_t *sep1 = NULL;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &HyphenMin, &sep1 ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
					goto ErrorExit;
				}
				else if ( ( *sep1 ) && ( *sep1 != ',' ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
					goto ErrorExit;
				}
				else if ( HyphenMin > MAX_COLUMNS ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Hyphen-break minimum must be 0 - %d\r\n", MAX_COLUMNS );
					goto ErrorExit;
				}

				HyphenMin2 = 0;

				if ( *sep1 == ',' ) {
					if ( ParseInt( sep1 + 1, &HyphenMin2, NULL ) ) {
						DisplayErrorHeader();
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
						goto ErrorExit;
					}
					else if ( HyphenMin2 > MAX_COLUMNS ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Hyphen-break minimum must be 0 - %d\r\n", MAX_COLUMNS );
						goto ErrorExit;
					}
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &Pause ) )
					goto ErrorExit;
			}
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Flags |= WRAP_FIX_QUOTES;
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Flags |= WRAP_REMOVE_HYPHENS;
			else if ( !_wcsnicmp( arg[i], L"/T", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &TabSize, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
					goto ErrorExit;
				}
				else if ( TabSize > MAX_TAB_SIZE ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Tab size must be 0 - %d\r\n", MAX_TAB_SIZE );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/W" ) )
				Width = ConsoleWidth();
			else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseInt( arg[i] + j, &Width, NULL ) ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
					goto ErrorExit;
				}
				else if ( ( Width < MIN_COLUMNS ) || ( Width > MAX_COLUMNS ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Width must be %d - %d\r\n", MIN_COLUMNS, MAX_COLUMNS );
					goto ErrorExit;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				Flags |= WRAP_IGNORE_EOLS;
			else if ( !_wcsicmp( arg[i], L"/Y" ) )
				CommentChar = WRAP_UNDEFINED;
			else if ( !_wcsnicmp( arg[i], L"/Y", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] && !iswdigit( arg[i][j] ) && ( arg[i][j + 1] == '\0' ) )
					CommentChar = ( arg[i][j] );
				else {
					if ( ParseInt( arg[i] + j, &CommentChar, NULL ) ) {
						DisplayErrorHeader();
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
						goto ErrorExit;
					}
					else if ( CommentChar > 0xffff ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Char must be 0 - 65535\r\n" );
						goto ErrorExit;
					}
				}
			}
			else if ( !_wcsicmp( arg[i], L"/Z" ) )
				ForceEOL = WRAP_UNDEFINED;
			else if ( !_wcsnicmp( arg[i], L"/Z", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] && !iswdigit( arg[i][j] ) && ( arg[i][j + 1] == '\0' ) )
					ForceEOL = ( arg[i][j] );
				else {
					if ( ParseInt( arg[i] + j, &ForceEOL, NULL ) ) {
						DisplayErrorHeader();
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						Qprintf( ERRHANDLE, MsgSyntaxError, arg[i] );
						goto ErrorExit;
					}
					else if ( ForceEOL > 0xffff ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Char must be 0 - 65535\r\n" );
						goto ErrorExit;
					}
				}
			}
			else if ( fhrv = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( fhrv < 0 )
					goto ErrorExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto ErrorExit;
			}
			else {
				if ( _wcsicmp( arg[i], StdinFilename ) )
					FilenameSpecified = TRUE;
				else
					StdinSpecified = TRUE;
			}

		}
	}


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );


	if ( !Width ) {
		if ( ConsoleHandle( STD_OUTPUT_HANDLE ) )
			Width = ConsoleWidth();
		else
			Width = EnvironmentNumericValue( L"COLUMNS", MIN_COLUMNS, MAX_COLUMNS );

		if ( ( Width < MIN_COLUMNS ) || ( Width > MAX_COLUMNS ) )
			Width = DEFAULT_COLUMNS;
	}

	if ( TabSize > MAX_TAB_SIZE )
		TabSize = EnvironmentNumericValue( L"TABSIZE", 0, MAX_TAB_SIZE );

	if ( TabSize > MAX_TAB_SIZE )
		TabSize = 4;

	if ( _MMIsEOL( ForceEOL ) )
		ForceEOL = 0;

	if ( CommentChar != WRAP_UNDEFINED )
		if ( _MMIsEOL( CommentChar ) || ( CommentChar == ForceEOL ) )
			CommentChar = WRAP_UNDEFINED;

	if ( HyphenMin >= ( Width - 1 ) )
		HyphenMin = 0;

	if ( HyphenMin2 == 0 )
		HyphenMin2 = HyphenMin;


	if ( Flags & WRAP_FIX_QUOTES )
		InitializeFixQuotesVars();


	PassArgs.Width			= Width;
	PassArgs.TabSize		= TabSize;
	PassArgs.HyphenMin		= HyphenMin;
	PassArgs.HyphenMin2		= HyphenMin2;
	PassArgs.MaxSpaces		= MaxSpaces;
	PassArgs.Indent			= Indent;
	PassArgs.Indent2		= Indent2;
	PassArgs.UserFormat		= UserFormat;
	PassArgs.CommentChar	= CommentChar;
	PassArgs.ForceEOL		= ForceEOL;


	if ( WriteBOM && !ConsoleHandle( STD_OUTPUT_HANDLE ) && OutputIsUnicode() )
		QPutc( GetStdHandle( STD_OUTPUT_HANDLE ), BOM );



	if ( !StdinSpecified ) {
		if ( ( ConsoleHandle( STD_INPUT_HANDLE ) == 0 ) || !FilenameSpecified ) {
			fhrv = WrapFile( (LPTSTR) StdinFilename, Flags, &PassArgs );
			if ( fhrv == USER_ABORT )
				goto UserAbort;
			else if ( fhrv )
				rv = 2;
		}
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				fhrv = HandleMatchingFiles( arg[i], Flags, &PassArgs, WrapFile, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );
				if ( fhrv == USER_ABORT )
					goto UserAbort;
				else if ( fhrv )
					rv = 2;
			}
		}
	}


	FH_FreeLinkedList( &FirstRange );
	return rv;

ErrorExit :

	FH_FreeLinkedList( &FirstRange );
	return 1;

UserAbort:

	FH_FreeLinkedList( &FirstRange );
	return 3;
}


#endif	//	WRAP_COMMAND_VER


/*

	-			2021-10-15

	1.0.0.0		2023-07-27
				Adds a version number to this source file.  Help text tweaks.

	1.0.1.0		2023-09-28
				WrapFile() filename is defined as a constant:  fix for mmfiles 1.4.4.2+

	1.0.1.1		2024-01-04
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
