
//
//		Metaphone.cpp     2012-05-24     C. Dye
//

#define METAPHONE_MAX		256
#define METAPHONE_MIN		4
#define MP_SHORT_BUF_LEN	1024


int Metaphone( LPTSTR inWord, LPTSTR outCode, int Length, BOOL Compatible )
{
	if ( ( inWord == NULL ) || ( outCode == NULL ) )
		return -1;
	else if ( inWord[0] == 0x00 )
		return -1;

	if ( ( Length < METAPHONE_MIN ) || ( Length > METAPHONE_MAX ) )
		Length = METAPHONE_MIN;

	wchar_t Upper[MP_SHORT_BUF_LEN] = L"";
	wchar_t Temp[METAPHONE_MAX + 1] = L"";
	wchar_t ch = 0x00, ch2 = 0x00, prev = 0x00;
	int i = 0, j = 0;

	const wchar_t Unaccented[65] = L"AAAAAA\u00c6CEEEEIIII\u00d0NOOOOO\u00d7\u00d8UUUUY\u00de\u00dfaaaaaa\u00e6ceeeeiiii\u00f0nooooo\u00f7\u00f8uuuuy\u00fey";

	for ( i = 0; inWord[i]; i++ ) {
		ch = inWord[i];
		if ( ( ch >= 0x00c0 ) && ( ch <= 0x00ff ) )
			ch = Unaccented[ch - 0x00c0];

		if ( iswalpha( ch ) && ( j < ( MP_SHORT_BUF_LEN - 1 ) ) ) {
			Upper[j++] = towupper( ch );
			Upper[j] = 0x00;
		}
	}

	if ( Upper[0] == 0x00 )
		return -1;

	//  transformations for the first letter(s):

	i = 0;

	if ( ( ( Upper[0] == 'G' ) || ( Upper[0] == 'K' ) || ( Upper[0] == 'P' ) ) && ( Upper[1] == 'N' ) )
		i++;
	else if ( ( Upper[0] == 'A' ) && ( Upper[1] == 'E' ) )
		i++;
	else if ( ( Upper[0] == 'W' ) && ( Upper[1] == 'R' ) )
		i++;
	else if ( ( Upper[0] == 'W' ) && ( Upper[1] == 'H' ) ) {
		Upper[1] = 'W';
		i++;
		ch2 = 'X';
	}
	else if ( Upper[0] == 'X' )
		Upper[0] = 'S';

	if ( i )
		Upper[0] = 0x00;

	//  transformations for the remainder of the word:

	ch = 0x00;

	for ( j = 0; Upper[i]; i++ ) {
		prev = ch2;

		ch = ch2 = Upper[i];

		if ( ( ch == prev ) && ( ch != 'G' ) && ( ch != 'C' ) )
			continue;

		switch ( ch ) {
			case 'B' :
				if ( ( prev == 'M' ) && ( Compatible || ( Upper[i + 1] ) == 0 ) )
					ch = 0x00;

				break;

			case 'C' :
				if ( Upper[i + 1] == 'I' )
					if ( Upper[i + 2] == 'A' )
						ch = 'X';

				if ( Upper[i + 1] == 'H' ) {
					ch = 'X';
					i++;

					if ( !Compatible && ( Upper[i + 1] == 'R' ) )
						ch = 'K';
				}

				if ( ( ch == 'C' ) && ( ( Upper[i + 1] == 'I' ) || ( Upper[i + 1] == 'E' ) || ( Upper[i + 1] == 'Y' ) ) ) {
					ch = 'S';
					if ( prev == 'S' )
						ch = 0x00;
				}

				if ( ( ch == 'C' ) && ( Upper[i + 1] == 'K' ) )
					i++;

				if ( ch == 'C' )
					ch = 'K';

				break;

			case 'D' :
				if ( Upper[i + 1] == 'G' )
					if ( ( Upper[i + 2] == 'E' ) || ( Upper[i + 2] == 'I' ) || ( Upper[i + 2] == 'Y' ) ) {
						ch = 'J';
						i++;
					}

				if ( ch == 'D' )
					ch = 'T';

				break;

			case 'G' :
				if ( Upper[i + 1] == 'H' ) {
					ch = 0x00;

					if ( Compatible || ( Upper[i + 2] == 0x00 ) || ( Upper[i + 2] == 'A' ) || ( Upper[i + 2] == 'E' ) || ( Upper[i + 2] == 'I' ) || ( Upper[i + 2] == 'O' ) || ( Upper[i + 2] == 'U' ) ) {
						ch = 'F';

						if ( i >= 3 )
							if ( ( Upper[i - 3] == 'B' ) || ( Upper[i - 3] == 'D' ) || ( Upper[i - 3] == 'H' ) )
								ch = 0x00;

						if ( i >= 4 )
							if ( Upper[i - 4] == 'H' )
								ch = 0x00;
					}
					else if ( Upper[i + 2] == 'T' )
						ch2 = '*';

					i++;
				}
				else if ( Upper[i + 1] == 'N' ) {
					if ( Upper[i + 2] == 0x00 )
						ch = 0x00;

					if ( Upper[i + 2] == 'E' )
						if ( Upper[i + 3] == 'D' )
							if ( Upper[i + 4] == 0x00 )
								ch = 0x00;
				}
				else if ( ( Upper[i + 1] == 'E' ) || ( Upper[i + 1] == 'I' ) || ( Upper[i + 1] == 'Y' ) ) {
					ch = 'J';

					if ( prev == 'G' )
						ch = 0x00;
				}

				if ( ch == 'G' ) {
					ch = 'K';
					if ( prev == 'G' )
						ch = 0x00;
				}

				break;

			case 'H' :
				if ( Compatible ) {
					if ( ( Upper[i + 1] != 'A' ) && ( Upper[i + 1] != 'E' ) && ( Upper[i + 1] != 'I' ) && ( Upper[i + 1] != 'O' ) && ( Upper[i + 1] != 'U' ) )
						ch = 0x00;
					else if ( ( prev == 'C' ) || ( prev == 'G' ) || ( prev == 'P' ) || ( prev == 'S' ) || ( prev == 'T' ) )
						ch = 0x00;
				}
				else if ( ( prev == 'A' ) || ( prev == 'E' ) || ( prev == 'I' ) || ( prev == 'O' ) || ( prev == 'U' ) )
					if ( ( Upper[i + 1] != 'A' ) && ( Upper[i + 1] != 'E' ) && ( Upper[i + 1] != 'I' ) && ( Upper[i + 1] != 'O' ) && ( Upper[i + 1] != 'U' ) )
						ch = 0x00;

				break;

			case 'P' :
				if ( Upper[i + 1] == 'H' ) {
					ch = 'F';
					i++;
				}

				break;

			case 'Q' :
				ch = 'K';
				break;

			case 'S' :
				if ( Upper[i + 1] == 'I' )
					if ( ( Upper[i + 2] == 'A' ) || ( Upper[i + 2] == 'O' ) )
						ch = 'X';

				if ( Upper[i + 1] == 'H' ) {
					ch = 'X';
					i++;
				}

				if ( !Compatible && Upper[i + 1] == 'C' )
					if ( Upper[i + 2] == 'H' ) {
						if ( ( Upper[i + 3] == 'M' ) || ( Upper[i + 3] == 'N' ) || ( Upper[i + 3] == 'W' ) )
							ch = 'X';
						else {
							Temp[j++] = ch;
							Temp[j] = 0x00;
							if ( j >= Length )
								return 0;

							ch = 'K';
						}

						i += 2;
					}

				break;

			case 'T' :
				if ( Upper[i + 1] == 'I' )
					if ( ( Upper[i + 2] == 'A' ) || ( Upper[i + 2] == 'O' ) )
						ch = 'X';

				if ( ( Upper[i + 1] == 'H' ) && ( prev != '*' ) ) {
					ch = ch2 = '0';
					i++;
				}

				if ( Upper[i + 1] == 'T' )
					if ( Upper[i + 2] == 'H' ) {
						ch = ch2 = '0';
						i += 2;
					}

				if ( Upper[i + 1] == 'C' )
					if ( Upper[i + 2] == 'H' )
						ch = 0x00;

				break;

			case 'V' :
				ch = 'F';
				break;

			case 'W' :
				if ( ( prev != 'X' ) && ( Upper[i + 1] != 'A' ) && ( Upper[i + 1] != 'E' ) && ( Upper[i + 1] != 'I' ) && ( Upper[i + 1] != 'O' ) && ( Upper[i + 1] != 'U' ) )
					ch = 0x00;

				break;

			case 'X' :
				Temp[j++] = 'K';
				Temp[j] = 0x00;

				ch = 'S';
				if ( j >= Length )
					ch = 0x00;

				break;

			case 'Y' :
				if ( ( Upper[i + 1] != 'A' ) && ( Upper[i + 1] != 'E' ) && ( Upper[i + 1] != 'I' ) && ( Upper[i + 1] != 'O' ) && ( Upper[i + 1] != 'U' ) )
					ch = 0x00;
				else if ( !Compatible && ( prev == 'R' ) )
					if ( Upper[i + 1] == 'A' )
						if ( Upper[i + 2] == 'N' )
							ch = 0x00;

				break;

			case 'Z' :
				ch = 'S';
				break;

			case 'F' :
			case 'J' :
			case 'K' :
			case 'L' :
			case 'M' :
			case 'N' :
			case 'R' :
				break;

			default:
				if ( prev )
					ch = 0x00;
		}

		if ( ch ) {
			Temp[j++] = ch;
			Temp[j] = 0x00;

			if ( j >= Length )
				break;
		}

	}

	wcscpy_s( outCode, MP_SHORT_BUF_LEN, Temp );
	return 0;
}


int Metaphone( LPTSTR inWord, LPTSTR outCode )
{
	if ( ( inWord == NULL ) || ( outCode == NULL ) )
		return -1;
	else if ( inWord[0] == 0x00 )
		return -1;

	unsigned long int Length = METAPHONE_MIN;
	BOOL Compatible = FALSE;

	wchar_t temp[MP_SHORT_BUF_LEN] = L"";
	wchar_t *sep1 = NULL;

	GetEnvironmentVariable( L"METAPHONE_LENGTH", temp, MP_SHORT_BUF_LEN );

	_set_errno( 0 );

	if ( ( temp[0] == '0' ) && ( towupper( temp[1] ) == 'X' ) && ( iswxdigit( temp[2] ) ) )
		Length = wcstoul( temp + 2, &sep1, 16 );				//  parse hexadecimal
	else if ( iswdigit( temp[0] ) )
		Length = wcstoul( temp, &sep1, 10 );					//  parse decimal

	if ( errno )
		Length = 0;

	if ( sep1 != NULL )
		if ( *sep1 != '\0' )
			Length = 0;

	if ( ( Length < METAPHONE_MIN ) || ( Length > METAPHONE_MAX ) )
		Length = METAPHONE_MIN;

	temp[0] = 0x00;

	GetEnvironmentVariable( L"METAPHONE_COMPATIBLE", temp, MP_SHORT_BUF_LEN );
	if ( temp[0] )
		Compatible = TRUE;

	return Metaphone( inWord, outCode, Length, Compatible );
}
